#!/usr/bin/perl -w

use Carp;
use Grid qw(:db :types :logics);
use Term::ReadLine;
use File::PathConvert qw(splitpath);
use XML::Parser;

$_plot_file = '.grid-tmp';

=pod

=head1 NAME

gdsh - grid shell - a Perl shell for raster-based GIS work

=head1 SYNOPSIS

    gdsh.pl <options>

=head1 DESCRIPTION

gdsh is an extension of the basic Perl shell:

    while (<>) {
	eval;
	print $@;
    }

=head1 BASIC FUNCTIONALITY

gdsh use's Grid, thus grids and PGPLOT graphics can be used
readily. gdsh adds command line editing and history with
Term::ReadLine. Database connectivity is added using the functionality
provided by Grid module. gdsh can open a pipe to gnuplot for plotting
data.

=head1 OPTIONS AND ARGUMENTS

=item B<-a>

Same as specifying both B<--db-connection> and B<--gnuplot>. Tries to
open a database connection into Grid module and a pipe to gnuplot.

=item B<--db-connection>

Tries to open a database connection into Grid module. The default
hostname ('' by Grid.pm) can be overridden with option B<-h>
I<hostname> or argument B<--hostname=>I<hostname>. The default
database (the name of the current directory by Grid.pm) can be
overridden with argument B<--database=>I<database>. The default
username with which to connect (the effective userid by Grid.pm) can
be overridden with argument B<--username=>I<username>.

=item B<--gnuplot>

Tries to open (an unbuffered) pipe GNUPLOT to gnuplot.

Options can be set from the gdsh command line with command
C<options(..list of options and arguments...)>

=cut

sub options {
    close(GNUPLOT) if $_options{gnuplot};
    db_close() if $_options{db_connection};
    %_options = ();
    my $i = 0;
    while ($i <= $#_) {
	$_ = $_[$i];
	if (/^--database=(\w+)/) {
	    $_options{database} = $1;
	} elsif (/^--username=(\w+)/) {
	    $_options{username} = $1;
	} elsif (/^--hostname=(\w+)/) {
	    $_options{hostname} = $1;
	} elsif (/^-h/) {
	    $i++;
	    $_options{hostname} = $_[$i];
	} elsif (/^--db-connection/) {
	    $_options{db_connection} = 1;
	} elsif (/^--gnuplot/) {
	    $_options{gnuplot} = 1;
	} elsif (/^-a/) {
	    $_options{db_connection} = 1;
	    $_options{gnuplot} = 1;
	} else {
	    carp "unknown argument: $_\n";
	}
	$i++;
    }
    if ($_options{gnuplot}) {
	open GNUPLOT,"| gnuplot" or carp "can't open gnuplot: $!\n";
	select(GNUPLOT); $| = 1;
	select(STDOUT);
    }
    db_connect(\%_options,{PrintError=>1}) if $_options{db_connection};
}

=pod

=head1 COMMANDS

=head2 output

Command C<output(filename)> directs all output to file F<filename>.
If filename is not given, directs all output to STDOUT.

=cut

sub output {
    my($fn) = @_;
    if ($fn) {
	open OUTPUT, ">$fn" or croak "can't open $fn: $!\n";
	select OUTPUT;
    } else {
	close(OUTPUT);
	select STDOUT;
    }
}

=pod

=head2 p

Command C<p> is the standard perl print except when the first
parameter is a reference to a hash or to an array.  If the first
parameter is a hash reference, C<p> sorts the hash numerically and
prints it using two columns (separated by single space) or several
columns if the values are references to arrays. If the first parameter
is a hash reference, C<p> prints the array elements, each on its own
row. If the element is a reference to an array, it is expanded.

=cut

sub p {
    my($ref) = @_;
    if (ref($ref) eq 'HASH') {
	foreach (sort {$a<=>$b} keys %{$ref}) {
	    my $v = $$ref{$_};
	    if (ref($v) eq 'ARRAY') {
	      print "$_ @{$v}\n";
	    } else {
	      print "$_ $v\n";
	    }
	}
    } elsif (ref($ref) eq 'ARRAY') {
	foreach (@{$ref}) {
	    if (ref($_) eq 'ARRAY') {
	      print "@{$_}\n";
	    } else {
	      print "$_\n";
	    }
	}
    } else {
      print @_;
    }
}

=pod

=head2 plot

Command C<plot(argument,options)> plots the argument using
gnuplot. The argument should be either a reference to a hash (the
returned value of grid method C<contents>, C<histogram>, or some such)
or a string. The options should be a hash, i.e, a list of key, value
-pairs written using the format: key1=>value1, key1=>value1, ...

For example:

C<plot($grid->contents,title=>'cell count of grid',with=>'impulses');>

A hash argument is printed to a temporary file (F<.grid-tmp>)
and the command line

C<plot I<xrange> ".grid-tmp" I<title> with I<with>>

I<range> = '' or [I<keymin-1>:I<keymax+1>] if with equals 'impulses'

I<title> = 'notitle' or 'title B<title>' if B<title> is given in the command line as an option title=>'B<title>'

I<with> = 'lines' or what is given as an option with=>'B<with>'

is piped to gnuplot.

A string argument is piped to gnuplot as a part of the command line

C<plot I<xrange> I<yrange> argument I<title>>

xrange, yrange, and title are empty strings or those given as options.
If argument is a readable file, it is surrounded by double quotation marks.

Gnuplot can be instructed to plot into a png-file using option file=>1.

=cut

sub plot {
    my($this,%o) = @_;
    if ($o{file}) {
	print GNUPLOT "set terminal png\n";
	print GNUPLOT "set output \"plot.png\"\n";
    }
    my $with = $o{with} ? $o{with} : 'lines';
    my $xrange = $o{xrange} ? $o{xrange} : '';
    my $yrange = $o{yrange} ? $o{yrange} : '';
    my $title = $o{title} ? 'title "' . $o{title} . '"' : 'notitle';
    my $using = $o{using} ? 'using ' . $o{using} : 'using 1:2';
    print GNUPLOT "set xdata\n";
    print GNUPLOT "set format x\n";
    if (ref($this) eq 'HASH' or ref($this) eq 'ARRAY') {
	output $_plot_file;
	p $this;
	output;
	if ($with eq 'impulses') {
	    my($minx,$maxx);
	    if (ref($this) eq 'HASH') {
		foreach (keys %{$this}) {
		    $minx = $_ if !defined($minx) or $_ < $minx;
		    $maxx = $_ if !defined($maxx) or $_ > $maxx;
		}
	    } else {
		foreach (@{$this}) {
		    $minx = $$_[0] if !defined($minx) or $$_[0] < $minx;
		    $maxx = $$_[0] if !defined($maxx) or $$_[0] > $maxx;
		}
	    }
	    $minx--;
	    $maxx++;
	    $xrange = "[$minx:$maxx]";
	}
	$this = "\"$_plot_file\"";
    } elsif(ref($this) eq 'Timeseries') {
	$this->save($_plot_file);
	print GNUPLOT "set xdata time\n";
	print GNUPLOT "set timefmt \"%Y%m%d\"\n";
	print GNUPLOT "set format x \"%d.%m\\n%y\"\n";
	$this = "\"$_plot_file\"";
    } else {
	if (-r $this) {
	    $this = "\"$this\"";
	} else {
	    $using = '';
	}
	$title = '' unless $o{title};
    }
    print GNUPLOT "plot $xrange$yrange $this $using $title with $with\n";
    print GNUPLOT "set xdata\n" if ref($this) eq 'Timeseries';
    print GNUPLOT "set terminal x11\n" if $o{file};
}

=pod

=head2 slurp

Command C<slurp(filename,options)> reads the contents of a file into a
hash (or an array, if option array=>1 is given). It is assumed that
the file contains data in two (for hash) or more columns (for a
array). In the case of an array the array values are references to
arrays which each hold the values of one row of data.

=cut

sub slurp {
    my($file,%o) = @_;
    if (!open S,$file) {
	print("$file: $!\n");
	return;
    }
    my $ret;
    if ($o{array}) {
	$ret = [];
    } else {
	$ret = {};
    }
    while (<S>) {
	chomp;
	s/^\s+//;
	s/\s+$//;
	next if /^#/;
	next if $_ eq '';
	last if $_ eq '__END__';
	my(@l) = split /[\s,]+/;
	if ($o{array}) {
	    my $l = [@l];
	    $ret->[$#$ret+1] = $l;
	} else {
	    $ret->{$l[0]} = $l[1];
	}
    }
    close S;
    return $ret;
}

sub nh {
    my($gd,$i0,$j0,$k) = @_;
    $k = 1 if !$k;
    my($i,$j);
    for $i ($i0-$k..$i0+$k) {
	for $j ($j0-$k..$j0+$k) {
	    print $gd->get($i,$j),' ';
	}
	print("\n");
    }
}

=pod

=head1 vars

Command C<vars> lists all variables, and if they are references, the
type of data which they refer to.

=cut

sub vars {
    my @scalars;
    my @arrays;
    my @hashes;
    foreach (sort keys %main::) {
	next if /^_/;
#	next if /^[A-Z]/;
	next if /^$/;
	next if /^scalars$/;
	next if /^[^a-z]/;
	next if /::/;
	local *sym = $main::{$_};
	my $scalar = ref($sym);
	$scalar = 'scalar' unless $scalar;
	push @scalars, "$scalar \$$_ is defined\n" if defined($sym);
	push @arrays, "array \@$_ is defined\n" if defined(@sym);
	push @hashes, "hash \%$_ is defined\n" if defined(%sym);
    }
    foreach (@scalars) {
	print;
    }
    foreach (@arrays) {
	print;
    }
    foreach (@hashes) {
	print;
    }
}

=pod

=head1 SIGINT

gdsh installs a SIGINT handler which may used to cancel lengthy grid
operations.

=cut

sub cntrlc {
    my($sig) = @_;
    &Grid::gdsigint(1);
}

=pod

=head1 HELP

Commands C<?> and C<help> run C<perldoc gdsh.pl> thus showing this
manual page.

Command C<? Grid> and C<help Grid> run C<man Grid> thus showing the
manual page of the Grid module.

=head1 EXECUTING SYSTEM COMMANDS

A command line which begins with '!' is interpreted as a system
command.

=cut

options(@ARGV);
$SIG{INT} = 'cntrlc';
$_term = new Term::ReadLine 'Grid shell';
while ( defined ($_ = $_term->readline('>')) ) {
    chomp;
    if (/^\?$/ or /^help$/i) {
	system "perldoc $0";
    } elsif (/^\? Grid$/ or /^help Grid$/i) {
	system "man Grid";
    } elsif (/^\!(.*)/) {
	system $1;
    } else {
	eval;
	print $@;
    }
}
db_close() if $_options{db_connection};
if ($_options{gnuplot}) {
    close(GNUPLOT);
    unlink $_plot_file;
}
print "\n";

package Timeseries;

use Carp;
use strict;
use warnings;
use Date::Calc qw/check_date Delta_Days Add_Delta_Days/;

BEGIN {
    use Exporter   ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);

    # set the version for version checking
    $VERSION     = 1.00;

    @ISA         = qw(Exporter);
    @EXPORT      = qw();
    %EXPORT_TAGS = ( );     # eg: TAG => [ qw!name1 name2! ],
    
    # your exported package globals go here,
    # as well as any optionally exported functions
    @EXPORT_OK   = qw();
}
our @EXPORT_OK;

sub new {
    my $class = shift;
    my $self = {};
    bless($self, $class);
    $self->set(@_) if @_;
    return $self;
}

sub plot {
    my $self = shift;
    &main::plot($self,@_);
}

sub set {
    my $self = shift;
    my($data) = @_;
    if (ref($data) eq 'HASH') {
	delete $self->{DATA}->{ARRAY};
	delete $self->{DATA}->{HASH};
	foreach (keys %{$data}) {
	    my $key = $_;
	    my $val = $data->{$key};
	    delete $data->{$key};
	    $key =~ s/[^\d]//g;
	    $self->{DATA}->{HASH}->{$key} = $val;
	}
    } elsif (ref($data) eq 'ARRAY') {
	delete $self->{DATA}->{HASH};
	foreach (@{$data}) {
	    $$_[0] =~ s/[^\d]//g;
	}
	$self->{DATA}->{ARRAY} = $data;
    } else {
	my(%data) = @_;
	# this should check whether data is in array or in hash
	foreach (keys %data) {
	    $self->{DATA}->{HASH}->{$_} = $data{$_};
	}
    }
}

sub at {
    my ($self,$time) = @_;
    if (exists $self->{DATA}->{HASH}) {
	if (exists $self->{DATA}->{HASH}->{$time}) {
	    return $self->{DATA}->{HASH}->{$time};
	}
    } elsif (exists $self->{DATA}->{ARRAY}) {
	foreach (@{$self->{DATA}->{ARRAY}}) {
	    return $$_[1] if $time eq $$_[0];
	}
    }
}

sub time_array {
    my $self = shift;
    my $t;
    if (exists $self->{DATA}->{ARRAY}) {
	foreach (@{$self->{DATA}->{ARRAY}}) {
	    push @{$t},$$_[0];
	}
    } elsif (exists $self->{DATA}->{HASH}) {
	foreach (sort {$a<=>$b} keys %{$self->{DATA}->{HASH}}) {
	    push @{$t},$_;
	}
    }
    return $t;
}

sub interpolate {
    my $self = shift;
    if (exists $self->{DATA}->{ARRAY}) {
	foreach (@{$self->{DATA}->{ARRAY}}) {
	    $self->{DATA}->{HASH}->{$$_[0]} = $$_[1];
	}
	delete $self->{DATA}->{ARRAY};
    }
    # assuming timestep is one day and date is in format YYYYMMDD
    my $prev_d;
    my @t;
    foreach my $d (sort {$a<=>$b} keys %{$self->{DATA}->{HASH}}) {
	my @d = $d =~ /(\d\d\d\d)(\d\d)(\d\d)/;
	croak "interpolate: not a valid date: @d\n" unless check_date(@d);
	my $val = $self->{DATA}->{HASH}->{$d};
	my $prev_val;
	unless ($prev_d) {
	    @t = @d;
	    push @{$self->{DATA}->{ARRAY}},[$d,$val];
	} else {
	    my $k = Delta_Days(@t,@d);
	    croak "interpolate: refusing to interpolate over $k days\n" if $k > 1000;
	    for my $i (1..$k) {
		@t = Add_Delta_Days(@t,1);
		my $t = 10000*$t[0]+100*$t[1]+$t[2];
		my $x;
		$x = $k == 1 ? $val : $prev_val + $i * ($val-$prev_val)/$k;
		push @{$self->{DATA}->{ARRAY}},[$t,$x];
	    }
	}
	$prev_d = $d;
	$prev_val = $val;
    }
    delete $self->{DATA}->{HASH};
    foreach (@{$self->{DATA}->{ARRAY}}) {
	$self->{DATA}->{HASH}->{$$_[0]} = $$_[1];
    }
}

sub print {
    my $self = shift;
    if (exists $self->{DATA}->{ARRAY}) {
	foreach (@{$self->{DATA}->{ARRAY}}) {
	    print "@{$_}\n";
	}
    } elsif (exists $self->{DATA}->{HASH}) {
	my $data = $self->{DATA}->{HASH};
	foreach (sort {$a<=>$b} keys %{$data}) {
	    print $_,' ',$data->{$_},"\n";
	}
    }
}

sub save {
    my($self,$file) = @_;
    open(FILE,">$file") or croak("Can't open $file: $!\n");
    my $s = select;
    select FILE;
    $self->print;
    close(FILE);
    select $s;
}

sub divide_by {
    my($self, $second) = @_;
    if (exists $self->{DATA}->{HASH}) {
	my $data = $self->{DATA}->{HASH};
	foreach (keys %{$data}) {
	    $data->{$_} /= $second;
	}
    } 
    if (exists $self->{DATA}->{ARRAY}) {
	foreach (@{$self->{DATA}->{ARRAY}}) {
	    $_->[1] /= $second;
	}
    }
    return $self;
}

END { }       # module clean-up code here (global destructor)


