#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include "grid.h"
#include "catchment.h"


/* 
   a border cell of an area is a grid cell which has at least one cell
   in its 8-neighborhood which has a different value than the area
   value or the cell is on the border of the grid
   
   this algorithm walks the border of an area and tests each
   bordercell with a given test function
*/
/* 
   should return true if the cell is to be remembered 
   should return -1 in the case of an error (report it!)
 */
typedef int testfct(void *fctparam1, cell c); 
/* 
   should return true if the cell is area cell 
*/
typedef int areafct(void *fctparam1, cell c); 
/* 
   a is the place where to start walking
   out is a direction which is away from the area whose border is to be walked
*/
cell borderwalk(cell a, int out, void *fctparam1, areafct is_area, testfct test)
{
    static char *fct = "borderwalk";
    cell c = a;
    int da = 0;
    int d = out;
    /* search for the first direction which is 
       clockwise an area cell just after direction out */
    cell t = cellmove(c, out);
    cell x = c;

    debug(20,"borderwalk called at %i,%i\n", a.i, a.j);
    if (!is_area(fctparam1, c)) {
	/* a is not on the grid or not on the area! */
	cell error = {-1, -1};
	fprintf(stderr,"%s: the initial cell is not on the area\n",fct);
	return error;
    }
    if (is_area(fctparam1, t)) {
	/* a is on the area! */
	cell error = {-1, -1};
	fprintf(stderr,"%s: the cell in direction out is on the area\n",fct);
	return error;
    }
    while (!is_area(fctparam1, t)) {
	d = NEXTDIR(d);
	if (d == out) /* a is the area */
	    break;
	t = cellmove(c, d);
    }
    if (test(fctparam1, c) == -1) {
	cell error = {-1, -1};
	return error;
    }
    if (d == out) /* a is the area */
	return c;
    while (1) {
	t = cellmove(c, d);
	while (!is_area(fctparam1, t)) {
	    d = NEXTDIR(d);
	    /* there _IS_ a 8-neighbor which is on the area */
	    t = cellmove(c, d);
	}
	/* this is a ghost by feng shui definition, 
	   i.e. this prefers walking to dir 1, 3, 5, or 7 */
	if (EVEN(d)) {
	    int od = NEXTDIR(d);
	    t = cellmove(c, od);
	    if (is_area(fctparam1, t))
		d = od;
	}
	/* test if we are done */
	if (SAME_CELL(c, a)) {
	    if (d == da)
		return x;
	    else if (!da)
		da = d;
	}
	c = cellmove(c, d);
	debug(30,"borderwalk at %i,%i d = %i da = %i x = %i,%i\n", c.i, c.j, d, da, x.i, x.j);
	switch (test(fctparam1, c)) {
	case -1: {
	    cell error = {-1, -1};
	    return error;
	}
	case 0:break;
	case 1:x = c;
	}
	d += 6;
	if (d > 8) d -= 8;
    }
}


/*

  Fits a 9-term quadratic polynomial 

  NOTE: assumes dx = dy as in grid

z = A * x^2y^2 + B * x^2y + C * xy^2 + D * x^2 + E * y^2 + F * xy + G * x + H *y + I

a[1] = A
a[2] = B
a[3] = C
a[4] = D
a[5] = E
a[6] = F
a[7] = G
a[8] = H
a[9] = I

  to a 3*3 square grid centered at the center point of cell c

  z's:   1 2 3
         4 5 6
         7 8 9

  ^
  |
y |
  |
   ---> 
    x

z1 =  1 * A +  1 * B + -1 * C + 1 * D + 1 * E + -1 * F + -1 * G +  1 * H + I
z2 =  0 * A +  0 * B +  0 * C + 0 * D + 1 * E +  0 * F +  0 * G +  1 * H + I
z3 =  1 * A +  1 * B +  1 * C + 1 * D + 1 * E +  1 * F +  1 * G +  1 * H + I
z4 =  0 * A +  0 * B +  0 * C + 1 * D + 0 * E +  0 * F + -1 * G +  0 * H + I
z5 =  0 * A +  0 * B +  0 * C + 0 * D + 0 * E +  0 * F +  0 * G +  0 * H + I
z6 =  0 * A +  0 * B +  0 * C + 1 * D + 0 * E +  0 * F +  1 * G +  0 * H + I
z7 =  1 * A + -1 * B + -1 * C + 1 * D + 1 * E +  1 * F + -1 * G + -1 * H + I
z8 =  0 * A +  0 * B +  0 * C + 0 * D + 1 * E +  0 * F +  0 * G + -1 * H + I
z9 =  1 * A + -1 * B +  1 * C + 1 * D + 1 * E + -1 * F +  1 * G + -1 * H + I

A =  0.25 * z1 +  0.5 * z2 + 0.25 * z3 + -0.5 * z4 +  1 * z5 + -0.5 * z6 +  0.25 * z7 + -0.5 * z8 +  0.25 * z9
B =  0.25 * z1 + -0.5 * z2 + 0.25 * z3 +  0   * z4 +  0 * z5 +  0   * z6 + -0.25 * z7 +  0.5 * z8 + -0.25 * z9
C = -0.25 * z1 +  0   * z2 + 0.25 * z3 +  0.5 * z4 +  0 * z5 + -0.5 * z6 + -0.25 * z7 +  0   * z8 +  0.25 * z9
D =  0    * z1 +  0   * z2 + 0    * z3 +  0.5 * z4 + -1 * z5 +  0.5 * z6 +  0    * z7 +  0   * z8 +  0    * z9
E =  0    * z1 +  0.5 * z2 + 0    * z3 +  0   * z4 + -1 * z5 +  0   * z6 +  0    * z7 +  0.5 * z8 +  0    * z9
F = -0.25 * z1 +  0   * z2 + 0.25 * z3 +  0   * z4 +  0 * z5 +  0   * z6 +  0.25 * z7 +  0   * z8 + -0.25 * z9
G =  0    * z1 +  0   * z2 + 0    * z3 + -0.5 * z4 +  0 * z5 +  0.5 * z6 +  0    * z7 +  0   * z8 +  0    * z9
H =  0    * z1 +  0.5 * z2 + 0    * z3 +  0   * z4 +  0 * z5 +  0   * z6 +  0    * z7 + -0.5 * z8 +  0    * z9
I =  0    * z1 +  0   * z2 + 0    * z3 +  0   * z4 +  1 * z5 +  0   * z6 +  0    * z7 +  0   * z8 +  0    * z9

z_factor is the unit of z dived by the unit of x and y

*/

int fitpoly(grid *dem, double a[], cell c, double z_factor) 
{
    int ix = 1;
    int ii, jj;
    double z[10];
    z[0] = 0;
    for (ii = c.i-1; ii <= c.i+1; ii++) {
	for (jj = c.j-1; jj <= c.j+1; jj++) {
	    cell c2;
	    if (ii < 0) {
		c2.i = 0;
	    } else if (ii > dem->M-1) {
		c2.i = dem->M-1;
	    } else {
		c2.i = ii;
	    }
	    if (jj < 0) {
		c2.j = 0;
	    } else if (jj > dem->N-1) {
		c2.j = dem->N-1;
	    } else {
		c2.j = jj;
	    }
	    if (GD_DATACELL(dem, c2)) 
		z[ix] =  GD_CELL(dem, c2);
	    else
		z[ix] = GD_CELL(dem, c);
	    z[ix] *= (z_factor / dem->unitdist);
	    ix++;
	}
    }
    a[0] = 0;
    a[1] =  0.25 * z[1] +  0.5 * z[2] + 0.25 * z[3] -  0.5 * z[4] + z[5] -  0.5 * z[6] +  0.25 * z[7] -  0.5 * z[8] +  0.25 * z[9];
    a[2] =  0.25 * z[1] -  0.5 * z[2] + 0.25 * z[3]                                    -  0.25 * z[7] +  0.5 * z[8] -  0.25 * z[9];
    a[3] = -0.25 * z[1]               + 0.25 * z[3] +  0.5 * z[4]        -  0.5 * z[6] -  0.25 * z[7]               +  0.25 * z[9];
    a[4] =                                             0.5 * z[4] - z[5] +  0.5 * z[6]                                            ;
    a[5] =                 0.5 * z[2]                             - z[5]                              +  0.5 * z[8]               ;
    a[6] = -0.25 * z[1]               + 0.25 * z[3]                                    +  0.25 * z[7]               -  0.25 * z[9];
    a[7] =                                            -0.5 * z[4]        +  0.5 * z[6]                                            ;
    a[8] =                 0.5 * z[2]                                                                 -  0.5 * z[8]               ;
    a[9] =                                                          z[5]                                                          ;
    return 1;
}


/*
        j
        ->      ^
     i |        | 0  -> Pi/2   | Pi   <- 3*Pi/2
       V                       V
  
*/
double aspect(double a[]) 
{
    double asp;
    if (fabs(a[8]) < EPSILON) {
	if (fabs(a[7]) < EPSILON) return -1;
	if (a[7] < 0) return M_PI/2.0;
	return 3.0/2.0 * M_PI;
    }
    if (fabs(a[7]) < EPSILON) {
	if (a[8] > 0) return M_PI;
	if (a[7] >= 0) return 2.0 * M_PI;
	return 0;
    }
    asp = M_PI - atan2(a[8]/a[7],1) + M_PI/2.0 * (a[7]/fabs(a[7]));
    if (asp<0) return 0;
    if (asp>2*M_PI) return 2*M_PI;
    return asp;
}


grid *dem2aspect(grid *dem)
{
    grid *ag;
    cell c;
    PROGRESS_BAR_VARS(dem->M/40);
    if (!(ag = gdnew2(REAL_GRID, dem))) return NULL;
    FOREACH(c, dem) {
	PROGRESS_CHECK(c.i);
	if (GD_DATACELL(dem, c)) {
	    double a[10];
	    SIGCANCEL;
	    fitpoly(dem, a, c, 1);
	    RGD_CELL(ag, c) = aspect(a);
	    debug(2,"%i,%i G=%f H=%f aspect=%f\n", c.i, c.j, a[7], a[8], RGD_CELL(ag, c));
	} else {
	    RGD_SETNODATACELL(ag, c);
	}
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    return ag;
}


grid *dem2slope(grid *dem, double z_factor)
{
    grid *sg;
    cell c;
    PROGRESS_BAR_VARS(dem->M/40);
    if (!(sg = gdnew2(REAL_GRID, dem))) return NULL;
    FOREACH(c, dem) {
	PROGRESS_CHECK(c.i);
	if (GD_DATACELL(dem, c)) {
	    double a[10];
	    SIGCANCEL;
	    fitpoly(dem, a, c, z_factor);
	    RGD_CELL(sg, c) = atan(sqrt(a[7]*a[7]+a[8]*a[8]));
	} else {
	    RGD_SETNODATACELL(sg, c);
	}
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    return sg;
}


double flatness_threshold(grid *dem)
{
    if (dem->datatype == INTEGER_GRID)
	return 0;
    else
	return 0.001;
}


grid *dem2fdg(grid *dem, int method) 
{
    static char *fct = "dem2fdg";
    grid *fdg;
    cell c;
    double dz = flatness_threshold(dem);
    PROGRESS_BAR_VARS(dem->M/40);
    if (!(fdg = gdnew2(INTEGER_GRID, dem))) return NULL;
    FOREACH(c, dem) {
	PROGRESS_CHECK(c.i);
	if (GD_NODATACELL(dem, c)) 
	    IGD_SETNODATACELL(fdg, c);
	else {
	    /* 
	       test for all eight-neighbors of c and 
	       remember the direction to the one with steepest descent to 
	       if a nodata available flow there
	       if on the border of the grid, flow outside
	    */
	    int dir, found = 0, dir_t = 0, has_non_higher = 0, many = 0;
	    double zc = GD_CELL(dem, c), descent = 0;
	    SIGCANCEL;

	    FOREACHDIR(dir) {
		cell t = cellmove(c, dir);
		if (GD_CELLIN(dem, t) AND GD_DATACELL(dem, t)) {
		    double zt = GD_CELL(dem, t);
		    if (zt < zc - dz) {
			double coeff = 1;
			if (EVEN(dir)) {
			    if (method == RHO8) {
				double rho = ((double)rand()) / RAND_MAX;
				coeff = 1.0/(2.0-rho);
			    } else {
				coeff = sqrt(1.0/2.0);
			    }
			}
			if (!found OR coeff*(zc-zt) > descent) {
			    found = 1;
			    descent = coeff*(zc-zt);
			    dir_t = dir;
			}
			many |= 1 << (dir-1); /* bits in the byte tell downslope cells */
		    } else if (zt <= zc + dz) {
			has_non_higher = 1;
		    }
		} else {
		    found = 1;
		    dir_t = dir;
		    many = 1 << (dir-1);
		    break;
		}
	    }
	    if (found) {
		if (method == MANY8)
		    dir = many;
		else
		    dir = dir_t;
	    } else if (has_non_higher)
		dir = FLAT_AREA;
	    else 
		dir = PIT_CELL;
	    debug(10,"%s: %i,%i dir=%i\n", fct, c.i, c.j, dir);
	    IGD_CELL(fdg, c) = dir;
	}
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    return fdg;
}


grid *dem2ucg(grid *dem) 
{
    grid *ucg;
    cell c;
    double dz = flatness_threshold(dem);
    PROGRESS_BAR_VARS(dem->M/40);
    if (!(ucg = gdnew2(INTEGER_GRID, dem))) return NULL;
    FOREACH(c, dem) {
	PROGRESS_CHECK(c.i);
	if (GD_NODATACELL(dem, c)) 
	    IGD_SETNODATACELL(ucg, c);
	else {
	    int dir, many = 0;
	    double zc = GD_CELL(dem, c);
	    SIGCANCEL;
	    FOREACHDIR(dir) {
		cell t = cellmove(c, dir);
		if (GD_CELLIN(dem, t) AND GD_DATACELL(dem, t)) {
		    double zt = GD_CELL(dem, t);
		    if (zt > zc + dz) {
			many |= 1 << (dir-1); /* bits in the byte tell upslope cells */
		    } 
		}
	    }
	    IGD_CELL(ucg, c) = many;
	}
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    return ucg;
}


int init_pour_point_struct(pour_point_struct *pp, grid *fdg, grid *dem, grid *mark)
{
    static char *fct = "init_pour_point_struct";
    if (!mark) return 0;
    if ((dem AND !gdoverlayable(fdg, dem)) OR !gdoverlayable(fdg, mark)) {
	fprintf(stderr,"%s: grids are not overlayable\n", fct);
	return 0;
    }
    if (fdg->datatype != INTEGER_GRID OR mark->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: fdg and mark have to be integer grids\n", fct);
	return 0;
    }
    pp->fdg = fdg;
    pp->dem = dem;
    pp->mark = mark;
    if (dem) pp->dz = flatness_threshold(dem);
    pp->bc_found = 0;
    pp->test_inner = 1;
    pp->pp_found = 0;
    pp->pour_to_nodata = 0;
    return 1;
}


int markflat(pour_point_struct *pp, cell c)
{
    int size = 1, d;
    debug(15,"markflat at %i,%i\n", c.i, c.j);
    IGD_CELL(pp->mark, c) = 1;
    if (!pp->bc_found OR c.i < pp->bc.i) {
	debug(20,"markflat new bc %i,%i\n", c.i, c.j);
	pp->bc_found = 1;
	pp->bc = c;
	pp->dir_out = 1;  /* this finds the topmost cell (min i) so 1 is away from the area */
    }
    for (d = 1; d < 9; d++) {
	cell t = cellmove(c, d);
	if (GD_CELLIN(pp->fdg, t) AND IGD_DATACELL(pp->fdg, t) AND !IGD_CELL(pp->mark, t) AND 
	    fabs(GD_CELL(pp->dem, t) - pp->zf) <= pp->dz)
	    size += markflat(pp, t);
    }
    return size;
}


/* this is called for each candidate inner pour point (border cell) */
int test_pour_point(void *fctparam1, cell c) {
    pour_point_struct *pp = (pour_point_struct *)fctparam1;
    int ret = 0;
    int d, d_in;
    double zc, slope_in = 0;

    zc = GD_CELL(pp->dem, c);
    d_in = IGD_CELL(pp->fdg, c);
    if (d_in) slope_in = (zc - GD_CELL(pp->dem, cellmove(c, d_in)))/DISTANCE_UNIT(d_in);

    debug(15,"searching for pour point at %i,%i z=%f\n", c.i, c.j, RGD_CELL(pp->dem,c));
    if (pp->pour_to_nodata) {
	if (zc < pp->z_ipp) {
	    pp->z_ipp = zc;
	} else
	    return 0;
    }
    /*if (pp->pour_to_nodata) return 0;*/
    for (d = 1; d < 9; d++) {
	cell t = cellmove(c, d);
	double zt, slope_out;
	if (GD_CELLOUT(pp->fdg, t) OR IGD_NODATACELL(pp->fdg, t)) {
	    debug(20,"  new ipp (pour to nodata) %i,%i\n", c.i, c.j);
	    pp->pp_found = 1;
	    pp->ipp = c;
	    pp->z_ipp = zc;
	    pp->in2out = d;
	    pp->pour_to_nodata = 1;
	    ret = 1;
	    break;
	}
	if (IGD_CELL(pp->mark, t)) continue;
	zt = GD_CELL(pp->dem, t);
	slope_out = (zc - zt)/DISTANCE_UNIT(d);
	if (!pp->pp_found OR 
	    (pp->test_inner AND 
	     (zc < pp->z_ipp OR 
	      (zc == pp->z_ipp AND slope_out > pp->slope_out) OR
	      (zc == pp->z_ipp AND slope_out == pp->slope_out AND d_in AND slope_in > pp->slope_in))) OR
	    (!pp->test_inner AND 
	     (zt < pp->z_opp OR 
	      (zt == pp->z_opp AND DISTANCE_UNIT(d) < pp->dio)))) {
	    debug(20,"  new ipp %i,%i\n", c.i, c.j);
	    pp->pp_found = 1;
	    pp->ipp = c;
	    pp->z_ipp = zc;
	    pp->slope_in = slope_in;
	    pp->slope_out = slope_out;
	    pp->dio = DISTANCE_UNIT(d);
	    pp->in2out = d;
	    pp->opp = t;
	    pp->z_opp = zt;
	    ret = 1;
	}
    }
    return ret;
}


int is_marked(void *fctparam1, cell c) {
    return GD_CELLIN(((pour_point_struct *)fctparam1)->mark, c) AND 
	IGD_CELL(((pour_point_struct *)fctparam1)->mark, c);
}


void drain_flat_area_to_pour_point(pour_point_struct *pp, cell c) 
{
    int d, invdir = 5;
    debug(15,"drain flat to pour point at %i,%i\n", c.i, c.j);
    IGD_CELL(pp->mark, c) = 0;
    for (d = 1; d < 9; d++) {
	cell t = cellmove(c, d);
	if (GD_CELLIN(pp->fdg, t) AND IGD_DATACELL(pp->fdg, t) AND IGD_CELL(pp->mark, t)) {
	    IGD_CELL(pp->fdg, t) = invdir;
	    drain_flat_area_to_pour_point(pp, t);
	}
	invdir++;
	if (invdir > 8) invdir = 1;
    }
}


void raise_area(pour_point_struct *pp, cell c, double z) 
{
    int d;
    debug(5,"raise area at %i,%i from %f to %f\n", c.i, c.j, GD_CELL(pp->dem, c), z);
    IGD_CELL(pp->mark, c) = 0;
    GD_CELL(pp->dem, c) = z;
    for (d = 1; d < 9; d++) {
	cell t = cellmove(c, d);
	if (GD_CELLIN(pp->fdg, t) AND IGD_DATACELL(pp->fdg, t) AND IGD_CELL(pp->mark, t)) {
	    raise_area(pp, t, z);
	}
    }
}


int flat_and_pit_cells(grid *fdg)
{
    cell c;
    int count = 0;
    FOREACH(c, fdg)
	if (IGD_CELL(fdg, c) == FLAT_AREA OR IGD_CELL(fdg, c) == PIT_CELL) count++;
    return count;
}


int fdg_fixflats1(grid *fdg, grid *dem) 
{
    static char *fct = "fdg_fixflats1";
    double dz = flatness_threshold(dem);
    double zf; /* elevation of the flat area */
    cell c;
    int done = -1;
    int fixed_flats = 0;
    PROGRESS_BAR_VARS(flat_and_pit_cells(fdg)/40);
    if (!gdoverlayable(fdg, dem)) {
	fprintf(stderr,"%s: dem and fdg are not overlayable\n", fct);
	return -1;
    }
    if (fdg->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: fdg has to be an integer grid\n", fct);
	return -1;
    }
    while (done != fixed_flats) {
	done = fixed_flats;
	PROGRESS_CHECK(fixed_flats);
	FOREACH(c, fdg) {
	    SIGCANCEL;
	    if (IGD_CELL(fdg, c) == FLAT_AREA OR IGD_CELL(fdg, c) == PIT_CELL) {	    
		/* 
		   are there non-higher neigbors with dge resolved?
		   or are there outside cells in the neighborhood?
		*/
		double zl = 0, zt;
		int dir = 0, d, invdir = 5;
		int is_on_border = 0;
		zf = GD_CELL(dem, c);
		for (d = 1; d < 9; d++) {
		    int dt;
		    cell t = cellmove(c, d);
		    if (GD_CELLOUT(fdg, t) OR IGD_NODATACELL(fdg, t)) {
			is_on_border = d;
			continue;
		    }
		    zt = GD_CELL(dem, t);
		    dt = IGD_CELL(fdg, t);
		    if (dt > 0 AND dt != invdir AND
			zt <= zf + dz AND
			(!dir OR zt < zl)) {
			zl = zt;
			dir = d;
		    }
		    invdir++;
		    if (invdir > 8) invdir = 1;
		}
		if (dir) {
		    IGD_CELL(fdg, c) = dir;
		    fixed_flats++;
		} else if (is_on_border) {
		    IGD_CELL(fdg, c) = is_on_border;
		    fixed_flats++;
		}
	    }
	}
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    return fixed_flats;
}


int fdg_fixflats2(grid *fdg, grid *dem) 
{
    cell c;
    int fixed_flats = 0;
    pour_point_struct pp;
    PROGRESS_BAR_VARS(fdg->M/40);
    if (!init_pour_point_struct(&pp, fdg, dem, gdnew2(INTEGER_GRID, fdg))) return -1;
    FOREACH(c, fdg) {
	PROGRESS_CHECK(c.i);
	SIGCANCEL;
	if (IGD_CELL(fdg, c) != FLAT_AREA) continue;

	pp.zf = GD_CELL(pp.dem, c);

	/* mark the flat area and find a definitive border cell */
	pp.bc_found = 0;
	markflat(&pp, c);

	/* walk the border */
	pp.test_inner = 0;
	pp.pp_found = 0;
	pp.pour_to_nodata = 0;
	borderwalk(pp.bc, pp.dir_out, &pp, &is_marked, &test_pour_point);

	/* if the elevation of the outer pour point is not higher than inner pour point */
	/* or the outer pour point is nodata */
	/* drain there, else make inner pour point a pit cell */
	if (pp.pour_to_nodata OR pp.z_opp <= GD_CELL(pp.dem, pp.ipp))
	    IGD_CELL(pp.fdg, pp.ipp) = pp.in2out;
	else
	    IGD_CELL(pp.fdg, pp.ipp) = PIT_CELL;

	/* starting from the inner pour point make the whole flat area
	drain into it, this unmarks the flat area along the way */

	drain_flat_area_to_pour_point(&pp, pp.ipp); 

	fixed_flats++;
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    gddestroy(pp.mark);
    return fixed_flats;
}


int rmb(pour_point_struct *pp, cell c, int m) 
{
    int size = 1;
    int dir, invdir = 5;
    IGD_CELL(pp->mark, c) = m;
    if (!pp->bc_found OR c.i < pp->bc.i) {
	pp->bc_found = 1;
	pp->bc = c;
	pp->dir_out = 1;  /* this finds the topmost cell (min i) so 1 is away from the area */
    }
    debug(23,"marking basin, at %i,%i\n", c.i, c.j);
    FOREACHDIR(dir) {
	cell t = cellmove(c, dir);
	if (
	    GD_CELLIN(pp->mark, t) AND IGD_DATACELL(pp->fdg, t) AND !IGD_CELL(pp->mark, t) AND 
	    IGD_CELL(pp->fdg, t) == invdir
	    )
	    size += rmb(pp, t, m);
	invdir++;
	if (invdir > 8) invdir = 1;
    }
    return size;
}


/* recursively mark a basin, using fdg and dem, return the size */
int rmb2(pour_point_struct *pp, cell c, int m) 
{
    int size = 1;
    int dir, invdir = 5;
    double zc = GD_CELL(pp->dem, c);
    IGD_CELL(pp->mark, c) = m;
    if (!pp->bc_found OR c.i < pp->bc.i) {
	pp->bc_found = 1;
	pp->bc = c;
	pp->dir_out = 1;  /* this finds the topmost cell (min i) so 1 is away from the area */
    }
    debug(13,"marking basin, at %i,%i\n", c.i, c.j);
    FOREACHDIR(dir) {
	cell t = cellmove(c, dir);
	double zt = GD_CELL(pp->dem, t);
	if (
	    GD_CELLIN(pp->mark, t) AND IGD_DATACELL(pp->fdg, t) AND !IGD_CELL(pp->mark, t) AND 
	    (IGD_CELL(pp->fdg, t) == invdir OR 
	     (IGD_CELL(pp->fdg, t) == FLAT_AREA AND zt > zc - pp->dz))
	    )
	    size += rmb2(pp, t, m);
	invdir++;
	if (invdir > 8) invdir = 1;
    }
    return size;
}


int fillpit(pour_point_struct *pp, grid *gd, cell c, int m, double zmax, int fill) 
{
    int size = 1;
    int dir, invdir = 5;
    IGD_CELL(gd, c) = m;
    debug(13,"filling pit below zmax, at %i,%i %f->%f\n", c.i, c.j, GD_CELL(pp->dem, c), zmax);
    if (fill) GD_CELL(pp->dem, c) = zmax;
    FOREACHDIR(dir) {
	cell t = cellmove(c, dir);
	double zt = GD_CELL(pp->dem, t);
	if (GD_CELLIN(gd, t) AND IGD_DATACELL(pp->fdg, t) AND !IGD_CELL(gd, t) AND
	    IGD_CELL(pp->fdg, t) == invdir AND zt < zmax)
	    size += fillpit(pp, gd, t, m, zmax, fill);
	invdir++;
	if (invdir > 8) invdir = 1;
    }
    return size;
}


grid *dem_depressions(grid *dem, grid *fdg, int inc_m)
{
    static char *fct = "dem_depressions";
    long m = 1;
    cell c;
    pour_point_struct pp;
    grid *gd;
    PROGRESS_BAR_VARS(fdg->M/40);
    if (!init_pour_point_struct(&pp, fdg, dem, gdnew2(INTEGER_GRID, fdg))) return NULL;
    if (!(gd = gdnew2(INTEGER_GRID, fdg))) {
	gddestroy(pp.mark);
	return NULL;
    }
    FOREACH(c, fdg) {
	PROGRESS_CHECK(c.i);
	SIGCANCEL;
	if (IGD_CELL(fdg, c) == PIT_CELL) {
	    double zmax;
	    debug(3,"pit at %i,%i\n", c.i, c.j);

	    pp.bc_found = 0;
	    rmb(&pp, c, 1);
	    
	    pp.pp_found = 0;
	    pp.pour_to_nodata = 0;
	    borderwalk(pp.bc, pp.dir_out, &pp, &is_marked, &test_pour_point);

	    igdfloodfill(pp.mark, NULL, c, 0, 8);
	    zmax = GD_CELL(dem, pp.ipp);
	    if (!pp.pour_to_nodata) zmax = max(zmax, GD_CELL(dem, pp.opp));
	    fillpit(&pp, gd, c, m, zmax, 0);

	    if (inc_m) m++;
	    if (m > INTEGER_MAX) {
		PROGRESS_BAR_CLOSE;
		gddestroy(gd);
		gddestroy(pp.mark);
		fprintf(stderr,"%s: too many pits\n",fct);
		return NULL;
	    }
	} 
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    gddestroy(pp.mark);
    return gd;
}


int raise_pitcells(grid *dem, grid *fdg)
{
    static char *fct = "raise_pitcells";
    cell c;
    FOREACH(c, fdg) {
	if (IGD_CELL(fdg, c) == PIT_CELL) {
	    double zmin = 0;
	    int zmin_found = 0, d;
	    for (d = 1; d < 9; d++) {
		cell t = cellmove(c, d);
		double zt = GD_CELL(dem, t);
		if (GD_CELLIN(fdg, t) AND GD_DATACELL(fdg, t) AND
		    (!zmin_found OR zt < zmin)) {
		    zmin_found = 1;
		    zmin = GD_CELL(dem, t);
		}
	    }
	    if (zmin_found) {
		GD_CELL(dem, c) = zmin;
		IGD_CELL(fdg, c) = FLAT_AREA;
	    } else {
		fprintf(stderr, "%s: can't fill all single cell pits\n", fct);
		return 0;
	    }
	}
    }
    return 1;
}


int dem_fillpits(grid *dem, double z_limit)
{
    cell c;
    int pits_filled = 0;
    FOREACH(c, dem) {
	int d, f = 0;
	double zc, z_lowest_nbor = 0;
	if (GD_NODATACELL(dem, c)) continue;
	zc = GD_CELL(dem, c);
	FOREACHDIR(d) {
	    cell t = cellmove(c, d);
	    if (GD_CELLIN(dem, t) AND GD_DATACELL(dem, t)) {
		double zt = GD_CELL(dem, t);
		if (!f OR zt < z_lowest_nbor) {
		    f = 1;
		    z_lowest_nbor = zt;
		    if (z_lowest_nbor < zc + z_limit) {
			f = 0;
			break;
		    }
		}
	    }
	}
	if (f) {
	    GDSET(dem, c, z_lowest_nbor);
	    pits_filled++;
	}
    }
    return pits_filled;
}


int dem_cutpeaks(grid *dem, double z_limit)
{
    cell c;
    int peaks_cut = 0;
    FOREACH(c, dem) {
	int d, f = 0;
	double zc, z_highest_nbor = 0;
	if (GD_NODATACELL(dem, c)) continue;
	zc = GD_CELL(dem, c);
	FOREACHDIR(d) {
	    cell t = cellmove(c, d);
	    if (GD_CELLIN(dem, t) AND GD_DATACELL(dem, t)) {
		double zt = GD_CELL(dem, t);
		if (!f OR zt > z_highest_nbor) {
		    f = 1;
		    z_highest_nbor = zt;
		    if (z_highest_nbor > zc - z_limit) {
			f = 0;
			break;
		    }
		}
	    }
	}
	if (f) {
	    GDSET(dem, c, z_highest_nbor);
	    peaks_cut++;
	}
    }
    return peaks_cut;
}


int dem_filldepressions(grid *dem, grid *fdg)
{
    /*static char *fct = "dem_filldepressions";*/
    cell c;
    int pits_filled = 0;
    pour_point_struct pp;
    PROGRESS_BAR_VARS(fdg->M/40);
    if (!raise_pitcells(dem, fdg)) return 0;
    if (!init_pour_point_struct(&pp, fdg, dem, gdnew2(INTEGER_GRID, fdg))) return 0;
    FOREACH(c, fdg) {
	PROGRESS_CHECK(c.i);
	SIGCANCEL;
	if (IGD_CELL(fdg, c) == FLAT_AREA) {
	    pp.zf = GD_CELL(dem, c);
	    debug(3,"pit at %i,%i z=%f\n", c.i, c.j, pp.zf);
	    while(1) {
		double z = 0;
		pp.bc_found = 0;
		markflat(&pp, c);
	    
		pp.test_inner = 0;
		pp.pp_found = 0;
		pp.pour_to_nodata = 0;
		borderwalk(pp.bc, pp.dir_out, &pp, &is_marked, &test_pour_point);
		if (pp.pour_to_nodata OR GD_CELL(dem, pp.opp) <= pp.zf) {
		    igdfloodfill(pp.mark, NULL, c, 0, 8);
		    break;
		}

		z = pp.z_opp;
		raise_area(&pp, c, z);
		pp.zf = z;
	    }
	    pits_filled++;

	} 
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    gddestroy(pp.mark);
    return pits_filled;
}

/* fdg should have only pits and valid dirs */
int dem_breach(grid *dem, grid *fdg, int limit)
{
    /*static char *fct = "dem_breach";*/
    cell c;
    pour_point_struct pp;
    PROGRESS_BAR_VARS(fdg->M/40);
    if (!init_pour_point_struct(&pp, fdg, dem, gdnew2(INTEGER_GRID, fdg))) return 0;
    FOREACH(c, fdg) {
	PROGRESS_CHECK(c.i);
	SIGCANCEL;
	if (IGD_CELL(fdg, c) == PIT_CELL) {

	    double z1 = GD_CELL(dem, c), z2, l_in, l_out, l;
	    cell flow;
	    int dir, b_count = 0;

	    pp.bc_found = 0;
	    rmb(&pp, c, 1);

	    pp.pp_found = 0;
	    pp.pour_to_nodata = 0;
	    borderwalk(pp.bc, pp.dir_out, &pp, &is_marked, &test_pour_point);
	    igdfloodfill(pp.mark, NULL, c, 0, 8);

	    /* The breach should now be done from ipp to the pit cell
	       (elev = z1) and from opp to the border of data or to
	       cell with elev same or lower than pit elev (elev =
	       z2). Take z1 and z2 and interpolate z's along the
	       breach. */

	    flow = pp.ipp;
	    l_in = 0;
	    while (!SAME_CELL(flow, c)) {
		if (GD_CELL(dem, flow) > z1 + pp.dz) b_count++;
		l_in += DISTANCE_UNIT(IGD_CELL(fdg, flow));
		flow = FLOW(fdg, flow);
	    }
	    
	    if (pp.pour_to_nodata) {
		z2 = GD_CELL(dem, pp.ipp);
		l_out = 0;
	    } else {
		l_out = DISTANCE_UNIT(pp.in2out);
		flow = pp.opp;
		b_count++;
		while (1) {
		    z2 = GD_CELL(dem, flow);
		    if (z2 <= z1 + pp.dz) break;
		    flow = FLOW(fdg, flow);
		    b_count++;
		    /* stop searching for the end of breaching if */
		    if (GD_CELLOUT(fdg, flow) OR GD_NODATACELL(fdg, flow) OR IGD_CELL(fdg, flow) < 1) break;
		    l_out += DISTANCE_UNIT(IGD_CELL(fdg, flow));
		}
	    }

	    /* give up if breaching would require too many cells to be
	       breached or z2 > z1 and not drain to nodata */
	    if (limit > 0 AND b_count > limit) continue;
	    if (z1 < z2-pp.dz AND GD_CELLIN(fdg, flow) AND GD_DATACELL(fdg, flow)) continue;
	    
	    flow = pp.ipp;
	    dir = pp.in2out;
	    l = 0;
	    if (SAME_CELL(flow, c)) {
		IGD_CELL(fdg, flow) = dir;
	    } else {
		while (!SAME_CELL(flow, c)) {
		    int tmp = IGD_CELL(fdg, flow);
		    cell down = FLOW(fdg, flow);
		    
		    double z = z1 < z2-pp.dz ? z1 : z1 - (z1-z2)*(l_in-l)/(l_in+l_out);
		    GDSET(dem, flow, z);
		    
		    /* invert the path */
		    IGD_CELL(fdg, flow) = dir;
		    dir = INVDIR(tmp);
		    
		    l += DISTANCE_UNIT(IGD_CELL(fdg, flow));
		    flow = down;
		}
	    }
		     
	    if (!pp.pour_to_nodata) {
		l = DISTANCE_UNIT(pp.in2out);
		flow = pp.opp;
		while (1) {
		    double old_z = GD_CELL(dem, flow), z;
		    if (old_z <= z1 + pp.dz) break;
		    z = z1 < z2-pp.dz ? z1 : z1 - (z1-z2)*(l_in+l)/(l_in+l_out);
		    GDSET(dem, flow, z);
		    flow = FLOW(fdg, flow);
		    /* stop breaching if */
		    if (GD_CELLOUT(fdg, flow) OR GD_NODATACELL(fdg, flow) OR IGD_CELL(fdg, flow) < 1) break;
		    l += DISTANCE_UNIT(IGD_CELL(fdg, flow));
		}
	    }

	} 
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    gddestroy(pp.mark);
    return 1;
}


int fdg_fixpits(grid *fdg, grid *dem) 
{
    cell c;
    int fixed_pits = 0;
    pour_point_struct pp;
    PROGRESS_BAR_VARS(fdg->M/40);
    if (!init_pour_point_struct(&pp, fdg, dem, gdnew2(INTEGER_GRID, fdg))) return -1;
    FOREACH(c, fdg) {
	PROGRESS_CHECK(c.i);
	SIGCANCEL;
	if (IGD_CELL(fdg, c) != PIT_CELL) continue;
	
	pp.bc_found = 0;
	rmb(&pp, c, 1);
	
	/* walk the border, start from border set by rmb (min i) */
	pp.pp_found = 0;
	pp.pour_to_nodata = 0;
	borderwalk(pp.bc, pp.dir_out, &pp, &is_marked, &test_pour_point);

	if (!SAME_CELL(pp.ipp, c)) {
	    /* invert the path from the pour point to the pit point */
	    /* from pour_point to pit */
	    cell a = pp.ipp;
	    int da = IGD_CELL(pp.fdg, a);
	    cell b = cellmove(a, da);
	    int db = IGD_CELL(pp.fdg, b);
	    while (db) {
		debug(2,"inverting a = %i,%i b = %i,%i da = %i db = %i\n", a.i, a.j, b.i, b.j, da, db);
		IGD_CELL(pp.fdg, b) = INVDIR(da);
		a = b;
		da = db;
		b = cellmove(a, da);
		db = IGD_CELL(pp.fdg, b); 
	    }
	    IGD_CELL(pp.fdg, b) = INVDIR(da);
	}

	IGD_CELL(pp.fdg, pp.ipp) = pp.in2out;
	igdfloodfill(pp.mark, NULL, c, 0, 8);
	fixed_pits++;
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    gddestroy(pp.mark);
    return fixed_pits;
}


float cuag(grid *fdg, grid *uag, grid *load, cell c) 
{
    int dir;
    float ua = load ? (GD_DATACELL(load, c) ? GD_CELL(load, c) : 0) : 1;
    FOREACHDIR(dir) {
	cell t = cellmove(c, dir);
	if (GD_CELLIN(fdg, t) AND IGD_DATACELL(fdg, t) AND IGD_CELL(fdg, t) == INVDIR(dir))
	    ua += cuag(fdg, uag, load, t);
    }
    RGD_CELL(uag, c) = ua;
    return ua;
}


grid *fdg2uag(grid *fdg, grid *load) 
{
    static char *fct = "fdg2uag";
    grid *uag;
    cell c;
    PROGRESS_BAR_VARS(fdg->M/40);
    if (load)
	ASSERT(gdoverlayable(fdg, load), "fdg and load are not overlayable\n");
    ASSERT(fdg->datatype == INTEGER_GRID, "fdg has to be an integer grid\n");
    if (!(uag = gdnew2(REAL_GRID, fdg))) return NULL;
    FOREACH(c, fdg) {
	PROGRESS_CHECK(c.i);
	SIGCANCEL;
	/* start at outlet cells */
	if (IGD_DATACELL(fdg, c)) {
	    cell down = cellmove(c, IGD_CELL(fdg, c));
	    if (GD_CELLOUT(fdg, down) OR IGD_NODATACELL(fdg, down))
		RGD_CELL(uag, c) = cuag(fdg, uag, load, c);
	}
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    return uag;
 fail:
    return NULL;
}


int rroute(grid *water, grid *dem, grid *fdg, grid *flow, cell c, grid *k, grid *d, int f, double r)
{
    int dir;
    cell up, down;
    debug(1, "rroute at %i,%i\n", c.i, c.j);
    FOREACHDIR(dir) {
	up = cellmove(c, dir);
	if (GD_CELLIN(fdg, up) AND (IGD_CELL(fdg, up) == INVDIR(dir))) {
	    rroute(water, dem, fdg, flow, up, k, d, f, r);
	}
    }
    debug(1, "rroute at %i,%i\n", c.i, c.j);
    if (f == ROUTE_TO_ALL_DOWNSTREAM_CELLS) { 
	/* this is broken since there may be contributing upslope
       cells which have not been calculated at this point */
	double h = GD_CELL(dem, c);
	double q;
	double k2 = 1;
	int i;
	for (i = 0; i < 2; i++) {
	    q = 0;
	    FOREACHDIR(dir) {
		down = cellmove(c, dir);
		if (GD_CELLIN(dem, down) AND GD_DATACELL(dem, down) AND (GD_CELL(dem, down) <= h)) {
		    double slope, qd;
		    slope =  r * (GD_CELL(dem, c) - GD_CELL(dem, down)) / (dem->unitdist * DISTANCE_UNIT(dir));
		    qd = k2 * RGD_CELL(k, c) * (slope + RGD_CELL(d, c)) * RGD_CELL(water, c);
		    if (i) RGD_CELL(water, down) += qd;
		    q += qd;
		}
	    }
	    if (q > RGD_CELL(water, c)) {
		k2 = RGD_CELL(water, c) / q;
		q = RGD_CELL(water, c);
	    }
	}
	debug(1, "  s is %f, q is %f\n", GDGET(water, c), q);
	RGD_CELL(flow, c) = q;
	RGD_CELL(water, c) -= q;
    } else {
	double slope, q;
	dir = IGD_CELL(fdg, c);
	down = cellmove(c, dir);
	if (GD_CELLIN(dem, down) AND GD_DATACELL(dem, down)) {
	    slope = r * (GD_CELL(dem, c) - GD_CELL(dem, down)) / (dem->unitdist * DISTANCE_UNIT(dir));
	    q = RGD_CELL(k, c) * (slope + RGD_CELL(d, c)) * RGD_CELL(water, c);
	} else {
	    q = RGD_CELL(water, c);
	}
	if (q > RGD_CELL(water, c))
	    q = RGD_CELL(water, c);
	debug(1, "  s is %f, q is %f\n", GDGET(water, c), q);
	RGD_CELL(flow, c) = q;
	RGD_CELL(water, c) -= q;
	if (GD_CELLIN(water, down) AND RGD_DATACELL(water, down))
	    RGD_CELL(water, down) += q;
    }
    return 1;
}


int water_route(grid *water, grid *dem, grid *fdg, grid *flow, grid *k, grid *d, int f, double r)
{
    static char *fct = "route";
    cell c;
    int ret = 1;
    PROGRESS_BAR_VARS(dem->M/40);
    ASSERT(gdoverlayable(water, dem), "water and dem are not overlayable\n");
    ASSERT(gdoverlayable(water, fdg), "water and fdg are not overlayable\n");
    ASSERT(gdoverlayable(water, flow), "water and flow are not overlayable\n");
    ASSERT(gdoverlayable(water, k), "water and k are not overlayable\n");
    ASSERT(gdoverlayable(water, d), "water and d are not overlayable\n");
    ASSERT(water->datatype == REAL_GRID, "water has to be a floating point grid\n");
    ASSERT(fdg->datatype == INTEGER_GRID, "fdg has to be an integer grid\n");
    ASSERT(flow->datatype == REAL_GRID, "flow has to be a floating point grid\n");
    ASSERT(k->datatype == REAL_GRID, "k has to be a floating point grid\n");
    ASSERT(d->datatype == REAL_GRID, "d has to be a floating point grid\n");
    FOREACH(c, water) {
	PROGRESS_CHECK(c.i);
	SIGCANCEL;
	/* start at outlet cells */
	if (IGD_DATACELL(fdg, c)) {
	    cell down = cellmove(c, IGD_CELL(fdg, c));
	    if (GD_CELLOUT(fdg, down) OR IGD_NODATACELL(fdg, down))
		if (!(ret = rroute(water, dem, fdg, flow, c, k, d, f, r))) break;
	}
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    return ret;
 fail:
    return 0;
}


grid *fdg_distance_to_channel(grid *fdg, grid *streams) 
{
    char *fct = "fdg_distance_to_channel";
    grid *g;
    cell c;
    PROGRESS_BAR_VARS(fdg->M/40);
    if (!gdoverlayable(fdg, streams)) {
	fprintf(stderr,"%s: grids are not overlayable\n", fct);
	return NULL;
    }
    if (fdg->datatype != INTEGER_GRID OR streams->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: FDG and streams grids have to be integer grids\n",fct);
	return NULL;
    }
    if (!(g = gdnew2(REAL_GRID, fdg))) return NULL;
    FOREACH(c, fdg) {
	cell f = c;
	float d = 0;
	PROGRESS_CHECK(c.i);
	SIGCANCEL;
	while (GD_CELLIN(fdg, f) AND GD_DATACELL(fdg, f) AND 
	       RGD_CELL(g, f) <= 0 AND !IGD_CELL(streams, f)) {
	    d += DISTANCE_UNIT(IGD_CELL(fdg, f)) * fdg->unitdist;
	    f = FLOW(fdg, f);
	}
	RGD_CELL(g, c) = (GD_CELLIN(fdg, f) AND GD_DATACELL(fdg, f)) ? d + RGD_CELL(g, f) : d;
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    return g;
}


double fdd(grid *fdg, cell c)
{
    int count = 0, dir, invdir = 5;
    double d_from_here = 0;
    FOREACHDIR(dir) {
	cell t = cellmove(c, dir);
	if (GD_CELLIN(fdg, t) AND GD_DATACELL(fdg, t) AND IGD_CELL(fdg, t) == invdir)
	    count++;
    }
    invdir = 5;
    FOREACHDIR(dir) {
	cell t = cellmove(c, dir);
	if (GD_CELLIN(fdg, t) AND GD_DATACELL(fdg, t) AND IGD_CELL(fdg, t) == invdir)
	    d_from_here = 1/(double)count*fdd(fdg, t);
    }
    return d_from_here;
}


grid *fdg_distance_to_divide(grid *fdg) 
{
    char *fct = "fdg_distance_to_divide";
    grid *g;
    cell c;
    PROGRESS_BAR_VARS(fdg->M/40);
    if (fdg->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: FDG has to be an integer grid\n",fct);
	return NULL;
    }
    if (!(g = gdnew2(REAL_GRID, fdg))) return NULL;
    FOREACH(c, fdg) {
	PROGRESS_CHECK(c.i);
	SIGCANCEL;
	if (GD_DATACELL(fdg, c))
	    RGD_CELL(g, c) = fdd(fdg, c);
	else
	    RGD_SETNODATACELL(g, c);
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    return g;
}


typedef struct {
    grid *fdg;             /* flow dir grid */
    grid *dem;             /* dem grid */
    grid *flat;            /* all flat area */
    grid *ss;              /* slope sums */
    grid *uag;
    double dz;             /* flatness threshold */
} cuag2_struct;


int find_flats(cuag2_struct *pp) 
{
    cell c;
    FOREACH(c, pp->fdg) {
	if (GD_DATACELL(pp->fdg, c)) {
	    double zc = GD_CELL(pp->dem, c);
	    int dir, is_flat = 1;
	    FOREACHDIR(dir) {
		cell t = cellmove(c, dir);
		double zt = GD_CELL(pp->dem, t);
		/* this marks pits as flat but there shouldn't be any pits */
		if (GD_CELLIN(pp->fdg, t) AND GD_DATACELL(pp->fdg, t) AND zt < zc - pp->dz) {
		    is_flat = 0;
		    break;
		}
	    }
	    IGD_CELL(pp->flat, c) = is_flat;
	} else {
	    IGD_SETNODATACELL(pp->flat, c);
	}
    }
    return 1;
}


int slope_sums(cuag2_struct *pp)
{
    cell c;
    FOREACH(c, pp->fdg) {
	if (GD_DATACELL(pp->fdg, c)) {
	    double zc = GD_CELL(pp->dem, c);
	    double slope_sum = 0;
	    int dir;
	    FOREACHDIR(dir) {
		cell t = cellmove(c, dir);
		double zt;
		if (GD_CELLIN(pp->fdg, t) AND 
		    IGD_DATACELL(pp->fdg, t) AND
		    (zt = GD_CELL(pp->dem, t)) < zc - pp->dz)
		    slope_sum += (zc - zt)/DISTANCE_UNIT(dir);
	    }
	    RGD_CELL(pp->ss, c) = slope_sum;
	} else {
	    RGD_SETNODATACELL(pp->ss, c);
	}
    }
    return 1;
}


int cuag2(cuag2_struct *pp, cell c, int recursion) 
{
    int dir, invdir = 5;
    double ua = 1; 
    /* upslope area is 1 + that part of higher lying cells that comes
       to this cell

       if this cell is on flat area (i.e. the lowest 8-neighbor has
       the same z as this) then use the fdg to find the upslope cells
       and portion is 1 if the upslope cell is flat otherwise use the
       portion coeff */
    double zc = GD_CELL(pp->dem, c);
    FOREACHDIR(dir) {
	cell t = cellmove(c, dir);
	if (GD_CELLIN(pp->fdg, t) AND IGD_DATACELL(pp->fdg, t)) {
	    double zt = GD_CELL(pp->dem, t);
	    if (fabs(zc - zt) <= pp->dz AND IGD_CELL(pp->flat, t) AND IGD_CELL(pp->fdg, t) == invdir) {
		if (RGD_CELL(pp->uag, t) > 0 OR !recursion) {
		    ua += RGD_CELL(pp->uag, t);
		} else {
		    ua += RGD_CELL(pp->uag, t) = cuag2(pp, t, 1);
		}
	    } else if (zt > zc + pp->dz) {
		if (RGD_CELL(pp->uag, t) > 0 OR !recursion)
		    ua += ((zt - zc)/DISTANCE_UNIT(dir))/RGD_CELL(pp->ss, t)*RGD_CELL(pp->uag, t);
		else {
		    RGD_CELL(pp->uag, t) = cuag2(pp, t, 1);
		    ua += ((zt - zc)/DISTANCE_UNIT(dir))/RGD_CELL(pp->ss, t)*RGD_CELL(pp->uag, t);
		}
	    }
	}
	invdir++;
	if (invdir > 8) invdir = 1;
    }
    debug(5, "cuag2 at %i,%i is %f\n", c.i, c.j, ua);
    return ua;
}


/* fdg has fd's to flat areas */
grid *dem2uag(grid *dem, grid *fdg, int recursion)
{
    /*static char *fct = "dem2uag1";*/
    cell c;
    cuag2_struct pp;
    PROGRESS_BAR_VARS(fdg->M/40);
    pp.dem = dem;
    pp.fdg = fdg;
    if (!(pp.flat = gdnew2(INTEGER_GRID, dem))) return NULL;
    if (!(pp.ss = gdnew2(REAL_GRID, dem))) {
	gddestroy(pp.flat);
	return NULL;
    }
    if (!(pp.uag = gdnew2(REAL_GRID, dem))) {
	gddestroy(pp.flat);
	gddestroy(pp.ss);
	return NULL;
    }
    pp.dz = flatness_threshold(dem);
    find_flats(&pp);
    slope_sums(&pp);
    FOREACH(c, dem) {
	int dir, cont = 1;
	PROGRESS_CHECK(c.i);
	SIGCANCEL;
	if (GD_NODATACELL(fdg, c) OR RGD_CELL(pp.uag, c) > 0) continue;
	/* start only from cells which are next to nodata */
	FOREACHDIR(dir) {
	    cell t = cellmove(c, dir);
	    if (GD_CELLOUT(fdg, t) OR GD_NODATACELL(fdg, t)) cont = 0;
	}
	if (cont) continue;
	debug(1, "CALL cuag2 at %i,%i\n", c.i, c.j);
	RGD_CELL(pp.uag, c) = cuag2(&pp, c, recursion);
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    gddestroy(pp.flat);
    gddestroy(pp.ss);
    return pp.uag;
}


/* fdg has fd's to flat areas */
/* no recursion now, local estimate 
   downslope area = 
   n_lower*sum from all 8-neighbors lower(min(slope/slope_sum,1/n_lower))
*/
grid *dem2dag(grid *dem, grid *fdg)
{
    /*static char *fct = "dem2uag1";*/
    cell c;
    cuag2_struct pp;
    PROGRESS_BAR_VARS(fdg->M/40);
    pp.dem = dem;
    pp.fdg = fdg;
    if (!(pp.flat = gdnew2(INTEGER_GRID, dem))) return NULL;
    if (!(pp.ss = gdnew2(REAL_GRID, dem))) {
	gddestroy(pp.flat);
	return NULL;
    }
    if (!(pp.uag = gdnew2(REAL_GRID, dem))) {
	gddestroy(pp.flat);
	gddestroy(pp.ss);
	return NULL;
    }
    pp.dz = flatness_threshold(dem);
    find_flats(&pp);
    slope_sums(&pp);
    FOREACH(c, dem) {
	int dir, n = 0;
	double zc, sum = 0;
	PROGRESS_CHECK(c.i);
	SIGCANCEL;
	if (GD_NODATACELL(dem, c)) continue;
	zc = GD_CELL(dem, c);
	FOREACHDIR(dir) {
	    cell t = cellmove(c, dir);
	    if (GD_CELLIN(dem, t) AND GD_DATACELL(dem, t) AND 
		GD_CELL(dem, t) < zc - pp.dz) {
		n++;
	    }
	}
	FOREACHDIR(dir) {
	    cell t = cellmove(c, dir);
	    double zt;
	    if (GD_CELLIN(dem, t) AND GD_DATACELL(dem, t) AND 
		(zt = GD_CELL(dem, t)) < zc - pp.dz) {
		sum += min((zc - zt)/DISTANCE_UNIT(dir),1.0/(double)n);
	    }
	}
	RGD_CELL(pp.uag, c) = n * sum;
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    gddestroy(pp.flat);
    gddestroy(pp.ss);
    return pp.uag;
}


int mb(pour_point_struct *pp, grid *streams, cell c, int k) 
{
    cell c0 = c;
    int d = IGD_CELL(pp->fdg, c); /* direction at current cell */
    int n = 1;                /* nr. of possible directions to go */
    cell c2 = c;              /* moving cursor */
    int d2 = d; 
    debug(1,"mb %i,%i k=%i\n", c.i, c.j, k);
    SIGCANCEL;
    /* seek to the end of this stream */
    while (n == 1) {
	int dt = d2; /* test direction */
	dt = NEXTDIR(dt);
	debug(3,"on stream %i,%i\n", c2.i, c2.j);
	n = 0;
	c = c2;
	d = d2;
	while (dt != d) { /* test all directions */
	    cell t = c; /* test cell */
	    int di = INVDIR(dt); /* upstream cell has this fd */
	    t = cellmove(t, dt);
	    debug(4,"testing d=%i dt=%i %i,%i c=%i\n", d, dt, t.i, t.j, n);
	    if (GD_CELLIN(pp->fdg, t) AND 
		IGD_DATACELL(streams, t) AND IGD_CELL(streams, t) AND IGD_CELL(pp->fdg, t) == di) {
		n++;
		c2 = t;
		d2 = di;
	    }
	    dt++;
	    if (dt > 8) {
		dt = 1;
		if (d == 0) break;
	    }
	}
    }
    debug(2,"eor %i,%i c=%i d=%i\n", c.i, c.j, n, d);
    if (n > 1) { /* junction, recurse to all upstream streams */
	int dt = NEXTDIR(d);
	while (dt != d) { /* test all directions */
	    cell t = c; /* test cell */
	    int di = INVDIR(dt); /* upstream cell has this fd */
	    t = cellmove(t, dt);
	    if (GD_CELLIN(pp->fdg, t) AND 
		IGD_DATACELL(streams, t) AND IGD_CELL(streams, t) AND 
		IGD_CELL(pp->fdg, t) == di) {
		if (!(k = mb(pp, streams, t, k))) 
		    return 0;
	    }
	    dt++;
	    if (dt > 8) {
		dt = 1;
		if (d == 0) break;
	    }
	}
    }
    /* now mark this subcatchment and we are done */
    rmb(pp, c0, k);
    k++;
    return k;
 cancel:
    return 0;
}


grid *streams_subcatchments(grid *streams, grid *fdg, int i, int j) 
{
    static char *fct = "streams_subcatchments";
    cell c = {i, j};
    int k = 1;
    pour_point_struct pp;
    if (streams->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: streams grid has to be an integer grid\n", fct);
	return NULL;
    }
    if (IGD_CELL(streams, c) == 0) {
	fprintf(stderr,"%s: %i,%i is not a stream cell\n", fct, c.i, c.j);
	return NULL;
    }
    if (GD_CELLOUT(streams, c)) {
	fprintf(stderr,"%s: the cell %i,%i is not within the grid\n", fct, c.i, c.j);
	return NULL;
    }
    if (!init_pour_point_struct(&pp, fdg, NULL, gdnew2(INTEGER_GRID, fdg))) return NULL;
    if (!(k = mb(&pp, streams, c, k))) {
	gddestroy(pp.mark);
	return NULL;
    }
    return pp.mark;
}


int upstream_cells_of(grid *streams, grid *fdg, cell c, cell u[]) 
{
    int n = 0;
    int dir;
    FOREACHDIR(dir) {
	cell t = cellmove(c, dir);
	if (GD_CELLIN(streams, t) AND IGD_DATACELL(streams, t) AND 
	    IGD_CELL(streams, t) AND IGD_CELL(fdg, t) == INVDIR(dir)) {
	    u[n] = t;
	    n++;
	}
    }
    return n;
}


int number_tree(grid *streams, grid *fdg, cell c, int k) 
{
    while (1) {
	/* move along the stream from b onwards */    
	int n;
	cell u[8];  /* upstream cells */
	IGD_CELL(streams, c) = k;
	n = upstream_cells_of(streams, fdg, c, u);

	if (n == 0) break; /* b is the end cell of the stream */

	if (n == 1) { /* b is a stream cell */
	    c = u[0];

	} else { /* a cell where two or more streams join */
	    int i;
	    for (i = 0; i < n; i++)
		k = number_tree(streams, fdg, u[i], k+1);
	    break;
	}

    }
    return k;
}


int streams_number(grid *streams, grid *fdg, int i, int j, int sid0) 
{
    static char *fct = "streams_number";
    cell c = {i, j};
    if (!gdoverlayable(streams, fdg)) {
	fprintf(stderr,"%s: grids are not overlayable\n", fct);
	return 0;
    }
    if (streams->datatype != INTEGER_GRID OR fdg->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: streams and FDG grids are not integer grids\n", fct);
	return 0;
    }
    if (GD_CELLOUT(streams, c)) {
	fprintf(stderr,"%s: the cell %i,%i is not within the grid\n", fct, c.i, c.j);
	return 0;
    }
    number_tree(streams, fdg, c, sid0);
    return 1;
}


typedef struct {
    grid *lakes;
    grid *streams;
    grid *fdg;
    grid *uag;
    grid *mark;
    int lid;       /* lake, whose next-to-shore cells we are boating */
    int sid;       /* stored stream id */
    int nsid;      /* next available stream id */
    double min_l;  /* min_l in pruning */
    int outlet_found;
    double max_ua; /* maximum upslope area */
    cell outlet;   /* the cell with max_ua */
    int bc_found;  /* as bc in pour_point_struct */
    cell bc;
    int dir_out;
} lakedata;


int rprune(lakedata *ld, cell a, double l0);


/* borderwalk do-your-thing functions for lakedata struct: */

int prunelake(void *fctparam1, cell c) 
{
    lakedata *ld = (lakedata *)fctparam1;
    int dir;

    if (IGD_CELL(ld->mark, c)) return 0;
    IGD_CELL(ld->mark, c) = 1;
    
    /* neighboring non-lake stream cells */
    FOREACHDIR(dir) {

	cell t = cellmove(c, dir);

	if (GD_CELLOUT(ld->lakes, t) OR (IGD_CELL(ld->lakes, t) == ld->lid)) continue;

	if (IGD_DATACELL(ld->streams, t) AND IGD_CELL(ld->streams, t)) {

	    /* does the stream cell flow into this cell? */
	    
	    if (SAME_CELL(c, FLOW(ld->fdg, t))) {
		
		/* the test cell flows into the lake, prune this stream */
		int ret = rprune(ld, t, DISTANCE_UNIT(IGD_CELL(ld->fdg, t)) * ld->fdg->unitdist);
		if (ret == -1) return -1;

	    }
	    
	}
    }
    return 1;
}


int lakearea(void *fctparam1, cell c)
{
    return GD_CELLIN(((lakedata *)fctparam1)->lakes, c) AND 
	IGD_DATACELL(((lakedata *)fctparam1)->lakes, c) AND 
	IGD_CELL(((lakedata *)fctparam1)->lakes, c) == ((lakedata *)fctparam1)->lid;
}


/* 
   renumber the stream upstream from cell c with id id 
   stop at a lake if it is given 
*/
int renumber_stream(grid *streams, grid *fdg, grid *lakes, cell c, int id) 
{
    while (1) {
	/* move along the arc from c onwards */    
	int k;
	cell u[8]; /* upstream cells */
	IGD_CELL(streams, c) = id;
	k = upstream_cells_of(streams, fdg, c, u);
	if (k == 0) break; /* c is an end cell of a stream */
	if (k == 1) {      /* c is a stream cell */
	    if (lakes AND IGD_DATACELL(lakes, u[0]) AND IGD_CELL(lakes, u[0])) 
		break;
	    c = u[0];
	} else             /* a cell where two or more streams join */
	    break;
    }
    return 1;
}


/* returns true if the stream was 0-level stream and it was cut away */
int rprune(lakedata *ld, cell a, double l0) 
{
    static char *fct = "rprune";
    cell b = a;
    double l = l0; /* distance from where we were to a */
    debug(2,"rprune called at %i,%i\n", a.i, a.j);
    while (1) {

	/* move along the stream from a onwards */
	int n;

	cell u[8];  /* upstream cells */

	if (IGD_CELL(ld->mark, b)) {
	    fprintf(stderr,"%s: there is a loop in the system at %i,%i\n", fct, b.i, b.j);
	    return -1;
	}

	IGD_CELL(ld->mark, b) = 1;

	n = upstream_cells_of(ld->streams, ld->fdg, b, u);

	if (n == 0) { /* b is the end cell of the stream */

	    if (l < ld->min_l) { /* this stream is too short, prune it */
		cell x = b;
		while (1) {
		    IGD_CELL(ld->streams, x) = 0;
		    if (SAME_CELL(x, a)) break;
		    x = FLOW(ld->fdg, x);
		}
		return 1;
	    }
	    break;

	} else if (n == 1) { /* b is a regular stream cell */

	    if (ld->lakes AND IGD_DATACELL(ld->lakes, u[0]) AND IGD_CELL(ld->lakes, u[0])) { 
		/* with a lake cell just upstream */	
		/* backup lid, there may be lakes behind the lakes */
		cell ret;
		int lid = ld->lid;
		ld->lid = IGD_CELL(ld->lakes, u[0]);
		debug(2,"rprune: lake %i upstream of %i,%i\n", ld->lid, b.i, b.j);
		ret = borderwalk(u[0], IGD_CELL(ld->fdg, u[0]), ld, &lakearea, &prunelake);
		if (ret.i == -1) return -1;
		ld->lid = lid;
		break;
	    } else {
		l += DISTANCE_UNIT(IGD_CELL(ld->fdg, u[0])) * ld->fdg->unitdist;
		b = u[0];
	    }

	} else { /* b is a cell where two or more streams join */

	    int i;
	    for (i = 0; i < n; i++) {
		if (ld->lakes AND IGD_DATACELL(ld->lakes, u[i]) AND IGD_CELL(ld->lakes, u[i])) {
		    /* backup lid, there may be lakes behind the lakes */
		    cell ret;
		    int lid = ld->lid;
		    ld->lid = IGD_CELL(ld->lakes, u[i]);
		    debug(2,"rprune: lake %i upstream of %i,%i\n", ld->lid, b.i, b.j);
		    ret = borderwalk(u[i], IGD_CELL(ld->fdg, u[0]), ld, &lakearea, &prunelake);
		    if (ret.i == -1) return -1;
		    ld->lid = lid;
		} else {
		    double l0 = DISTANCE_UNIT(IGD_CELL(ld->fdg, u[i]))*ld->fdg->unitdist;
		    /* join the remaining upstream stream if there is now no junction */
		    int ret = rprune(ld, u[i], l0);
		    if (ret == -1) return -1;
		    if (ret == 1 AND n == 2) {
			int j = i == 0 ? 1 : 0;
			renumber_stream(ld->streams, ld->fdg, NULL, u[j], IGD_CELL(ld->streams, b));
		    }
		}
	    }
	    break;

	}

    }
    return 0;
}


int streams_prune(grid *streams, grid *fdg, grid *lakes, int i, int j, double min_l) 
{
    static char *fct = "streams_prune";
    cell c = {i, j};
    lakedata ld;
    int ret;
    ld.lakes = lakes;
    ld.streams = streams;
    ld.fdg = fdg;
    ld.min_l = min_l;
    if (!gdoverlayable(streams, fdg) OR (lakes AND !gdoverlayable(streams, lakes))) {
	fprintf(stderr,"%s: grids are not overlayable\n", fct);
	return 0;
    }
    if (fdg->datatype != INTEGER_GRID OR 
	streams->datatype != INTEGER_GRID OR
	(lakes AND lakes->datatype != INTEGER_GRID)) {
	fprintf(stderr,"%s: streams and FDG (and possibly lakes) grids are not all integer grids\n", fct);
	return 0;
    }
    if (GD_CELLOUT(streams, c)) {
	fprintf(stderr,"%s: the cell %i,%i is not within the grid\n", fct, c.i, c.j);
	return 0;
    }
    if (!(ld.mark = gdnew2(INTEGER_GRID, fdg))) return 0;
    ret = rprune(&ld, c, 0);
    gddestroy(ld.mark);
    return ret;
}


int testlake(void *fctparam1, cell c)
{
    lakedata *ld = (lakedata *)fctparam1;
    int dir;
    
    /* neighboring non-lake stream cells */
    FOREACHDIR(dir) {

	cell t = cellmove(c, dir);

	if (GD_CELLOUT(ld->lakes, t) OR (IGD_CELL(ld->lakes, t) == ld->lid)) continue;

	if (IGD_CELL(ld->streams, t) == ld->sid) {

	    /* it is a stream cell with same id, 
	       does the stream flow into this lake? */
	    
	    if (SAME_CELL(c, FLOW(ld->fdg, t))) {
		
		/* the test cell flows into the lake, renumber this stream */
		
		debug(1,"lake %i renumber stream %i to %i\n", ld->lid, ld->sid, ld->nsid);
		renumber_stream(ld->streams, ld->fdg, NULL, t, ld->nsid);
		ld->nsid++;
		
	    }
	}
    }
    return 0;
}


int streams_break(grid *streams, grid *fdg, grid *lakes, int nsid) 
{
    static char *fct = "streams_break";
    cell a;
    lakedata ld;

    ld.lakes = lakes;
    ld.streams = streams;
    ld.fdg = fdg;
    ld.nsid = nsid;

    if (!gdoverlayable(streams, fdg) OR !gdoverlayable(streams, lakes)) {
	fprintf(stderr,"%s: grids are not overlayable\n", fct);
	return 0;
    }
    if (fdg->datatype != INTEGER_GRID OR 
	streams->datatype != INTEGER_GRID OR 
	lakes->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: streams, FDG, and lakes grids are not integer grids\n", fct);
	return 0;
    }
    FOREACH(a, fdg) {
	debug(20,"break_streams1 at %i,%i\n", a.i, a.j);

	if (IGD_DATACELL(lakes, a) AND IGD_CELL(lakes, a) AND 
	    IGD_DATACELL(streams, a) AND IGD_CELL(streams, a)) {
		
	    /* cells of interest: a stream leaving a lake */

	    cell f = FLOW(fdg, a);                   /* the cell into which the stream flows to */
	    if (GD_CELLOUT(fdg, a) OR SAME_CELL(f, a)) 
	      continue;                              /* no flow */
	    ld.lid = IGD_CELL(lakes, a);             /* lake id */
	    ld.sid = IGD_CELL(streams, a);           /* stream id */
	    if (IGD_CELL(lakes, f) == ld.lid) continue; /* the stream is not an outflow */

	    borderwalk(a, IGD_CELL(fdg, a), &ld, &lakearea, &testlake);
	}
    }
    return 1;
}


int marklake(lakedata *ld, cell c)
{
    int size = 1, d;
    debug(25,"marklake at %i,%i\n", c.i, c.j);
    IGD_CELL(ld->mark, c) = 1;
    if (!ld->bc_found OR c.i < ld->bc.i) {
	debug(20,"marklake new bc %i,%i\n", c.i, c.j);
	ld->bc_found = 1;
	ld->bc = c;
	ld->dir_out = 1;  /* this finds the topmost cell (min i) so 1 is away from the area */
    }
    for (d = 1; d < 9; d++) {
	cell t = cellmove(c, d);
	if (GD_CELLIN(ld->lakes, t) AND IGD_DATACELL(ld->lakes, t) AND 
	    IGD_CELL(ld->lakes, t) == ld->lid AND !IGD_CELL(ld->mark, t))
	    size += marklake(ld, t);
    }
    return size;
}


int outlet(lakedata *ld, cell c) 
{
    cell f = FLOW(ld->fdg, c);          
    if (GD_CELLOUT(ld->fdg, f))
	return 1;
    return IGD_CELL(ld->lakes, f) != ld->lid;
}

int findoutlet(void *fctparam1, cell c)
{
    lakedata *ld = (lakedata *)fctparam1;
    double ua;
    if (!outlet(ld, c)) return 0;
    ua = GD_CELL(ld->uag, c);
    if (!ld->outlet_found OR ua > ld->max_ua) {
	ld->outlet_found = 1;
	ld->max_ua = ua;
	ld->outlet = c;
	return 1;
    }
    return 0;
}

void drain_lake_to_outlet(lakedata *ld, cell c) 
{
    int d, invdir = 5;
    debug(25,"drain_lake_to_outlet at %i,%i\n", c.i, c.j);
    IGD_CELL(ld->mark, c) = 1;
    for (d = 1; d < 9; d++) {
	cell t = cellmove(c, d);
	if (GD_CELLIN(ld->fdg, t) AND 
	    IGD_DATACELL(ld->fdg, t) AND 
	    IGD_CELL(ld->lakes, t) == ld->lid AND
	    !IGD_CELL(ld->mark, t)) {
	    IGD_CELL(ld->fdg, t) = invdir;
	    drain_lake_to_outlet(ld, t);
	}
	invdir++;
	if (invdir > 8) invdir = 1;
    }
}

int fdg_killoutlets(grid *fdg, grid *lakes, grid *uag) 
{
    static char *fct = "fdg_killoutlets";
    cell c;
    lakedata ld;

    ld.lakes = lakes;
    ld.fdg = fdg;
    ld.uag = uag;

    if (!gdoverlayable(lakes, fdg) OR !gdoverlayable(lakes, uag)) {
	fprintf(stderr,"%s: grids are not overlayable\n", fct);
	return 0;
    }
    if (fdg->datatype != INTEGER_GRID OR
	lakes->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: FDG and lakes grids are not integer grids\n", fct);
	return 0;
    }
    if (!(ld.mark = gdnew2(INTEGER_GRID, fdg))) return 0;
    FOREACH(c, fdg) {

	/* check all lakes once */
	if (IGD_DATACELL(lakes, c) AND IGD_CELL(lakes, c) AND !IGD_CELL(ld.mark, c)) {
	    
	    /* cells of interest: flow out of a lake */
	    /* find the one and only outlet: 
	       the one with largest upslope area */

	    ld.lid = IGD_CELL(lakes, c);
	    ld.bc_found = 0;
	    marklake(&ld, c); /* only to find a definitive border cell */
	    igdfloodfill(ld.mark, NULL, c, 0, 8);

	    ld.outlet_found = 0;
	    borderwalk(ld.bc, ld.dir_out, &ld, &lakearea, &findoutlet);
	    if (!ld.outlet_found) {
		fprintf(stderr,"%s: there is no outlet at lake %i\n", fct, ld.lid);
		gddestroy(ld.mark);
		return 0;
	    }

	    /* kill other outlets */
	    debug(22,"lake %i outlet at %i,%i\n", ld.lid, ld.outlet.i, ld.outlet.j);
	    drain_lake_to_outlet(&ld, ld.outlet); /* mark again */
	}

    }
    gddestroy(ld.mark);
    return 1;
}


int wsempty(ws *w)
{
    if (w) {
	if (w->outlet) free(w->outlet);
	w->outlet = NULL;
	if (w->down) free(w->down);
	w->down = NULL;
    }
    return 1;
}


cell wsdown(ws *w, cell outlet)
{
    int i;
    cell down = {-1, -1};
    for (i = 0; i < w->n; i++) {
	if (SAME_CELL(outlet, w->outlet[i])) return w->down[i];
    }
    return down;
}


int wsadd(ws *w, cell outlet, cell down) 
{
    cell test;
    debug(4,"wsadd outlet= %i,%i, down = %i,%i\n", outlet.i, outlet.j, down.i, down.j);
    if (!w) return 0;
    if (w->size == 0) {
	w->size = w->delta;
	if (!(w->outlet = (cell *)calloc(w->size,sizeof(cell)))) goto outofmem;
	if (!(w->down = (cell *)calloc(w->size,sizeof(cell)))) goto outofmem;
	w->n = 0;
    } else if (w->n >= w->size) {
	cell *tmp;
	w->size += w->delta;
	if (!(tmp = (cell *)realloc(w->outlet,w->size*sizeof(cell)))) goto outofmem;
	w->outlet = tmp;
	if (!(tmp = (cell *)realloc(w->down,w->size*sizeof(cell)))) goto outofmem;
	w->down = tmp;
    }
    /* testing for loops down -> -> outlet */
    if (!SAME_CELL(down, outlet)) {
	do {
	    test = wsdown(w, down);
	    if (SAME_CELL(test, outlet)) {
		fprintf(stderr,"a loop detected: catchment %i,%i already drains eventually to catchment %i,%i\n", 
			down.i, down.j, outlet.i, outlet.j);
		return 0;
	    }
	} while (test.i > 0);
    }
    w->outlet[w->n] = outlet;
    w->down[w->n] = down;
    w->n++;
    return 1;
 outofmem:
    fprintf(stderr,"Out of memory!\n");
    return 0;
}


typedef struct {
    pour_point_struct pp;
    /* fill these when calling tree or lake */
    ws *w;
    grid *subs;
    grid *streams;
    grid *lakes;
    int k;
    int headwaters; /* are headwaters marked as separate subs */
    /* these are needed only in internal routines */
    int lid;
    cell a;
} treedata;


int tree(treedata *td, cell a);
int lake(treedata *td, cell a);


int ws_subcatchments(ws *w, grid *subs, grid *streams, grid *fdg, grid *lakes, int i, int j, int headwaters) 
{
    static char *fct = "ws_subcatchments";
    int ret;
    cell c = {i, j};
    w->size = 0;
    w->delta = 50;	
    w->outlet = NULL;
    w->down = NULL;
    {
	treedata td = {{fdg, NULL, subs, 0, 0, 
			{-1, -1}, 0, 0, 1, 0, {-1, -1}, 0.0, 0.0, 0.0, 0, 0, 0, {-1, -1}, 0}, 
		       w, subs, NULL, lakes, 1, headwaters};
	if (!gdoverlayable(subs, streams) OR 
	    !gdoverlayable(subs, fdg) OR 
	    !gdoverlayable(subs, lakes)) {
	    fprintf(stderr,"%s: grids are not overlayable\n", fct);
	    return 0;
	}
	if (subs->datatype != INTEGER_GRID OR 
	    streams->datatype != INTEGER_GRID OR
	    fdg->datatype != INTEGER_GRID OR
	    lakes->datatype != INTEGER_GRID) {
	    fprintf(stderr,"%s: grids are not integer grids\n",fct);
	    return 0;
	}
	if (GD_CELLOUT(subs, c)) {
	    fprintf(stderr,"%s: the cell %i,%i is not within the grid\n",fct,i,j);
	    return 0;
	}
	if (!(streams = gdcreatecopy(streams,0))) 
	    return 0;
	td.streams = streams;
	ret = tree(&td, c);
	if (ret) {
	    rmb(&(td.pp), c, td.k);
	    ret = wsadd(td.w, c, c);
	}
	gddestroy(streams);
    }
    return ret;
}


/* callback for the treedata struct */

int testlake2(void *fctparam1, cell c)
{
    treedata *td = (treedata *)fctparam1;
    int d;
    
    /* neighboring non-lake stream cells */
    for (d = 1; d < 9; d++) {

	cell t = cellmove(c, d);

	if (GD_CELLOUT(td->lakes, t) OR (IGD_CELL(td->lakes, t) == td->lid)) continue;

	debug(30,"looking for incoming streams at %i,%i sid=%i\n", t.i, t.j, IGD_CELL(td->streams, t));

	if (IGD_DATACELL(td->streams, t) AND IGD_CELL(td->streams, t) > 0) { 

	    /* it is an unvisited stream cell, 
	       does the stream flow into this lake? */
	    
	    if (SAME_CELL(FLOW(td->pp.fdg, t), c)) {

		/* the test cell flows into the lake */
		
		debug(1,"stream %i enters the lake %i from %i,%i to %i,%i\n", 
		      IGD_CELL(td->streams, t), td->lid, t.i, t.j, c.i, c.j);
		if (!tree(td, t)) return -1;
		rmb(&(td->pp), t, td->k);
		td->k++;
		if (!wsadd(td->w, t, td->a)) return -1;
	    }
	    
	}
    }
    return 1;
}


int lakearea2(void *fctparam1, cell c)
{
    treedata *td = (treedata *)fctparam1;
    debug(23,"lakearea2 called with %i,%i\n", c.i, c.j);	
    return IGD_CELL(td->lakes, c) == td->lid;
}


int tree(treedata *td, cell a) 
{
    /* arc a..b , k is a running number for the subcatchments
       this is a recursive function
       let a = b = root point when calling from outside
    */
    static char *fct = "tree";
    cell b = a;
    debug(1,"tree called at %i %i k=%i sid=%i\n", a.i, a.j, td->k, IGD_CELL(td->streams, a));
    while (1) {
	/* move along the arc from b onwards */    
	int c;
	cell u[8]; /* upstream cells of b */
    
	/* check for loops, visited stream cells are marked by
	   multiplying them by -1 this has no effect to the original
	   streams grid since we use a copy here */

	if (IGD_DATACELL(td->streams, b)) {
	    if (IGD_CELL(td->streams, b) < 0) {
		fprintf(stderr,"%s: at (%i,%i): there is a loop in the system\n", fct, b.i, b.j);
		return 0;
	    } else 
		IGD_CELL(td->streams, b) *= -1;
	}

	c = upstream_cells_of(td->streams, td->pp.fdg, b, u);

	debug(14,"tree at %i %i, %i upstream cells\n", b.i, b.j, c);

	if (c == 0) { /* b is an end cell of an arc */
	    if (td->headwaters) {
		rmb(&(td->pp), b, td->k);
		td->k++;
		if (!wsadd(td->w, b, FLOW(td->pp.fdg, b))) return 0;
	    }
	    break;
	} else if (c == 1) {      /* b is an arc cell */
	    if (IGD_DATACELL(td->lakes, u[0]) AND IGD_CELL(td->lakes, u[0])) { /* but on a lake */
		if (!lake(td, u[0])) return 0;
		rmb(&(td->pp), u[0], td->k);
		td->k++;
		if (!wsadd(td->w, u[0], a)) return 0;		
		break;
	    } else { /* regular arc point */
		b = u[0];
	    }
	} else {           /* b is a cell where two or more arcs join */
	    int l;
	    for (l = 0; l < c; l++) {
		if (IGD_DATACELL(td->lakes, u[l]) AND IGD_CELL(td->lakes, u[l])) { /* a lake cell */
		    if (!lake(td, u[l])) return 0;
		} else {
		    if (!tree(td, u[l])) return 0;
		}
		rmb(&(td->pp), u[l], td->k);
		td->k++;
		if (!wsadd(td->w, u[l], a)) return 0;
	    }
	    break;
	}
    }
    debug(1,"tree done at %i,%i\n", a.i, a.j);
    return 1;
}


int lake(treedata *td, cell a) 
{
    cell r;
    /* backup the old values: */
    int lid = td->lid; 
    cell a_copy = td->a;

    td->lid = IGD_CELL(td->lakes, a);  /* lake id */    
    td->a = a;

    debug(1,"lake %i at %i %i k=%i\n", td->lid, a.i, a.j, td->k);

    r = borderwalk(a, IGD_CELL(td->pp.fdg, a), td, &lakearea2, &testlake2);

    debug(1,"lake %i done ret=%i\n", td->lid, r.i);

    if (r.i < 0) return 0;

    /* return the old values */
    td->lid = lid; 
    td->a = a_copy;
    return 1;
}


