#ifndef CATCHMENT_H
#define CATCHMENT_H


grid *dem2aspect(grid *dem);

/* z_factor is the unit of z dived by the unit of x and y */

grid *dem2slope(grid *dem, double z_factor);

/* 
   fdg is a Flow Direction Grid, an integer grid with flow dirs in cells 
   -1 denotes flat area, 
   0 is a pit (local minimum), 
   1 is up (-1,0), 
   2 is up-right (-1,1), 
   3 is right (0,1), etc */

#define FLOW(fdg, c) (cellmove((c), IGD_CELL((fdg), (c))))

#define FLAT_AREA -1
#define PIT_CELL 0

#define D8 1
#define RHO8 2
#define MANY8 3
/* 
   all flowpaths end in a pit, a nodata cell, or outside the grid
   methods are 
   D8: deterministic eight-neighbors
   RHO8: stochastic eight-neighbors (Fairfield & Leymarie, WRR 27(5) 709-717)
   MANY8: codes all 8 neighbor cells which are lower
   default method is D8
*/
grid *dem2fdg(grid *dem, int method);

/* 
   dem to upslope cells grid, upslope cells are all 8-neighbors which are higher
 */
grid *dem2ucg(grid *dem);

/* a struct for marking and then walking the border of a catchment
   what do we search for? (Martz and Garbrecht, 1998): the lowest
   border cell with steepest slope out of the area */

typedef struct {
    grid *fdg;             /* flow dir grid */
    grid *dem;             /* dem grid */
    grid *mark;            /* current flat area or pit cell's catchment */
    double zf;             /* elevation of the flat area */
    double dz;             /* flatness threshold */
    cell bc;               /* a border cell after marking the area, start the walk from here */
    int bc_found;          /* bc is valid */
    int dir_out;           /* direction away from the area, in topological sense?, suitable for borderwalk */
    int test_inner;        /* when looking for the pour point, do we
			      look for lowest border cell (default) or lowest cell adjacent to the area? */
    int pp_found;          /* inner pour point found? */
    cell ipp;              /* inner pour point */
    double z_ipp;          /* elevation of the inner pour point */
    double slope_in;       /* slope from the inner pour point to its downslope point in the depression */
    double slope_out;      /* slope from the inner pour point to the the outer pour point */
    int dio;               /* distance from the ipp to opp */
    int in2out;            /* direction from inner_pour_point to outer_pour_point */
    int pour_to_nodata;    /* pour to nodata? */
    cell opp;              /* is valid only if pour_to_nodata is false */
    double z_opp;          /* elevation of the outer pour point */
} pour_point_struct;

int init_pour_point_struct(pour_point_struct *pp, grid *fdg, grid *dem, grid *mark);

/* 
   recursively mark a basin, using only fdg, return the size 
   DOES NOT check if c is within the grids
*/
int rmb(pour_point_struct *pp, cell c, int m);

/* 
   fix fdg by iteratively draining flat cells to non-higher cells with dge resolved
*/
int fdg_fixflats1(grid *fdg, grid *dem);

/* 
   drain to inner pour point
   make inner pour point a pit if its pour point is higher
   else drain it to the outer pour point
*/
int fdg_fixflats2(grid *fdg, grid *dem);

/* fill or cut pits or peaks (single cells) */

int dem_fillpits(grid *dem, double z_limit);

int dem_cutpeaks(grid *dem, double z_limit);

/* 
   return the depressions in the DEM 
   each depression is marked with a unique integer if inc_m
*/
grid *dem_depressions(grid *dem, grid *fdg, int inc_m);

/* 
   fills the depressions in the dem, returns the number of filled depressions
*/
int dem_filldepressions(grid *dem, grid *fdg);

int dem_breach(grid *dem, grid *fdg, int limit);

/*  
    inverts the path from the pit cell to the lowest pour point of the depression
*/
int fdg_fixpits(grid *fdg, grid *dem);

#define ROUTE_TO_ALL_DOWNSTREAM_CELLS 1

int water_route(grid *water, grid *dem, grid *fdg, grid *flow, grid *k, grid *d, int f, double r);

grid *fdg2uag(grid *fdg, grid *load);

grid *fdg_distance_to_channel(grid *fdg, grid *streams);
grid *fdg_distance_to_divide(grid *fdg);

grid *dem2uag(grid *dem, grid *fdg, int recursion);
grid *dem2dag(grid *dem, grid *fdg);

grid *streams_subcatchments(grid *streams, grid *fdg, int i, int j);

/* 
   number each stream section with a unique id 
*/
int streams_number(grid *streams, grid *fdg, int i, int j, int sid0);

/*                          */
/* lakes grid is used below */
/*                          */

/* lakes should have (by default) only one outlet */

/* this removes other outlets than the one with max ua by changing the fdg */
int fdg_killoutlets(grid *fdg, grid *lakes, grid *uag);

/*
  pruning removes streams shorter than min_l (give min_l in grid scale)
*/
int streams_prune(grid *streams, grid *fdg, grid *lakes, int i, int j, double min_l);

/*
  renumbers the upstream reach of a stream which flows through a lake
  without junctions

  break streams actually broken by a lake 
  nsid is the first available stream id
*/
int streams_break(grid *streams, grid *fdg, grid *lakes, int nsid);

/* catchment structure: */

typedef struct {
    cell *outlet;  /* outlet cell of this catchment */
    cell *down;    /* outlet cell of the catchment into which this drains */
    int n;
    int size;
    int delta;
} ws;

int wsempty(ws *w);

/* 
   algorithm to make a subcatchment grid from 
   a flow direction grid
   a numbered stream grid 
   a numbered lake grid
*/
int ws_subcatchments(ws *w, grid *subs, grid *streams, grid *fdg, grid *lakes, int i, int j, int headwaters);

#endif
