#include <endian.h>
#include <byteswap.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <limits.h>
#include <math.h>

#include "hash.h"
#include "grid.h"

#define WHITESPACE " \t\r\n"

int debug = 0;
int hashmarks = 1;
int sigint = 0;

grid *mask = NULL;


int gddebug(int dbg) 
{
    int old = debug;
    debug = dbg;
    return old;
}


int gdhashmarks(int h) 
{
    hashmarks = h;
    return hashmarks;
}


int gdsigint(int s) 
{
    sigint = s;
    return sigint;
}


char *etrim(char *s) 
{
    char *e = s+strlen(s)-1;
    if (e != s) {
	e--;
	while (e != s AND strchr(WHITESPACE, *e)) e--;
	e++;
	*e = '\0';
    }
    return s;
}


char *btrim(char *s) 
{
    while (*s AND strchr(WHITESPACE, *s)) s++;
    return s;
}


char *w_ext(char *basename, char *ext) 
{
    char *we = malloc(strlen(basename)+strlen(ext)+1);
    if (!we) {
	fprintf(stderr,"Out of memory!\n");
	return NULL;
    }
    strcpy(we,basename);
    strcat(we,ext);
    return we;
}


char *getln(FILE *stream) 
{
    char *l = NULL, *s, *x, e;
    size_t sn, ln = 0;
    while (1) {
	s = malloc(1024);
	if (!s) {
	    fprintf(stderr,"Out of memory!\n");
	    break;
	}
	x = fgets(s,1024,stream);
	if (!x) {
	    free(s);
	    break;
	}
	sn = strlen(s);
	e = *(s+sn-1);
	if (!l) {
	    l = strdup(s);
	    free(s);
	    if (!l) {
		fprintf(stderr,"Out of memory!\n");
		break;
	    }
	    ln = sn;
	} else {
	    ln += sn;
	    l = realloc(l,ln+1);
	    if (!l) {
		fprintf(stderr,"Out of memory!\n");
		free(s);
		break;
	    } else {
		strcat(l,s);
		free(s);
	    }
	}
	if ((e == '\r') OR (e == '\n')) {
	    break;
	}
    }
    return l;
}


cell *cellnew()
{
    static char *fct = "gellnew";
    cell *c = (cell *)malloc(sizeof(cell));
    if (!c) {
	fprintf(stderr,"%s: Out of memory!\n", fct);
	return NULL;
    }
    c->i = 0;
    c->j = 0;
    return c;
}


int celldestroy(cell *c)
{
    if (c) free(c);
    return 1;
}


cell cellmove(cell c, int dir) 
{
    switch (dir) {
    case 1:{ c.i--;         break; }
    case 2:{ c.i--; c.j++;  break; }
    case 3:{ c.j++;         break; }
    case 4:{ c.i++; c.j++;  break; }
    case 5:{ c.i++;         break; }
    case 6:{ c.i++; c.j--;  break; }
    case 7:{ c.j--;         break; }
    case 8:{ c.i--; c.j--;  break; }
    }
    return c;
}


int celldirto(cell a, cell b)
{
    if (b.i < a.i AND b.j == a.j)
	return 1;
    else if (b.i < a.i AND b.j > a.j)
	return 2;
    else if (b.i == a.i AND b.j > a.j)
	return 3;
    else if (b.i > a.i AND b.j > a.j)
	return 4;
    else if (b.i > a.i AND b.j == a.j)
	return 5;
    else if (b.i > a.i AND b.j < a.j)
	return 6;
    else if (b.i == a.i AND b.j < a.j)
	return 7;
    else 
	return 8;
}


int gdsetmask(grid *gd)
{
    if (mask) gddestroy(mask);
    mask = NULL;
    if (!gd) return 1;
    mask = gdcreatecopy(gd, INTEGER_GRID);
    if (!mask) return 0;
    return 1;
}


grid *gdgetmask()
{
    return mask;
}


grid *gdinit() 
{
    static char *fct = "gdinit";
    grid *gd = (grid *)malloc(sizeof(grid));
    if (!gd) {
	fprintf(stderr,"%s: Out of memory!\n", fct);
	return NULL;
    }
    gd->title = NULL;
    gd->datatype = 0;
    gd->filetype = NULL;
    gd->M = 0;
    gd->N = 0;
    gd->unitdist = 1;
    gd->xy.minX = 0;
    gd->xy.maxX = 0;
    gd->xy.minY = 0;
    gd->xy.maxY = 0;
    gd->minval = 0;
    gd->maxval = 0;
    gd->sdata = NULL;
    gd->fdata = NULL;
    gd->snodata = -9999;
    gd->fnodata = -9999.0;
    return gd;
}


int gdcreate(grid *gd) 
{
    static char *fct = "gdcreate";
    if (gd->datatype == REAL_GRID) {
	gd->fdata = (REAL *)calloc(gd->M*gd->N,sizeof(REAL));
	if (!gd->fdata) {
	    fprintf(stderr,"%s: Out of memory!\n", fct);
	    return 0;
	}
    } else if (gd->datatype == INTEGER_GRID) {
	gd->sdata = (INTEGER *)calloc(gd->M*gd->N,sizeof(INTEGER));
	if (!gd->sdata) {
	    fprintf(stderr,"%s: Out of memory!\n", fct);
	    return 0;
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gddestroy(grid *gd) 
{
    if (gd) {
	if (gd->title) free(gd->title);
	if (gd->filetype) free(gd->filetype);
	if (gd->sdata) free(gd->sdata);
	if (gd->fdata) free(gd->fdata);
	free(gd);
    }
    return 1;
}


grid *gdnew(int datatype, int M, int N) 
{
    grid *gd = gdinit();
    if (!gd) return NULL;
    gd->datatype = datatype;
    gd->M = M;
    gd->N = N;
    if (!gdcreate(gd)) {
	gddestroy(gd);
	return NULL;
    }
    gdsetbounds(gd,1,0,0);
    return gd;
}


grid *gdnew2(int datatype, grid *gd)
{
    grid *g = gdinit();
    if (!g) return NULL;
    g->datatype = datatype;
    g->M = gd->M;
    g->N = gd->N;
    g->snodata = gd->snodata;
    g->fnodata = gd->fnodata;
    if (!gdcreate(g)) {
	gddestroy(g);
	return NULL;
    }
    g->unitdist = gd->unitdist;
    g->xy = gd->xy;
    return g;
}


int gdcopy(grid *to, grid *from) 
{
    static char *fct = "gdcopy";
    cell c;
    if (to->M != from->M OR to->N != from->N) {
	fprintf(stderr,"%s: ERROR: grids are of different size\n", fct);
	return 0;
    }
    if (to->datatype < from->datatype) {
	fprintf(stderr,"%s: WARNING: copying to lower datatype\n", fct);
    }
    if (to->datatype == INTEGER_GRID AND from->datatype == INTEGER_GRID) {
	FOREACH(c, from)
	    GDSET2(to, c, IGD_CELL(from, c));
    } else {
	FOREACH(c, from)
	    GDSET(to, c, GD_CELL(from, c));
    }
    to->minval = from->minval;
    to->maxval = from->maxval;
    to->snodata = from->snodata;
    to->fnodata = from->fnodata;
    return 1;
}


grid *gdcreatecopy(grid *gd, int datatype) 
{
    grid *g;
    if (datatype == 0) datatype = gd->datatype;
    if (!(g = gdnew2(datatype, gd))) return NULL;
    gdcopy(g, gd);
    return g;
}


grid *gdopen(char *basename) 
{
    static char *fct = "gdopen";
    grid *gd = gdinit();
    FILE *f;
    char *fn, *ln, *key, *val;
    if (!gd) return NULL;
    if (!(fn = w_ext(basename, ".doc"))) {
	gddestroy(gd);
	return NULL;
    }
    if (!(f = fopen(fn,"r"))) {
	fprintf(stderr,"%s: %s: %s\n",fct,fn,strerror(errno)); 
	gddestroy(gd);
	free(fn);
	return NULL;
    }
    free(fn);
    while ((ln = getln(f))) {
	key = btrim(etrim(strtok(ln,":")));
	val = btrim(etrim(strtok(NULL,":")));
	if ((strcmp(key,"file title") == 0)) {
	    gd->title = strdup(val);
	    if (!(gd->title)) {
		fprintf(stderr,"%s: Out of memory!\n", fct);
		gddestroy(gd);
		fclose(f);
		return NULL;
	    }
	} else if ((strcmp(key,"data type") == 0)) {
	    if ((strcmp(val,REAL_STRING) == 0)) {
		gd->datatype = REAL_GRID;
	    } else if ((strcmp(val,INTEGER_STRING) == 0)) {
		gd->datatype = INTEGER_GRID;
	    } else {
		fprintf(stderr,"%s: %s: unsupported data type\n",fct,val);
		gddestroy(gd);
		fclose(f);
		return NULL;
	    }
	} else if ((strcmp(key,"file type") == 0)) {
	    gd->filetype = strdup(val);
	    if (!(gd->filetype)) {
		fprintf(stderr,"%s: Out of memory!\n", fct);
		gddestroy(gd);
		fclose(f);
		return NULL;
	    }
	} else if ((strcmp(key,"columns") == 0)) {
	    gd->N = atoi(val);
	} else if ((strcmp(key,"rows") == 0)) {
	    gd->M = atoi(val);
	} else if ((strcmp(key,"unit dist.") == 0)) {
	    sscanf(val, "%lf", &(gd->unitdist));
	} else if ((strcmp(key,"min. X") == 0)) {
	    sscanf(val, "%f", &(gd->xy.minX));
	} else if ((strcmp(key,"max. X") == 0)) {
	    sscanf(val, "%f", &(gd->xy.maxX));
	} else if ((strcmp(key,"min. Y") == 0)) {
	    sscanf(val, "%f", &(gd->xy.minY));
	} else if ((strcmp(key,"max. Y") == 0)) {
	    sscanf(val, "%f", &(gd->xy.maxY));
	} else if ((strcmp(key,"min. value") == 0)) {
	    sscanf(val, "%lf", &(gd->minval));
	} else if ((strcmp(key,"max. value") == 0)) {
	    sscanf(val, "%lf", &(gd->maxval));
	}
	free(ln);
    }
    fclose(f);
    /* unit. dist. in IDRISI? */
    gdsetbounds4(gd, gd->xy.minX, gd->xy.maxX, gd->xy.minY);
    if (!gdcreate(gd)) {
	gddestroy(gd);
	return NULL;
    }
    {
	int el_size = 0;
	switch (gd->datatype) {
	case INTEGER_GRID: 
	    el_size = sizeof(INTEGER);
	    break;
	case REAL_GRID: 
	    el_size = sizeof(REAL);
	    break;
	}
	if (!(gdread(gd, basename, ".img", el_size, LITTLE_ENDIAN))) {
	    gddestroy(gd);
	    return NULL;
	}
    }
    return gd;
}


int gdread(grid *gd, char *basename, char *ext, int el_size, int byteorder) 
{
    static char *fct = "gdread";
    FILE *f;
    size_t l = 0;
    REAL *f_tmp = NULL;
    INTEGER *s_tmp = NULL;
    char *b_tmp = NULL;
    char *fn;
    int tmp = 1;

    switch (el_size) {
    case sizeof(REAL):
	if (gd->datatype != REAL_GRID)
	    tmp = (int)f_tmp  = (REAL *)calloc(gd->M*gd->N,sizeof(REAL));
	break;
    case sizeof(INTEGER):
	if (gd->datatype != INTEGER_GRID)
	    tmp = (int)s_tmp  = (INTEGER *)calloc(gd->M*gd->N,sizeof(INTEGER));
	break;
    case 1:
	tmp = (int)b_tmp  = (char *)calloc(gd->M*gd->N,1);
	break;
    default:
	fprintf(stderr,"%s: %i: unsupported element size\n", fct, el_size); 
	return 0;
    }
    if (!tmp) {
	fprintf(stderr,"%s: Out of memory!\n", fct);
	return 0;
    }
    
    if (!(fn = w_ext(basename, ext))) return 0;
    if (!(f = fopen(fn,"r"))) {
	fprintf(stderr,"%s: %s: %s\n",fct,fn,strerror(errno)); 
	free(fn);
	return 0;
    }
    free(fn);
    
    if (gd->datatype == REAL_GRID) {
	int i;
	switch (el_size) {
	case sizeof(REAL):
	    l = fread(gd->fdata,sizeof(REAL),gd->M*gd->N,f);
	    break;
	case sizeof(INTEGER):
	    l = fread(s_tmp,sizeof(INTEGER),gd->M*gd->N,f);
	    for (i = 0; i < gd->M*gd->N; i++) 
		gd->fdata[i] = s_tmp[i];
	    break;
	case 1:
	    l = fread(b_tmp,1,gd->M*gd->N,f);
	    for (i = 0; i < gd->M*gd->N; i++) 
		gd->fdata[i] = b_tmp[i];
	    break;
	}
    } else if (gd->datatype == INTEGER_GRID) {
	int i;
	switch (el_size) {
	case sizeof(REAL):
	    l = fread(f_tmp,sizeof(REAL),gd->M*gd->N,f);
	    for (i = 0; i < gd->M*gd->N; i++) {
		long a = round(f_tmp[i]);
		if (a < INTEGER_MIN OR a > INTEGER_MAX) {
		    fprintf(stderr,"%s: value %f out of bounds\n",fct,f_tmp[i]);
		    l = i-1;
		    break;
		}
		gd->sdata[i] = a;
	    }
	    break;
	case sizeof(INTEGER):
	    l = fread(gd->sdata,sizeof(INTEGER),gd->M*gd->N,f);
	    if (byteorder != BYTE_ORDER) {
		int i;
		for (i = 0; i < gd->M*gd->N; i++) {
		    gd->sdata[i] = bswap_16(gd->sdata[i]);
		}
	    }
	    break;
	case 1:
	    l = fread(b_tmp,1,gd->M*gd->N,f);
	    for (i = 0; i < gd->M*gd->N; i++) 
		gd->sdata[i] = b_tmp[i];
	    break;
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n",fct); 
    }
    if (f_tmp) free(f_tmp);
    if (s_tmp) free(s_tmp);
    if (b_tmp) free(b_tmp);
    fclose(f);
    if (l != gd->M*gd->N) {
	fprintf(stderr,"%s: got only %i items out of %i\n",fct,l,gd->M*gd->N); 
	return 0;
    }
    return 1;
}

int gdsave(grid *gd, char *basename) 
{
    static char *fct = "gdsave";
    FILE *f;
    char *fn;
    if (!(fn = w_ext(basename, ".doc")))
	return 0;
    if (!(f = fopen(fn,"w"))) {
	fprintf(stderr,"%s: %s: %s\n",fct,fn,strerror(errno)); 
	free(fn);
	return 0;
    }
    free(fn);
    fprintf(f,"file title  : %s\n",gd->title);
    if (gd->datatype == REAL_GRID) {
	fprintf(f,"data type   : %s\n",REAL_STRING);
    } else if (gd->datatype == INTEGER_GRID) {
	fprintf(f,"data type   : %s\n",INTEGER_STRING);
    } else {
	fprintf(f,"data type   : ?\n");
    }
    fprintf(f,"file type   : binary\n");
    fprintf(f,"columns     : %i\n",gd->N);
    fprintf(f,"rows        : %i\n",gd->M);
    fprintf(f,"ref. system : plane\n");
    fprintf(f,"ref. units  : m\n");
    fprintf(f,"unit dist.  : %f\n",gd->unitdist);
    fprintf(f,"min. X      : %f\n",gd->xy.minX);
    fprintf(f,"max. X      : %f\n",gd->xy.maxX);
    fprintf(f,"min. Y      : %f\n",gd->xy.minY);
    fprintf(f,"max. Y      : %f\n",gd->xy.maxY);
    fprintf(f,"pos'n error : unknown\n");
    fprintf(f,"resolution  : unknown\n");
    if (gd->datatype == REAL_GRID) {
	fprintf(f,"min. value  : %f\n",gd->minval);
	fprintf(f,"max. value  : %f\n",gd->maxval);
    } else if (gd->datatype == INTEGER_GRID) {
	fprintf(f,"min. value  : %i\n",(int)gd->minval);
	fprintf(f,"max. value  : %i\n",(int)gd->maxval);
    } else {
	fprintf(f,"min. value  : \n");
	fprintf(f,"max. value  : \n");
    }
    fprintf(f,"value units : unspecified\n");
    fprintf(f,"value error : unknown\n");
    fprintf(f,"flag value  : none\n");
    fprintf(f,"flag def'n  : none\n");
    fprintf(f,"legend cats : 0\n");
    fclose(f);
    if (!(gdwrite(gd, basename, ".img")))
	return 0;
    return 1;
}

int gdwrite(grid *gd, char *basename, char *ext)
{
    static char *fct = "gdwrite";
    FILE *f;
    size_t l;
    char *fn;

    if (!(fn = w_ext(basename, ext))) return 0;
    if (!(f = fopen(fn,"w"))) {
	fprintf(stderr,"%s: %s: %s\n",fct,fn,strerror(errno)); 
	free(fn);
	return 0;
    }
    free(fn);

    if (gd->datatype == REAL_GRID) {
	l = fwrite(gd->fdata,sizeof(REAL),gd->M*gd->N,f);
    } else if (gd->datatype == INTEGER_GRID) {
	l = fwrite(gd->sdata,sizeof(INTEGER),gd->M*gd->N,f);
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	fclose(f);
	return 0;
    }

    fclose(f);
    if (l != gd->M*gd->N) {
	fprintf(stderr,"%s: wrote only %i items out of %i\n",fct,l,gd->M*gd->N); 
	return 0;
    }
    return 1;
}


int gdsetnodata(grid *gd, double nodata)
{
    static char *fct = "gdsetnodata";
    if (gd->datatype == REAL_GRID) {
	gd->fnodata = nodata;
    } else if (gd->datatype == INTEGER_GRID) {
	long a = round(nodata);
	if (a < INTEGER_MIN OR a > INTEGER_MAX) {
	    fprintf(stderr,"%s: nodata value %f out of bounds\n",fct,nodata);
	    return 0;
	}
	gd->snodata = a;
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


double gdgetnodata(grid *gd)
{
    static char *fct = "gdgetnodata";
    if (gd->datatype == REAL_GRID) {
	return gd->fnodata;
    } else if (gd->datatype == INTEGER_GRID) {
	return gd->snodata;
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
}


int gdoverlayable(grid *g1, grid *g2)
{
    return (g1->M == g2->M AND g1->N == g2->N AND 
	    fabs(g1->unitdist - g2->unitdist) < EPSILON AND
	    fabs(g1->xy.minX - g2->xy.minX) < EPSILON AND
	    fabs(g1->xy.minY - g2->xy.minY) < EPSILON);
}


int gdgetM(grid *gd) 
{
    return gd->M;
}


int gdgetN(grid *gd) 
{
    return gd->N;
}


int gddatatype(grid *gd) 
{
    return gd->datatype;
}


double gdunitdist(grid *gd)
{  
    return gd->unitdist;
}


double gdminX(grid *gd)
{
    return gd->xy.minX;
}


double gdmaxX(grid *gd)
{
    return gd->xy.maxX;
}


double gdminY(grid *gd)
{
    return gd->xy.minY;
}


double gdmaxY(grid *gd)
{
    return gd->xy.maxY;
}


int gdsetbounds(grid *gd, double unitdist, double minX, double minY)
{
    gd->unitdist = unitdist;
    gd->xy.minX = minX;
    gd->xy.minY = minY;
    gd->xy.maxX = minX+gd->N*unitdist;
    gd->xy.maxY = minY+gd->M*unitdist;
    return 1;
}


int gdsetbounds2(grid *gd, double unitdist, double minX, double maxY)
{
    gd->unitdist = unitdist;
    gd->xy.minX = minX;
    gd->xy.minY = maxY-gd->M*unitdist;
    gd->xy.maxX = minX+gd->N*unitdist;
    gd->xy.maxY = maxY;
    return 1;
}


int gdsetbounds3(grid *gd, double unitdist, double maxX, double minY)
{
    gd->unitdist = unitdist;
    gd->xy.minX = maxX-gd->N*unitdist;
    gd->xy.minY = minY;
    gd->xy.maxX = maxX;
    gd->xy.maxY = minY+gd->M*unitdist;
    return 1;
}


int gdsetbounds4(grid *gd, double minX, double maxX, double minY)
{
    gd->unitdist = (maxX-minX)/gd->N;
    gd->xy.minX = minX;
    gd->xy.minY = minY;
    gd->xy.maxX = maxX;
    gd->xy.maxY = minY+gd->M*gd->unitdist;
    return 1;
}


int gdsetbounds5(grid *gd, double minX, double maxX, double maxY)
{
    gd->unitdist = (maxX-minX)/gd->N;
    gd->xy.minX = minX;
    gd->xy.minY = maxY-gd->M*gd->unitdist;
    gd->xy.maxX = maxX;
    gd->xy.maxY = maxY;
    return 1;
}


int gdsetbounds6(grid *gd, double minX, double minY, double maxY)
{
    gd->unitdist = (maxY-minY)/gd->M;
    gd->xy.minX = minX;
    gd->xy.minY = minY;
    gd->xy.maxX = minX+gd->N*gd->unitdist;
    gd->xy.maxY = maxY;
    return 1;
}


int gdsetbounds7(grid *gd, double maxX, double minY, double maxY)
{
    gd->unitdist = (maxY-minY)/gd->M;
    gd->xy.minX = maxX-gd->N*gd->unitdist;
    gd->xy.minY = minY;
    gd->xy.maxX = maxX;
    gd->xy.maxY = maxY;
    return 1;
}


int gdcopybounds(grid *from,grid *to)
{
    to->unitdist = from->unitdist;
    to->xy = from->xy;
    return 1;
}


int gdx2j(grid *gd, double x)
{
    static char *fct = "gdx2j";
    if (x < gd->xy.minX) {
	fprintf(stderr,"%s: x (%f) out of bounds [%f,%f]\n",fct,x,gd->xy.minX,gd->xy.maxX);
	return 0;
    }
    if (x > gd->xy.maxX) {
	fprintf(stderr,"%s: x (%f) out of bounds [%f,%f]\n",fct,x,gd->xy.minX,gd->xy.maxX);
	return gd->N-1;
    }
    x -= gd->xy.minX;
    x /= gd->unitdist;
    return floor(x);
}


int gdy2i(grid *gd, double y)
{
    static char *fct = "gdy2i";
    if (y < gd->xy.minY) {
	fprintf(stderr,"%s: y (%f) out of bounds [%f,%f]\n",fct,y,gd->xy.minY,gd->xy.maxY);
	return gd->M-1;
    }
    if (y > gd->xy.maxY) {
	fprintf(stderr,"%s: y (%f) out of bounds [%f,%f]\n",fct,y,gd->xy.minY,gd->xy.maxY);
	return 0;
    }
    y = gd->xy.maxY - y;
    y /= gd->unitdist;
    return floor(y);
}


double gdj2x(grid *gd, int j)
{
    static char *fct = "gdj2x";
    double x = j;
    if (j < 0) {
	fprintf(stderr,"%s: index j (%i) out of bounds [%i,%i]\n",fct,j,0,gd->N-1);
	return gd->xy.minX;
    }
    if (j >= gd->N) {
	fprintf(stderr,"%s: index j (%i) out of bounds [%i,%i]\n",fct,j,0,gd->N-1);
	return gd->xy.maxX;
    }
    x *= gd->unitdist;
    x += gd->xy.minX+gd->unitdist/2;
    return x;
}


double gdi2y(grid *gd, int i)
{
    static char *fct = "gdi2y";
    double y = gd->M-i;
    if (i < 0) {
	fprintf(stderr,"%s: index i (%i) out of bounds [%i,%i]\n",fct,i,0,gd->M-1);
	return gd->xy.maxY;
    }
    if (i >= gd->M) {
	fprintf(stderr,"%s: index i (%i) out of bounds [%i,%i]\n",fct,i,0,gd->M-1);
	return gd->xy.minY;
    }
    y *= gd->unitdist;
    y += gd->xy.minY-gd->unitdist/2;
    return y;
}


double gdget(grid *gd, int i, int j)
{
    static char *fct = "gdget";
    if (i < 0 OR i >= gd->M OR j < 0 OR j >= gd->N) {
	fprintf(stderr,"%s: index out of bounds (i,j)=(%i,%i) M=%i, N=%i\n",fct,i,j,gd->M,gd->N);
	return 0;
    }
    if (gd->datatype == REAL_GRID) {
	return RGD_AT(gd,i,j);
    } else if (gd->datatype == INTEGER_GRID) {
	return IGD_AT(gd,i,j);
    }
    fprintf(stderr,"%s: grid datatype error\n", fct);
    return 0;
}


int gdget2(grid *gd, int i, int j)
{
    static char *fct = "gdget2";
    if (i < 0 OR i >= gd->M OR j < 0 OR j >= gd->N) {
	fprintf(stderr,"%s: index out of bounds (i,j)=(%i,%i) M=%i, N=%i\n",fct,i,j,gd->M,gd->N);
	return 0;
    }
    if (gd->datatype == REAL_GRID) {
	return round(RGD_AT(gd,i,j));
    } else if (gd->datatype == INTEGER_GRID) {
	return IGD_AT(gd,i,j);
    }
    fprintf(stderr,"%s: grid datatype error\n", fct);
    return 0;
}


int gdset(grid *gd, int i, int j, double x)
{
    static char *fct = "gdset";
    if (i < 0 OR i >= gd->M OR j < 0 OR j >= gd->N) {
	fprintf(stderr,"%s: index out of bounds (i,j)=(%i,%i) M=%i, N=%i\n",fct,i,j,gd->M,gd->N);
	return 0;
    }
    if (gd->datatype == REAL_GRID) {
	RGD_AT(gd,i,j) = x;
    } else if (gd->datatype == INTEGER_GRID) {
	long a = round(x);
	if (a < INTEGER_MIN OR a > INTEGER_MAX) {
	    fprintf(stderr,"%s: can't insert value %f to an integer grid\n",fct,x);
	    return 0;
	}
	IGD_AT(gd,i,j) = a;
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdset2(grid *gd, int i, int j, int x)
{
    static char *fct = "gdset2";
    if (i < 0 OR i >= gd->M OR j < 0 OR j >= gd->N) {
	fprintf(stderr,"%s: index out of bounds (i,j)=(%i,%i) M=%i, N=%i\n",fct,i,j,gd->M,gd->N);
	return 0;
    }
    if (gd->datatype == REAL_GRID) {
	RGD_AT(gd,i,j) = x;
    } else if (gd->datatype == INTEGER_GRID) {
	if (x < INTEGER_MIN OR x > INTEGER_MAX) {
	    fprintf(stderr,"%s: %i: integer out of bounds\n",fct,x);
	    return 0;
	}
	IGD_AT(gd,i,j) = x;
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdsetminmax(grid *gd)
{
    double min, max;
    cell c;
    int f = 0;
    min = max = gdgetnodata(gd);
    FOREACH(c, gd) {
	if (GD_DATACELL(gd,c)) {
	    double a = gdget(gd,c.i,c.j);
	    if (!f) {
		min = max = a;
		f = 1;
	    }
	    if (a < min) min = a;
	    if (a > max) max = a;
	}
    }
    gd->minval = min;
    gd->maxval = max;
    return 1;
}


double gdgetminval(grid *gd)
{
    return gd->minval;
}


double gdgetmaxval(grid *gd)
{
    return gd->maxval;
}


cell gdgetmin(grid *gd) 
{
    double min = 0;
    cell c, x;
    int f = 0;
    FOREACH(c, gd) {
	if (GD_DATACELL(gd,c)) {
	    double a = gdget(gd,c.i,c.j);
	    if (!f OR a < min) {
		min = a;
		x = c;
		f = 1;
	    }
	}
    }
    return x;
}


cell gdgetmax(grid *gd)
{
    double max = 0;
    cell c, x;
    int f = 0;
    FOREACH(c, gd) {
	if (GD_DATACELL(gd,c)) {
	    double a = gdget(gd,c.i,c.j);
	    if (!f OR a > max) {
		max = a;
		x = c;
		f = 1;
	    }
	}
    }
    return x;
}


int gdsetall(grid *gd, double x)
{
    static char *fct = "gdsetall";
    cell c;
    if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd)
	    RGD_CELL(gd, c) = x;
    } else if (gd->datatype == INTEGER_GRID) {
	long a = round(x);
	if (a < INTEGER_MIN OR a > INTEGER_MAX) {
	    fprintf(stderr,"%s: %li: integer out f bounds\n",fct,a);
	    return 0;
	}
	FOREACH(c, gd)
	    IGD_CELL(gd, c) = a;
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdstealdata(grid *to, grid *from)
{
    if (to->fdata) free(to->fdata);
    to->fdata = NULL;
    if (to->sdata) free(to->sdata);
    to->sdata = NULL;
    to->fdata = from->fdata;
    from->fdata = NULL;
    to->sdata = from->sdata;
    from->sdata = NULL;
    to->datatype = from->datatype;
    return 1;
}


int gddata(grid *gd)
{
    static char *fct = "gddata";
    cell c;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd) {
	    if (IGD_DATACELL(gd, c)) 
		IGD_CELL(gd, c) = 1;
	    else
		IGD_CELL(gd, c) = 0;
	}
    } else if (gd->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd);
	if (!g) return 0;
	FOREACH(c, gd) {
	    if (RGD_DATACELL(gd, c)) 
		IGD_CELL(g, c) = 1;
	    else
		IGD_CELL(g, c) = 0;
	}
	gdstealdata(gd, g);
	gddestroy(g);
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdnot(grid *gd)
{
    static char *fct = "gdnot";
    cell c;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c)) 
		IGD_CELL(gd, c) = !IGD_CELL(gd, c);
    } else {
	fprintf(stderr,"%s: not an integer grid\n", fct);
	return 0;
    }
    return 1;
}


int gdandgd(grid *gd1, grid *gd2)
{
    static char *fct = "gdandgd";
    if (gd1->M != gd2->M OR gd1->N != gd2->N) {
	fprintf(stderr,"%s: the grids have different sizes\n", fct);
	return 0;
    }
    if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	cell c;
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c)) {
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) AND IGD_CELL(gd2, c);
	    } else 
		IGD_SETNODATACELL(gd1, c);
	}
	return 1;
    }
    fprintf(stderr,"%s: the grids are not both integer grids\n", fct);
    return 0;
}


int gdorgd(grid *gd1, grid *gd2) 
{  
    static char *fct = "gdorgd";
    if (gd1->M != gd2->M OR gd1->N != gd2->N) {
	fprintf(stderr,"%s: the grids have different sizes\n", fct);
	return 0;
    }
    if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	cell c;
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c)) {
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) OR IGD_CELL(gd2, c);
	    } else 
		IGD_SETNODATACELL(gd1, c);
	}
	return 1;
    }
    fprintf(stderr,"%s: the grids are not both integer grids\n", fct);
    return 0;
}


int gdaddsv(grid *gd, double x)
{
    static char *fct = "gdaddsv";
    cell c;
    if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd)
	    if (RGD_DATACELL(gd, c))
		RGD_CELL(gd, c) +=  x;
    } else if (gd->datatype == INTEGER_GRID) {
	long a = round(x);
	if (a < INTEGER_MIN OR a > INTEGER_MAX) {
	    fprintf(stderr,"%s: %li: integer out of bounds\n",fct,a);
	    return 0;
	}
	FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) += x;
    } else {
	fprintf(stderr,"%s: grid datatype error\n",fct);
	return 0;
    }
    return 1;
}


int gdaddgd(grid *gd1, grid *gd2)
{
    static char *fct = "gdaddgd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == REAL_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		RGD_CELL(gd1, c) += RGD_CELL(gd2, c);
	    else
		RGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == REAL_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		RGD_CELL(gd1, c) += IGD_CELL(gd2, c);
	    else
		RGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c)) {
		long a = IGD_CELL(gd1, c) + round(RGD_CELL(gd2, c));
		if (a < INTEGER_MIN OR a > INTEGER_MAX) {
		    fprintf(stderr,"%s: %li: integer out of bounds\n",fct,a);
		    return 0;
		}
		IGD_CELL(gd1, c) = a;
	    } else
		IGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) += IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdsubgd(grid *gd1, grid *gd2)
{
    static char *fct = "gdsubgd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == REAL_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		RGD_CELL(gd1, c) -= RGD_CELL(gd2, c);
	    else
		RGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == REAL_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		RGD_CELL(gd1, c) -= IGD_CELL(gd2, c);
	    else
		RGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c)) {
		long a = IGD_CELL(gd1, c) - round(RGD_CELL(gd2, c));
		if (a < INTEGER_MIN OR a > INTEGER_MAX) {
		    fprintf(stderr,"%s: %li: integer out of bounds\n",fct,a);
		    return 0;
		}
		IGD_CELL(gd1, c) = a;
	    } else
		IGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) -= IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdmultsv(grid *gd, double x) 
{
    static char *fct = "gdmultsv";
    cell c;
    if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd)
	    if (RGD_DATACELL(gd, c))
		RGD_CELL(gd, c) *= x;
    } else if (gd->datatype == INTEGER_GRID) {
	long a = round(x);
	if (a < INTEGER_MIN OR a > INTEGER_MAX) {
	    fprintf(stderr,"%s: %li: integer out of bounds\n",fct,a);
	    return 0;
	}
	FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) *= a;
    } else {
	fprintf(stderr,"%s: grid datatype error\n",fct);
	return 0;
    }
    return 1;
}


int gdmultgd(grid *gd1, grid *gd2) 
{
    static char *fct = "gdmultgd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == REAL_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		RGD_CELL(gd1, c) *= RGD_CELL(gd2, c);
	    else
		RGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == REAL_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		RGD_CELL(gd1, c) *= IGD_CELL(gd2, c);
	    else
		RGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c)) {
		long a = IGD_CELL(gd1, c) * round(RGD_CELL(gd2, c));
		if (a < INTEGER_MIN OR a > INTEGER_MAX) {
		    fprintf(stderr,"%s: %li: integer out of bounds\n",fct,a);
		    return 0;
		}
		IGD_CELL(gd1, c) = a;
	    } else
		IGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) *= IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gddivsv(grid *gd, double x) 
{
    static char *fct = "gdivsv";
    cell c;
    if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd)
	    if (RGD_DATACELL(gd, c))
		RGD_CELL(gd, c) /= x;
    } else if (gd->datatype == INTEGER_GRID) {
	long a = round(x);
	if (a == 0) {
	    fprintf(stderr,"%s: divide by zero\n",fct);
	    return 0;
	}
	if (a < INTEGER_MIN OR a > INTEGER_MAX) {
	    fprintf(stderr,"%s: %li: integer out of bounds\n",fct,a);
	    return 0;
	}
	FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) /= a;
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int svdivgd(double x, grid *gd)
{
    static char *fct = "svdivgd";
    cell c;
    if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd) {
	    if (RGD_CELL(gd, c) == 0) {
		fprintf(stderr,"%s: divide by zero\n",fct);
		return 0;
	    }
	    if (RGD_DATACELL(gd, c))
		RGD_CELL(gd, c) = x / RGD_CELL(gd, c);
	}
    } else if (gd->datatype == INTEGER_GRID) {
	long a = round(x);
	if (a < INTEGER_MIN OR a > INTEGER_MAX) {
	    fprintf(stderr,"%s: %li: integer out of bounds\n",fct,a);
	    return 0;
	}
	FOREACH(c, gd) {
	    if (IGD_DATACELL(gd, c)) {
		INTEGER i = IGD_CELL(gd, c);
		if (i == 0) {
		    fprintf(stderr,"%s: divide by zero\n",fct);
		    return 0;
		}
		IGD_CELL(gd, c) = a/i;
	    }
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gddivgd(grid *gd1, grid *gd2) 
{
    static char *fct = "gdivgd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == REAL_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c)) {
		if (RGD_CELL(gd2, c) == 0) {
		    fprintf(stderr,"%s: divide by zero\n",fct);
		    return 0;
		}
		RGD_CELL(gd1, c) /= RGD_CELL(gd2, c);
	    } else
		RGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == REAL_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c)) {
		if (IGD_CELL(gd2, c) == 0) {
		    fprintf(stderr,"%s: divide by zero\n",fct);
		    return 0;
		}
		RGD_CELL(gd1, c) /= IGD_CELL(gd2, c);
	    } else
		RGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c)) {
		long a = round((double)IGD_CELL(gd1, c) / RGD_CELL(gd2, c));
		if (a < INTEGER_MIN OR a > INTEGER_MAX) {
		    fprintf(stderr,"%s: %li: integer out of bounds\n",fct,a);
		    return 0;
		}
		IGD_CELL(gd1, c) = a;
	    } else
		IGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c)) {
		if (IGD_CELL(gd2, c) == 0) {
		    fprintf(stderr,"%s: divide by zero\n",fct);
		    return 0;
		}
		IGD_CELL(gd1, c) /= IGD_CELL(gd2, c);
	    } else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdmodulussv(grid *gd, int x) 
{
    static char *fct = "gmodulussv";
    cell c;
    if (x == 0) {
	fprintf(stderr,"%s: modulus by zero\n",fct);
	return 0;
    }
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) %= x;
    } else {
	fprintf(stderr,"%s: not an integer grid\n", fct);
	return 0;
    }
    return 1;
}


int svmodulusgd(int x, grid *gd)
{
    static char *fct = "svmodulusgd";
    cell c;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd) {
	    if (IGD_DATACELL(gd, c)) {
		if (IGD_CELL(gd, c) == 0) {
		    fprintf(stderr,"%s: modulus by zero\n",fct);
		    return 0;
		}
		IGD_CELL(gd, c) = x % IGD_CELL(gd, c);
	    }
	}
    } else {
	fprintf(stderr,"%s: not an integer grid\n", fct);
	return 0;
    }
    return 1;
}


int gdmodulusgd(grid *gd1, grid *gd2) 
{
    static char *fct = "gmodulusgd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c)) {
		if (IGD_CELL(gd2, c) == 0) {
		    fprintf(stderr,"%s: modulus by zero\n",fct);
		    return 0;
		}
		IGD_CELL(gd1, c) %= IGD_CELL(gd2, c);
	    } else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: the grids are not both integer grids\n", fct);
	return 0;
    }
    return 1;
}


int gdpowersv(grid *gd, double x) 
{
    static char *fct = "gpowersv";
    cell c;
    if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd)
	    if (RGD_DATACELL(gd, c))
		RGD_CELL(gd, c) = pow(RGD_CELL(gd, c), x);
    } else if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd) {
	    if (IGD_DATACELL(gd, c)) {
		long a = round(pow((double)IGD_CELL(gd, c), x));
		if (a < INTEGER_MIN OR a > INTEGER_MAX) {
		    fprintf(stderr,"%s: %li: integer out of bounds\n",fct,a);
		    return 0;
		}
		IGD_CELL(gd, c) = a;
	    }
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int svpowergd(double x, grid *gd)
{
    static char *fct = "svgivgd";
    cell c;
    if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd) {
	    if (RGD_DATACELL(gd, c))
		RGD_CELL(gd, c) = pow(x, RGD_CELL(gd, c));
	}
    } else if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd) {
	    if (IGD_DATACELL(gd, c)) {
		long a = pow(x, (double)IGD_CELL(gd, c));
		if (a < INTEGER_MIN OR a > INTEGER_MAX) {
		    fprintf(stderr,"%s: %li: integer out of bounds\n",fct,a);
		    return 0;
		}
		IGD_CELL(gd, c) = a;
	    }
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdpowergd(grid *gd1, grid *gd2) 
{
    static char *fct = "gpowergd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == REAL_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		RGD_CELL(gd1, c) = pow(RGD_CELL(gd1, c), RGD_CELL(gd2, c));
	    else
		RGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == REAL_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		RGD_CELL(gd1, c) = pow(RGD_CELL(gd1, c), (double)IGD_CELL(gd2, c));
	    else
		RGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c)) {
		long a = round(pow((double)IGD_CELL(gd1, c), RGD_CELL(gd2, c)));
		if (a < INTEGER_MIN OR a > INTEGER_MAX) {
		    fprintf(stderr,"%s: %li: integer out of bounds\n",fct,a);
		    return 0;
		}
		IGD_CELL(gd1, c) = a;
	    } else
		IGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c)) {
		long a = round(pow((double)IGD_CELL(gd1, c), (double)IGD_CELL(gd2, c)));
		if (a < INTEGER_MIN OR a > INTEGER_MAX) {
		    fprintf(stderr,"%s: %li: integer out of bounds\n",fct,a);
		    return 0;
		}
		IGD_CELL(gd1, c) = a;
	    } else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdabs(grid *gd) 
{
    static char *fct = "gdabs";
    cell c;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) = abs(IGD_CELL(gd, c));
    } else if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd)
	    if (RGD_DATACELL(gd, c))
		RGD_CELL(gd, c) = fabs(RGD_CELL(gd, c));
    }
    else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdacos(grid *gd) 
{
    static char *fct = "gdacos";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c))
	    RGD_CELL(gd, c) = acos(RGD_CELL(gd, c));
    return 1;
}


int gdatan(grid *gd) 
{
    static char *fct = "gdatan";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c))
	    RGD_CELL(gd, c) = atan(RGD_CELL(gd, c));
    return 1;
}


int gdatan2(grid *gd1, grid *gd2) 
{
    static char *fct = "gdatan2";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype != REAL_GRID OR gd2->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grids are not both floating point grids\n", fct);
	return 0;
    }
    FOREACH(c, gd1) {
	if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
	    RGD_CELL(gd1, c) = atan2(RGD_CELL(gd1, c),RGD_CELL(gd2, c));
	else
	    RGD_SETNODATACELL(gd1, c);
    }
    return 1;
}


int gdceil(grid *gd) 
{
    static char *fct = "gdceil";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c))
	    RGD_CELL(gd, c) = ceil(RGD_CELL(gd, c));
    return 1;
}


int gdcos(grid *gd) 
{
    static char *fct = "gdcos";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c))
	    RGD_CELL(gd, c) = cos(RGD_CELL(gd, c));
    return 1;
}


int gdcosh(grid *gd) 
{
    static char *fct = "gdcosh";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c))
	    RGD_CELL(gd, c) = cosh(RGD_CELL(gd, c));
    return 1;
}


int gdexp(grid *gd) 
{
    static char *fct = "gdexp";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c))
	    RGD_CELL(gd, c) = exp(RGD_CELL(gd, c));
    return 1;
}


int gdfloor(grid *gd) 
{
    static char *fct = "gdfloor";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c))
	    RGD_CELL(gd, c) = floor(RGD_CELL(gd, c));
    return 1;
}


int gdlog(grid *gd)
{
    static char *fct = "gdlog";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c))
	    RGD_CELL(gd, c) = log(RGD_CELL(gd, c));
    return 1;
}


int gdlog10(grid *gd)
{
    static char *fct = "gdlog10";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c)) {
	    double x = RGD_CELL(gd, c);
	    if (x <= 0) {
		fprintf(stderr,"%s: %f: can't take log from this\n", fct, x);
		return 0;
	    }
	    RGD_CELL(gd, c) = log10(x);
	}
    return 1;
}


int gdpow(grid *gd, double b)
{
    static char *fct = "gdpow";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c))
	    RGD_CELL(gd, c) = pow(RGD_CELL(gd, c),b);
    return 1;
}


int gdsin(grid *gd)
{
    static char *fct = "gdsin";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c))
	    RGD_CELL(gd, c) = sin(RGD_CELL(gd, c));
    return 1;
}


int gdsinh(grid *gd)
{
    static char *fct = "gdsinh";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c))
	    RGD_CELL(gd, c) = sinh(RGD_CELL(gd, c));
    return 1;
}


int gdsqrt(grid *gd)
{
    static char *fct = "gdsqrt";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c))
	    RGD_CELL(gd, c) = sqrt(RGD_CELL(gd, c));
    return 1;
}


int gdtan(grid *gd)
{
    static char *fct = "gdtan";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c))
	    RGD_CELL(gd, c) = tan(RGD_CELL(gd, c));
    return 1;
}


int gdtanh(grid *gd)
{
    static char *fct = "gdtanh";
    cell c;
    if (gd->datatype != REAL_GRID) {
	fprintf(stderr,"%s: the grid is not a floating point grid\n", fct);
	return 0;
    }
    FOREACH(c, gd)
	if (RGD_DATACELL(gd, c))
	    RGD_CELL(gd, c) = tanh(RGD_CELL(gd, c));
    return 1;
}


grid *gdround(grid *gd)
{
    static char *fct = "gdround";
    cell c;
    grid *g;
    if (!(g = gdnew2(INTEGER_GRID, gd))) return NULL;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd)
	    GDSET2(g, c, IGD_CELL(gd, c));
    } else if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd)
	    GDSET2(g, c, round(RGD_CELL(gd, c)));
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	gddestroy(g);
	return NULL;
    }
    g->minval = gd->minval;
    g->maxval = gd->maxval;
    g->snodata = gd->snodata;
    g->fnodata = gd->fnodata;
    return g;
}


int gdltsv(grid *gd, double x)
{
    static char *fct = "gdltsv";
    cell c;
    if (gd->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd);
	if (!g) return 0;
        FOREACH(c, gd) {
	    if (RGD_DATACELL(gd, c))
		IGD_CELL(g, c) = RGD_CELL(gd, c) < x;
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd, g);
	gddestroy(g);
    } else if (gd->datatype == INTEGER_GRID) {
        long a = round(x);
        FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) = IGD_CELL(gd, c) < a;
    } else {
        fprintf(stderr,"%s: grid datatype error\n", fct);
        return 0;
    }
    return 1;
}


int gdgtsv(grid *gd, double x)
{
    static char *fct = "gdgtsv";
    cell c;
    if (gd->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd);
	if (!g) return 0;
	FOREACH(c, gd) {
	    if (RGD_DATACELL(gd, c))
		IGD_CELL(g, c) = RGD_CELL(gd, c) > x;
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd, g);
	gddestroy(g);
    } else if (gd->datatype == INTEGER_GRID) {
	long a = round(x);
	FOREACH(c, gd) {
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) = IGD_CELL(gd, c) > a;
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdlesv(grid *gd, double x)
{
    static char *fct = "gdlesv";
    cell c;
    if (gd->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd);
	if (!g) return 0;
        FOREACH(c, gd) {
	    if (RGD_DATACELL(gd, c))
		IGD_CELL(g, c) = RGD_CELL(gd, c) <= x;
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd, g);
	gddestroy(g);
    } else if (gd->datatype == INTEGER_GRID) {
        long a = round(x);
        FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) = IGD_CELL(gd, c) <= a;
    } else {
        fprintf(stderr,"%s: grid datatype error\n", fct);
        return 0;
    }
    return 1;
}


int gdgesv(grid *gd, double x)
{
    static char *fct = "gdgesv";
    cell c;
    if (gd->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd);
	if (!g) return 0;
        FOREACH(c, gd) {
	    if (RGD_DATACELL(gd, c))
		IGD_CELL(g, c) = RGD_CELL(gd, c) >= x;
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd, g);
	gddestroy(g);
    } else if (gd->datatype == INTEGER_GRID) {
        long a = round(x);
        FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) = IGD_CELL(gd, c) >= a;
    } else {
        fprintf(stderr,"%s: grid datatype error\n", fct);
        return 0;
    }
    return 1;
}


int gdeqsv(grid *gd, double x)
{
    static char *fct = "gdeqsv";
    cell c;
    if (gd->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd);
	if (!g) return 0;
        FOREACH(c, gd) {
	    if (RGD_DATACELL(gd, c))
		IGD_CELL(g, c) = RGD_CELL(gd, c) == x;
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd, g);
	gddestroy(g);
    } else if (gd->datatype == INTEGER_GRID) {
        long a = round(x);
        FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) = IGD_CELL(gd, c) == a;
    } else {
        fprintf(stderr,"%s: grid datatype error\n", fct);
        return 0;
    }
    return 1;
}


int gdnesv(grid *gd, double x)
{
    static char *fct = "gdnesv";
    cell c;
    if (gd->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd);
	if (!g) return 0;
        FOREACH(c, gd) {
	    if (RGD_DATACELL(gd, c))
		IGD_CELL(g, c) = RGD_CELL(gd, c) != x;
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd, g);
	gddestroy(g);
    } else if (gd->datatype == INTEGER_GRID) {
        long a = round(x);
        FOREACH(c, gd) {
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) = IGD_CELL(gd, c) != a;
	}
    } else {
        fprintf(stderr,"%s: grid datatype error\n", fct);
        return 0;
    }
    return 1;
}


int gdcmpsv(grid *gd, double x)
{
    static char *fct = "gdcmpsv";
    cell c;
    if (gd->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd);
	if (!g) return 0;
	FOREACH(c, gd) {
	    if (RGD_DATACELL(gd, c))
		IGD_CELL(g, c) = RGD_CELL(gd, c) > x ? 1 : (RGD_CELL(gd, c) == x ? 0 : -1);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd, g);
	gddestroy(g);
    } else if (gd->datatype == INTEGER_GRID) {
	long a = round(x);
	FOREACH(c, gd) {
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) = IGD_CELL(gd, c) > a ? 1 : (IGD_CELL(gd, c) == a ? 0 : -1);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdltgd(grid *gd1, grid *gd2)
{
    static char *fct = "gdltgd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == REAL_GRID AND gd2->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd1);
	if (!g) return 0;
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(g, c) = RGD_CELL(gd1, c) < RGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd1, g);
	gddestroy(g);
    } else if (gd1->datatype == REAL_GRID AND gd2->datatype == INTEGER_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd1);
	if (!g) return 0;
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(g, c) = RGD_CELL(gd1, c) < IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd1, g);
	gddestroy(g);
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) < round(RGD_CELL(gd2, c));
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) < IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdgtgd(grid *gd1, grid *gd2) 
{
    static char *fct = "gdgtgd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == REAL_GRID AND gd2->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd1);
	if (!g) return 0;
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(g, c) = RGD_CELL(gd1, c) > RGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd1, g);
	gddestroy(g);
    } else if (gd1->datatype == REAL_GRID AND gd2->datatype == INTEGER_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd1);
	if (!g) return 0;
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(g, c) = RGD_CELL(gd1, c) > IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd1, g);
	gddestroy(g);
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) > round(RGD_CELL(gd2, c));
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) > IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdlegd(grid *gd1, grid *gd2) 
{
    static char *fct = "gdlegd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == REAL_GRID AND gd2->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd1);
	if (!g) return 0;
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(g, c) = RGD_CELL(gd1, c) <= RGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd1, g);
	gddestroy(g);
    } else if (gd1->datatype == REAL_GRID AND gd2->datatype == INTEGER_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd1);
	if (!g) return 0;
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(g, c) = RGD_CELL(gd1, c) <= IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd1, g);
	gddestroy(g);
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) <= round(RGD_CELL(gd2, c));
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) <= IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdgegd(grid *gd1, grid *gd2) 
{
    static char *fct = "gdgegd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == REAL_GRID AND gd2->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd1);
	if (!g) return 0;
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(g, c) = RGD_CELL(gd1, c) >= RGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd1, g);
	gddestroy(g);
    } else if (gd1->datatype == REAL_GRID AND gd2->datatype == INTEGER_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd1);
	if (!g) return 0;
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(g, c) = RGD_CELL(gd1, c) >= IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd1, g);
	gddestroy(g);
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) >= round(RGD_CELL(gd2, c));
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) >= IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdeqgd(grid *gd1, grid *gd2)
{
    static char *fct = "gdeqgd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == REAL_GRID AND gd2->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd1);
	if (!g) return 0;
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(g, c) = RGD_CELL(gd1, c) == RGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd1, g);
	gddestroy(g);
    } else if (gd1->datatype == REAL_GRID AND gd2->datatype == INTEGER_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd1);
	if (!g) return 0;
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(g, c) = RGD_CELL(gd1, c) == IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd1, g);
	gddestroy(g);
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) == round(RGD_CELL(gd2, c));
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) == IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdnegd(grid *gd1, grid *gd2)
{
    static char *fct = "gdnegd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == REAL_GRID AND gd2->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd1);
	if (!g) return 0;
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(g, c) = RGD_CELL(gd1, c) != RGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd1, g);
	gddestroy(g);
    } else if (gd1->datatype == REAL_GRID AND gd2->datatype == INTEGER_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd1);
	if (!g) return 0;
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(g, c) = RGD_CELL(gd1, c) != IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd1, g);
	gddestroy(g);
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) != round(RGD_CELL(gd2, c));
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) != IGD_CELL(gd2, c);
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdcmpgd(grid *gd1, grid *gd2)
{
    static char *fct = "gdcmpgd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == REAL_GRID AND gd2->datatype == REAL_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd1);
	if (!g) return 0;
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(g, c) = RGD_CELL(gd1, c) > RGD_CELL(gd2, c) ? 
		    1 : (RGD_CELL(gd1, c) == RGD_CELL(gd2, c) ? 0 : -1);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd1, g);
	gddestroy(g);
    } else if (gd1->datatype == REAL_GRID AND gd2->datatype == INTEGER_GRID) {
	grid *g = gdnew2(INTEGER_GRID, gd1);
	if (!g) return 0;
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(g, c) = RGD_CELL(gd1, c) > IGD_CELL(gd2, c) ? 
		    1 : (RGD_CELL(gd1, c) == IGD_CELL(gd2, c) ? 0 : -1);
	    else
		IGD_SETNODATACELL(g, c);
	}
	gdstealdata(gd1, g);
	gddestroy(g);
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) > round(RGD_CELL(gd2, c)) ? 
		    1 : (IGD_CELL(gd1, c) == round(RGD_CELL(gd2, c)) ? 0 : -1);
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = IGD_CELL(gd1, c) > IGD_CELL(gd2, c) ? 
		    1 : (IGD_CELL(gd1, c) == IGD_CELL(gd2, c) ? 0 : -1);
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdminsv(grid *gd, double x)
{
    static char *fct = "gdminsv";
    cell c;
    if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd)
	    if (RGD_DATACELL(gd, c))
		RGD_CELL(gd, c) = min(RGD_CELL(gd, c),x);
    } else if (gd->datatype == INTEGER_GRID) {
	long a = round(x);
	FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) = min(IGD_CELL(gd, c),a);
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdmaxsv(grid *gd, double x)
{
    static char *fct = "gdmaxsv";
    cell c;
    if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd)
	    if (RGD_DATACELL(gd, c))
		RGD_CELL(gd, c) = max(RGD_CELL(gd, c),x);
    } else if (gd->datatype == INTEGER_GRID) {
	long a = round(x);
	FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) = max(IGD_CELL(gd, c),a);
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdmingd(grid *gd1, grid *gd2)
{
    static char *fct = "gdmingd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == REAL_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		RGD_CELL(gd1, c) = min(RGD_CELL(gd1, c),RGD_CELL(gd2, c));
	    else
		RGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == REAL_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		RGD_CELL(gd1, c) = min(RGD_CELL(gd1, c),IGD_CELL(gd2, c));
	    else
		RGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = min(IGD_CELL(gd1, c),round(RGD_CELL(gd2, c)));
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = min(IGD_CELL(gd1, c),IGD_CELL(gd2, c));
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdmaxgd(grid *gd1, grid *gd2)
{
    static char *fct = "gdmaxgd";
    cell c;
    if (!gdoverlayable(gd1, gd2)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (gd1->datatype == REAL_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		RGD_CELL(gd1, c) = max(RGD_CELL(gd1, c),RGD_CELL(gd2, c));
	    else
		RGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == REAL_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (RGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		RGD_CELL(gd1, c) = max(RGD_CELL(gd1, c),IGD_CELL(gd2, c));
	    else
		RGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == REAL_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND RGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = max(IGD_CELL(gd1, c),round(RGD_CELL(gd2, c)));
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else if (gd1->datatype == INTEGER_GRID AND gd2->datatype == INTEGER_GRID) {
	FOREACH(c, gd1) {
	    if (IGD_DATACELL(gd1, c) AND IGD_DATACELL(gd2, c))
		IGD_CELL(gd1, c) = max(IGD_CELL(gd1, c),IGD_CELL(gd2, c));
	    else
		IGD_SETNODATACELL(gd1, c);
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int cmp_int(const void *a, const void *b)
{
    if (*(int *)a < *(int *)b) return -1;
    if (*(int *)a > *(int *)b) return 1;
    return 0;
}


grid *gdcross(grid *a, grid *b)
{
    static char *fct = "gdcross";
    grid *c = NULL;
    cell p;
    hash atable,btable;
    int *ca = NULL, na, *cb = NULL, nb;
    if (a->datatype != INTEGER_GRID OR b->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: the grids have to be integer grids\n",fct);
	return NULL;
    }
    if (!gdoverlayable(a, b)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return NULL;
    }
    if (!hash_create(&atable, 200)) goto fail;
    if (!hash_create(&btable, 200)) goto fail;
    if (!(gdcontents(a, &atable))) goto fail;
    if (!(gdcontents(b, &btable))) goto fail;
    if (!(ca = hash_keys(&atable, &na))) goto fail;
    if (!(cb = hash_keys(&btable, &nb))) goto fail;
    if (!(c = gdnew2(INTEGER_GRID, a))) goto fail;
    qsort(ca, na, sizeof(int), &cmp_int);
    qsort(cb, nb, sizeof(int), &cmp_int);
    FOREACH(p, a) {
	if (IGD_DATACELL(a, p) AND IGD_DATACELL(b, p)) {
	    int ia = 0, ib = 0;
	    while (ia < na AND ca[ia] != IGD_CELL(a, p)) ia++;
	    while (ib < nb AND cb[ib] != IGD_CELL(b, p)) ib++;
	    IGD_CELL(c, p) = ib + ia*na + 1;
	} else 
	    IGD_SETNODATACELL(c, p);
    }
 fail:
    hash_destroy(&atable);
    hash_destroy(&btable);
    if (ca) free(ca);
    if (cb) free(cb);
    return c;
}


/* if a then b = c */
int gdif_thensv(grid *a, grid *b, double c)
{
    static char *fct = "gdif_thensv";
    cell p;
    if (a->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: the grid 'a' is not an integer grid\n",fct);
	return 0;
    }
    if (!gdoverlayable(a, b)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (b->datatype == INTEGER_GRID) {
	long x = round(c);
	if (x < INTEGER_MIN OR x > INTEGER_MAX) {
	    fprintf(stderr,"%s: %li: integer out of bounds\n",fct,x);
	    return 0;
	}
	FOREACH(p, a)
	    if (IGD_DATACELL(a, p) AND IGD_CELL(a, p)) IGD_CELL(b, p) = x;
    } else if (b->datatype == REAL_GRID) {
	FOREACH(p, a)
	    if (IGD_DATACELL(a, p) AND IGD_CELL(a, p)) RGD_CELL(b, p) = c;
    } else {
	fprintf(stderr,"%s: grid datatype error\n",fct);
	return 0;
    }
    return 1;
}

/* if a then b = c */
int gdif_thengd(grid *a, grid *b, grid *c)
{
    static char *fct = "gdif_thengd";
    cell p;
    if (a->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: the grid 'a' is not an integer grid\n",fct);
	return 0;
    }
    if (!gdoverlayable(a, b) OR !gdoverlayable(a, c)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (b->datatype == INTEGER_GRID) {
	if (c->datatype == INTEGER_GRID) {
	    FOREACH(p, a)
		if (IGD_DATACELL(a, p) AND IGD_DATACELL(c, p) AND IGD_CELL(a, p)) 
		    IGD_CELL(b, p) = IGD_CELL(c, p);
	} else if (c->datatype == REAL_GRID) {
	    FOREACH(p, a) {
		if (IGD_DATACELL(a, p) AND RGD_DATACELL(c, p) AND IGD_CELL(a, p)) {
		    long x = round(RGD_CELL(c, p));
		    if (x < INTEGER_MIN OR x > INTEGER_MAX) {
			fprintf(stderr,"%s: %li: integer out of bounds\n",fct,x);
			return 0;
		    }
		    IGD_CELL(b, p) = x;
		}
	    }
	} else {
	    fprintf(stderr,"%s: grid datatype error\n",fct);
	    return 0;
	}
    } else if (b->datatype == REAL_GRID) {
	if (c->datatype == INTEGER_GRID) {
	    FOREACH(p, a)
		if (IGD_DATACELL(a, p) AND IGD_DATACELL(c, p) AND IGD_CELL(a, p)) 
		    RGD_CELL(b, p) = IGD_CELL(c, p);
	} else if (c->datatype == REAL_GRID) {
	    FOREACH(p, a)
		if (IGD_DATACELL(a, p) AND RGD_DATACELL(c, p) AND IGD_CELL(a, p)) 
		    RGD_CELL(b, p) = RGD_CELL(c, p);
	} else {
	    fprintf(stderr,"%s: grid datatype error\n",fct);
	    return 0;
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n",fct);
	return 0;
    }
    return 1;
}


int gdzonal_if_then(grid *a, grid *b, int *k, double *v, int n)
{
    static char *fct = "gdzonal_if_then";
    cell p;
    if (a->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: the zones grid is not an integer grid\n",fct);
	return 0;
    }
    if (!gdoverlayable(a, b)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (b->datatype == INTEGER_GRID) {
	FOREACH(p, a) {
	    int i = 0;
	    long x;
	    if (IGD_NODATACELL(a, p)) continue;
	    while (i < n AND k[i] != IGD_CELL(a, p)) i++;
	    if (i == n)
		continue;
	    x = round(v[i]);
	    if (x < INTEGER_MIN OR x > INTEGER_MAX) {
		fprintf(stderr,"%s: %li: integer out of bounds\n",fct,x);
		return 0;
	    }
	    IGD_CELL(b, p) = x;
	}
    } else if (b->datatype == REAL_GRID) {
	FOREACH(p, a) {
	    int i = 0;
	    if (IGD_NODATACELL(a, p)) continue;
	    while (i < n AND k[i] != IGD_CELL(a, p)) i++;
	    if (i == n)
		continue;
	    RGD_CELL(b, p) = v[i];
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n",fct);
	return 0;
    }
    return 1;
}


int gdbinary(grid *gd)
{
    static char *fct = "gdbinary";
    if (gd->datatype == INTEGER_GRID) {
	cell c;
	FOREACH(c, gd) 
	    if (IGD_DATACELL(gd, c))
		IGD_CELL(gd, c) = IGD_CELL(gd, c) > 0;
    } else {
	fprintf(stderr,"%s: not an integer grid\n",fct);
	return 0;
    }
    return 1;
}


int gdapplytempl(grid *gd, int *templ, int new_val)
{
    static char *fct = "gdapplytempl";
    int matches = 0;
    if (gd->datatype == INTEGER_GRID) {
	cell c;
	FOREACH(c, gd) {
	    /* go through the 3x3 neighborhood 
	       in a bit different way than what's usual here */
	    cell a = c;
	    int match = 1;
	    debug(1,"apply templ %i,%i\n", c.i, c.j);
	    while (1) { /* the intention is not to loop 
			   but break out in the middle */
		/* 0 */
		a.i--;
		a.j--;
		if (templ[0] >= 0) {
		    if (a.i < 0 OR a.j < 0) 
			match = templ[0] == 0; /* we assume outside is 0 */
		    else 
			match = (templ[0] AND IGD_CELL(gd, a)) OR 
			    (!templ[0] AND !IGD_CELL(gd, a));
		    if (!match) break;	  
		}
		/* 1 */
		a.j++;
		if (templ[1] >= 0) {
		    if (a.i < 0) 
			match = templ[1] == 0; 
		    else 
			match = (templ[1] AND IGD_CELL(gd, a)) OR 
			    (!templ[1] AND !IGD_CELL(gd, a));
		    if (!match) break;	  
		}
		/* 2 */
		a.j++;
		if (templ[2] >= 0) {
		    if (a.i < 0 OR a.j == gd->N) 
			match = templ[2] == 0; 
		    else 
			match = (templ[2] AND IGD_CELL(gd, a)) OR 
			    (!templ[2] AND !IGD_CELL(gd, a));
		    if (!match) break;	  
		}
		/* 3 */
		a.i++;
		a.j-=2;
		if (templ[3] >= 0) {	    
		    if (a.j < 0) 
			match = templ[3] == 0; 
		    else 
			match = (templ[3] AND IGD_CELL(gd, a)) OR 
			    (!templ[3] AND !IGD_CELL(gd, a));
		    if (!match) break;	  
		}
		/* 4 */
		a.j++;
		if (templ[4] >= 0) {
		    match = (templ[4] AND IGD_CELL(gd, a)) OR 
			(!templ[4] AND !IGD_CELL(gd, a));
		    if (!match) break;	  
		}
		/* 5 */
		a.j++;
		if (templ[5] >= 0) {
		    if (a.j == gd->N) 
			match = templ[5] == 0; 
		    else 
			match = (templ[5] AND IGD_CELL(gd, a)) OR 
			    (!templ[5] AND !IGD_CELL(gd, a));
		    if (!match) break;	  
		}
		/* 6 */	  
		a.i++;
		a.j-=2;
		if (templ[6] >= 0) {
		    if (a.i == gd->M OR a.j < 0) 
			match = templ[6] == 0; 
		    else 
			match = (templ[6] AND IGD_CELL(gd, a)) OR 
			    (!templ[6] AND !IGD_CELL(gd, a));
		    if (!match) break;	  
		}
		/* 7 */
		a.j++;
		if (templ[7] >= 0) {
		    if (a.i == gd->M) 
			match = templ[7] == 0; 
		    else 
			match = (templ[7] AND IGD_CELL(gd, a)) OR 
			    (!templ[7] AND !IGD_CELL(gd, a));
		    if (!match) break;	  
		}
		/* 8 */
		a.j++;
		if (templ[8] >= 0) {
		    if (a.i == gd->M OR a.j == gd->N) 
			match = templ[8] == 0; 
		    else 
			match = (templ[8] AND IGD_CELL(gd, a)) OR 
			    (!templ[8] AND !IGD_CELL(gd, a));
		    if (!match) break;	  
		}
		break;
	    }
	    if (match) {
		IGD_CELL(gd, c) = 2;
		matches++;
	    }
	}
	FOREACH(c, gd) 
	    if (IGD_CELL(gd, c) > 1) IGD_CELL(gd, c) = new_val;
    } else {
	fprintf(stderr,"%s: not an integer grid\n",fct);
	return -1;
    }
    return matches;
}


int gdmap(grid *gd, int *s, int *d, int n)
{
    static char *fct = "gdmap";
    cell c;
    int k;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd) {
	    for (k=0;k<n;k++) {
		if (s[k] > IGD_CELL(gd, c))
		    break;
		else if (s[k] == IGD_CELL(gd, c)) {
		    IGD_CELL(gd, c) = d[k];
		    break;
		}
	    }
	}
    } else {
	fprintf(stderr,"%s: not an integer grid\n",fct);
	return 0;
    }
    return 1;
}


int _gdzonesize(grid *gd, grid *visited, cell c, int color, int connectivity)
{
    int lastBorder;
    int leftLimit, rightLimit;
    int i2;
    int size = 0;
    if (IGD_CELL(gd, c) != color) return 0;
    /* Seek left */
    leftLimit = -1;
    for (i2 = c.i; i2 >= 0; i2--) {
	if (IGD_AT(gd, i2, c.j) != color) break;
	size++;
	IGD_AT(visited, i2, c.j) = 1;
	if (connectivity == 8)
	    leftLimit = max(0,i2-1);
	else
	    leftLimit = i2;
    }
    if (leftLimit == -1) return 1;
    /* Seek right */
    if (connectivity == 8)
	rightLimit = min(gd->M-1,c.i+1);
    else
	rightLimit = c.i;
    for (i2 = c.i+1; i2 < gd->M; i2++) {	
	if (IGD_AT(gd, i2, c.j) != color) break;
	size++;
	IGD_AT(visited, i2, c.j) = 1;
	if (connectivity == 8)
	    rightLimit = min(gd->M-1,i2+1);
	else
	    rightLimit = i2;
    }
    /* Look at lines above and below */
    /* Above */
    if (c.j > 0) {
	cell c2;
	c2.j = c.j-1;
	lastBorder = 1;
	for (c2.i = leftLimit; c2.i <= rightLimit; c2.i++) {
	    int a;
	    a = IGD_CELL(gd, c2);
	    if (lastBorder) {
		if (a == color) {	
		    if (!IGD_CELL(visited, c2))
			size += _gdzonesize(gd, visited, c2, color, connectivity);
		    lastBorder = 0;
		}
	    } else if (a != color)
		lastBorder = 1;
	}
    }
    /* Below */
    if (c.j < gd->N - 1) {
	cell c2;
	c2.j = c.j+1;
	lastBorder = 1;
	for (c2.i = leftLimit; c2.i <= rightLimit; c2.i++) {
	    int a;
	    a = IGD_CELL(gd, c2);
	    if (lastBorder) {
		if (a == color) {
		    if (!IGD_CELL(visited, c2))
			size += _gdzonesize(gd, visited, c2, color, connectivity);
		    lastBorder = 0;
		}
	    } else if (a != color)
		lastBorder = 1;
	}
    }
    return size;
}


int _gdrzonesize(grid *gd, grid *visited, cell c, double color, int connectivity)
{
    int lastBorder;
    int leftLimit, rightLimit;
    int i2;
    int size = 0;
    if (RGD_CELL(gd, c) != color) return 0;
    /* Seek left */
    leftLimit = -1;
    for (i2 = c.i; i2 >= 0; i2--) {
	if (RGD_AT(gd, i2, c.j) != color) break;
	size++;
	IGD_AT(visited, i2, c.j) = 1;
	if (connectivity == 8)
	    leftLimit = max(0,i2-1);
	else
	    leftLimit = i2;
    }
    if (leftLimit == -1) return 1;
    /* Seek right */
    if (connectivity == 8)
	rightLimit = min(gd->M-1,c.i+1);
    else
	rightLimit = c.i;
    for (i2 = c.i+1; i2 < gd->M; i2++) {	
	if (RGD_AT(gd, i2, c.j) != color) break;
	size++;
	IGD_AT(visited, i2, c.j) = 1;
	if (connectivity == 8)
	    rightLimit = min(gd->M-1,i2+1);
	else
	    rightLimit = i2;
    }
    /* Look at lines above and below */
    /* Above */
    if (c.j > 0) {
	cell c2;
	c2.j = c.j-1;
	lastBorder = 1;
	for (c2.i = leftLimit; c2.i <= rightLimit; c2.i++) {
	    int a;
	    a = RGD_CELL(gd, c2);
	    if (lastBorder) {
		if (a == color) {
		    if (!IGD_CELL(visited, c2))
			size += _gdrzonesize(gd, visited, c2, color, connectivity);
		    lastBorder = 0;
		}
	    } else if (a != color)
		lastBorder = 1;
	}
    }
    /* Below */
    if (c.j < gd->N - 1) {
	cell c2;
	c2.j = c.j+1;
	lastBorder = 1;
	for (c2.i = leftLimit; c2.i <= rightLimit; c2.i++) {
	    int a;
	    a = RGD_CELL(gd, c2);
	    if (lastBorder) {
		if (a == color) {
		    if (!IGD_CELL(visited, c2))
			size += _gdrzonesize(gd, visited, c2, color, connectivity);
		    lastBorder = 0;
		}
	    } else if (a != color)
		lastBorder = 1;
	}
    }
    return size;
}


double gdzonesize(grid *gd, cell c)
{
    static char *fct = "gdsize_of_an_area";
    grid *visited = gdnew2(INTEGER_GRID, gd);
    double size = 0;
    if (!visited) return 0;
    if (GD_CELLOUT(gd, c)) {
	fprintf(stderr,"%s: %i,%i: initial cell not within the grid\n", fct, c.i, c.j);
	gddestroy(visited);
	return 0;
    }
    if (gd->datatype == INTEGER_GRID) {
	size = gd->unitdist*gd->unitdist*
	  (double)_gdzonesize(gd, visited, c, IGD_CELL(gd, c), 8);
    } else if (gd->datatype == REAL_GRID) {
	size = gd->unitdist*gd->unitdist*
	  (double)_gdrzonesize(gd, visited, c, RGD_CELL(gd, c), 8);
    } else {
	fprintf(stderr,"%s: grid datatype error\n",fct);
    }
    gddestroy(visited);
    return size;
}


grid *gdborders(grid *gd)
{
    static char *fct = "gdborders";
    grid *borders;
    cell c;
    if (gd->datatype == INTEGER_GRID) {
	if (!(borders = gdcreatecopy(gd,INTEGER_GRID))) return NULL;
	for (c.i = 0; c.i < gd->M; c.i++) {
	    int j0 = 0;
	    int on = IGD_AT(gd, c.i, 0);
	    for (c.j = 0; c.j < gd->N; c.j++)
		if (!mask OR (GD_CELLIN(mask, (c)) AND IGD_CELL(mask, (c)))) {
		    if (!IGD_CELL(gd, c)) {
			if (on) {
			    if ((c.i > 0) AND (c.i < gd->M)) {
				int jc;
				for (jc = j0+1; jc < c.j-1; jc++) {
				    if (IGD_AT(gd,c.i-1,jc) AND IGD_AT(gd,c.i+1,jc)) 
					IGD_AT(borders,c.i,jc) = 0;
				}
			    }
			    on = 0;
			}
		    } else if (!on) {
			on = 1;
			j0 = c.j;
		    }
		}
	    if (on) {
		if ((c.i > 0) AND (c.i < gd->M)) {
		    int jc;
		    for (jc = j0+1; jc < c.j-1; jc++) {
			if (IGD_AT(gd,c.i-1,jc) AND IGD_AT(gd,c.i+1,jc)) 
			    IGD_AT(borders,c.i,jc) = 0;
		    }
		}
	    }
	}
    } else {
	fprintf(stderr,"%s: not an integer grid\n",fct);
	return NULL;
    }
    return borders;
}


void gdmark_borders(grid *gd, cell c, grid *visited, grid *borders)
{
    int paint = IGD_CELL(gd, c);
    int d;
    debug(1,"gdmark_borders at %i,%i paint=%i\n", c.i, c.j, paint);
    IGD_CELL(visited, c) = 1;
    for (d = 0; d < 9; d++) {
	cell t = cellmove(c, d);
	if (GD_CELLIN(gd, t)) {
	    if (IGD_CELL(gd, t) != paint)
		IGD_CELL(borders, c) = paint;
	    else if (!IGD_CELL(visited, t))
		gdmark_borders(gd, t, visited, borders);
	} else 
	    IGD_CELL(borders, c) = paint;
    }
}


grid *gdborders2(grid *gd)
{
    static char *fct = "gdborders2";
    grid *borders, *visited;
    cell c;
    if (gd->datatype == INTEGER_GRID) {
	if (!(borders = gdnew2(INTEGER_GRID, gd))) return NULL;
	if (!(visited = gdnew2(INTEGER_GRID, gd))) {
	    gddestroy(borders);
	    return NULL;
	}
	FOREACH(c, gd)
	    if (IGD_CELL(gd, c) AND !IGD_CELL(visited, c))
		gdmark_borders(gd, c, visited, borders);
    } else {
	fprintf(stderr,"%s: not an integer grid\n", fct);
	return NULL;
    }
    debug(0, "gddestroy visited:\n");
    gddestroy(visited);
    debug(0, "%s: done\n", fct);
    return borders;
}


grid *gdareas(grid *gd, int k)
{
    static char *fct = "gdareas";
    grid *areas;
    if (gd->datatype == INTEGER_GRID) {
	cell c;
	if (!(areas = gdnew2(INTEGER_GRID, gd))) return NULL;
	FOREACH(c, gd) {
	    /* cell is part of an area 
	       if there is at least 
	       k consecutive nonzero cells as neighbors */
	    int r, d, co = 0, cm = 0;
	    if (!IGD_CELL(gd, c) OR IGD_NODATACELL(gd, c)) continue;
	    for (r = 0; r < 2; r++) {
		for (d = 1; d < 9; d++) {
		    cell x = cellmove(c, d);
		    if (GD_CELLIN(gd, x) AND IGD_CELL(gd, x))
			co++;
		    else {
			if (co > cm) cm = co;
			co = 0;
		    }
		}
	    }
	    if (co > cm) cm = co;
	    debug(0,"%i neighbors at %i,%i\n",cm, c.i, c.j);
	    if (cm >= k) IGD_CELL(areas, c) = IGD_CELL(gd, c);
	}
    } else {
	fprintf(stderr,"%s: not an integer grid\n",fct);
	return NULL;
    }
    return areas;
}


int gdconnect(grid *gd)
{
    static char *fct = "gdconnect";
    cell c;
    PROGRESS_BAR_VARS(gd->M/40);
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd) {
	    PROGRESS_CHECK(c.i);
	    /* connect within 3x3 mask */
	    if (IGD_CELL(gd, c)) continue;
	    else {
		int d, k = 0;
		/* d <-> d + 3,4,5 */
		for (d = 1; d < 9; d++) {
		    int e = d + 4;
		    cell cd = c, ce = c;
		    if (e > 8) e -= 8;  
		    cd = cellmove(cd, d);
		    ce = cellmove(ce, e);
		    if (GD_CELLIN(gd, cd) AND GD_CELLIN(gd, ce) AND 
			IGD_CELL(gd, cd) AND IGD_CELL(gd, cd) == IGD_CELL(gd, ce)) {
			k =  IGD_CELL(gd, cd);
			break;
		    }
		}
		IGD_CELL(gd, c) = k;
	    }
	}
    } else {
	fprintf(stderr,"%s: not an integer grid\n",fct);
	return 0;
    }
    PROGRESS_BAR_CLOSE;
    return 1;
}


int gdnrareas(grid *gd, int connectivity)
{
    static char *fct = "gdnrareas";
    int k = 1;
    cell c;
    if (gd->datatype == INTEGER_GRID) {
	grid *done = NULL;
	if (!(done = gdnew2(INTEGER_GRID, gd))) return 0;
	FOREACH(c, gd) {
	    if (IGD_CELL(gd, c) AND IGD_DATACELL(gd, c) AND !IGD_CELL(done, c)) {
		k++;
		if (k > INTEGER_MAX) {
		    fprintf(stderr,"%s: too many areas\n",fct);
		    gddestroy(done);
		    return 0;
		}
		debug(0,"mark area from %i,%i to %i\n", c.i, c.j, k);
		if (!igdfloodfill(gd, done, c, k, connectivity)) {
		    gddestroy(done);
		    return 0;
		}
	    }
	}
	gddestroy(done);
    } else {
	fprintf(stderr,"%s: not an integer grid\n", fct);
	return 0;
    }
    return 1;
}


grid *gdclip(grid *gd, int i1, int j1, int i2, int j2)
{
    static char *fct = "gdclip";
    grid *g;
    int i, j;
    if (i1 < 0 OR i2 >= gd->M OR i2 < i1 OR j1 < 0 OR j2 >= gd->N OR j2 < j1) {
	fprintf(stderr,"%s: can't clip %i,%i,%i,%i from a %i,%i grid\n",
		fct,i1,i2,j1,j2,gd->M,gd->N);
	return 0;
    }
    if (!(g = gdnew(gd->datatype, i2-i1+1, j2-j1+1))) return NULL;
    if (gd->datatype == REAL_GRID) {
	for (i=i1; i<=i2; i++)
	    for (j=j1; j<=j2; j++)
		RGD_AT(g,i-i1,j-j1) = RGD_AT(gd,i,j);
    } else if (gd->datatype == INTEGER_GRID) {
	for (i=i1; i<=i2; i++)
	    for (j=j1; j<=j2; j++)
		IGD_AT(g,i-i1,j-j1) = IGD_AT(gd,i,j);
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	gddestroy(g);
	return NULL;
    }
    g->unitdist = gd->unitdist;
    g->xy.minX = gd->xy.minX+gd->unitdist*j1;
    g->xy.maxX = gd->xy.minX+gd->unitdist*(j2+1);
    g->xy.minY = gd->xy.minY+gd->unitdist*(gd->M-1-i2);
    g->xy.maxY = gd->xy.minY+gd->unitdist*(gd->M-i1);
    return g;
}


grid *gdjoin(grid *g1, grid *g2)
{
    static char *fct = "gdjoin";
    grid *g;
    cell c;
    int M,N,mode;
    if (g1->unitdist != g2->unitdist) {
	fprintf(stderr,"%s: the unit distances are different (%f!=%f)\n",
		fct,g1->unitdist,g2->unitdist);
	return NULL;
    }
    if (!((g1->datatype == INTEGER_GRID OR g1->datatype == REAL_GRID) AND 
	  (g2->datatype == INTEGER_GRID OR g2->datatype == REAL_GRID))) {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return NULL;
    }
    if (g1->xy.minX == g2->xy.minX AND g1->xy.maxX == g2->xy.maxX) {
	N = g1->N;
	M = g1->M+g2->M;
	if (g1->xy.maxY == g2->xy.minY) {
	    mode = 1; /* g2 on the top of g1 */
	} else if (g1->xy.minY == g2->xy.maxY) {
	    mode = 2; /* g2 below g1 */
	} else {
	    fprintf(stderr,"%s: g1->xy.minY != g1->xy.maxY or g1->xy.maxY != g1->xy.minY\n",fct);
	    return NULL;
	}
    } else if (g1->xy.minY == g2->xy.minY AND g1->xy.maxY == g2->xy.maxY) {
	M = g1->M;
	N = g1->N+g2->N;
	if (g2->xy.maxX == g1->xy.minX) {      
	    mode = 3; /* g2 on the left of g1 */
	} else if (g2->xy.minX == g1->xy.maxX) {
	    mode = 4; /* g2 on the right of g1 */
	} else {
	    fprintf(stderr,"%s: g1->xy.minX != g1->xy.maxX or g1->xy.maxX != g1->xy.minX\n",fct);
	    return NULL;
	}
    } else {
	fprintf(stderr,"%s: the grids are not aligned\n",fct);
	return NULL;
    }
    if (!(g = gdnew(max(g1->datatype,g2->datatype),M,N))) return NULL;
    g->unitdist = g1->unitdist;
    g->xy.minX = min(g1->xy.minX,g2->xy.minX);
    g->xy.maxX = max(g1->xy.maxX,g2->xy.maxX);
    g->xy.minY = min(g1->xy.minY,g2->xy.minY);
    g->xy.maxY = max(g1->xy.maxY,g2->xy.maxY);
    if (g->datatype == INTEGER_GRID) {
	if (mode == 1) {
	    if (g2->datatype == REAL_GRID) {
		FOREACH(c, g2)
		    IGD_CELL(g, c) = RGD_CELL(g2, c);
	    } else if (g2->datatype == INTEGER_GRID) {
		FOREACH(c, g2)
		    IGD_CELL(g, c) = IGD_CELL(g2, c);
	    }
	    if (g1->datatype == REAL_GRID) {
		FOREACH(c, g1)
		    IGD_AT(g,g2->M+c.i, c.j) = RGD_CELL(g1, c);
	    } else if (g1->datatype == INTEGER_GRID) {
		FOREACH(c, g1)
		    IGD_AT(g,g2->M+c.i, c.j) = IGD_CELL(g1, c);
	    }
	} else if (mode == 2) {
	    if (g1->datatype == REAL_GRID) {
		FOREACH(c, g1)
		    IGD_CELL(g, c) = RGD_CELL(g1, c);
	    } else if (g1->datatype == INTEGER_GRID) {
		FOREACH(c, g1)
		    IGD_CELL(g, c) = IGD_CELL(g1, c);
	    }
	    if (g2->datatype == REAL_GRID) {
		FOREACH(c, g2)
		    IGD_AT(g, g1->M+c.i, c.j) = RGD_CELL(g2, c);
	    } else if (g2->datatype == INTEGER_GRID) {
		FOREACH(c, g2)
		    IGD_AT(g, g1->M+c.i, c.j) = IGD_CELL(g2, c);
	    }
	} else if (mode == 3) {
	    if (g2->datatype == REAL_GRID) {
		FOREACH(c, g2)
		    IGD_CELL(g, c) = RGD_CELL(g2, c);
	    } else if (g2->datatype == INTEGER_GRID) {
		FOREACH(c, g2)
		    IGD_CELL(g, c) = IGD_CELL(g2, c);
	    }
	    if (g1->datatype == REAL_GRID) {
		FOREACH(c, g1)
		    IGD_AT(g, c.i, g2->N+c.j) = RGD_CELL(g1, c);
	    } else if (g1->datatype == INTEGER_GRID) {
		FOREACH(c, g1)
		    IGD_AT(g, c.i, g2->N+c.j) = IGD_CELL(g1, c);
	    }
	} else if (mode == 4) {
	    if (g1->datatype == REAL_GRID) {
		FOREACH(c, g1)
		    IGD_CELL(g, c) = RGD_CELL(g1, c);
	    } else if (g1->datatype == INTEGER_GRID) {
		FOREACH(c, g1)
		    IGD_CELL(g, c) = IGD_CELL(g1, c);
	    }
	    if (g2->datatype == REAL_GRID) {
		FOREACH(c, g2)
		    IGD_AT(g, c.i, g1->N+c.j) = RGD_CELL(g2, c);
	    } else if (g2->datatype == INTEGER_GRID) {
		FOREACH(c, g2)
		    IGD_AT(g, c.i, g1->N+c.j) = IGD_CELL(g2, c);
	    }
	}
    } else if (g->datatype == REAL_GRID) {
	if (mode == 1) {
	    if (g2->datatype == REAL_GRID) {
		FOREACH(c, g2)
		    RGD_CELL(g, c) = RGD_CELL(g2, c);
	    } else if (g2->datatype == INTEGER_GRID) {
		FOREACH(c, g2)
		    RGD_CELL(g, c) = IGD_CELL(g2, c);
	    }
	    if (g1->datatype == REAL_GRID) {
		FOREACH(c, g2)
		    RGD_AT(g,g2->M+c.i, c.j) = RGD_CELL(g1, c);
	    } else if (g1->datatype == INTEGER_GRID) {
		FOREACH(c, g1)
		    RGD_AT(g,g2->M+c.i, c.j) = IGD_CELL(g1, c);
	    }
	} else if (mode == 2) {
	    if (g1->datatype == REAL_GRID) {
		FOREACH(c, g1)
		    RGD_CELL(g, c) = RGD_CELL(g1, c);
	    } else if (g1->datatype == INTEGER_GRID) {
		FOREACH(c, g1)
		    RGD_CELL(g, c) = IGD_CELL(g1, c);
	    }
	    if (g2->datatype == REAL_GRID) {
		FOREACH(c, g2)
		    RGD_AT(g,g1->M+c.i, c.j) = RGD_CELL(g2, c);
	    } else if (g2->datatype == INTEGER_GRID) {
		FOREACH(c, g2)
		    RGD_AT(g,g1->M+c.i, c.j) = IGD_CELL(g2, c);
	    }
	} else if (mode == 3) {
	    if (g2->datatype == REAL_GRID) {
		FOREACH(c, g2)
		    RGD_CELL(g, c) = RGD_CELL(g2, c);
	    } else if (g2->datatype == INTEGER_GRID) {
		FOREACH(c, g2)
		    RGD_CELL(g, c) = IGD_CELL(g2, c);
	    }
	    if (g1->datatype == REAL_GRID) {
		FOREACH(c, g1)
		    RGD_AT(g,c.i,g2->N+c.j) = RGD_CELL(g1, c);
	    } else if (g1->datatype == INTEGER_GRID) {
		FOREACH(c, g1)
		    RGD_AT(g,c.i,g2->N+c.j) = IGD_CELL(g1, c);
	    }
	} else if (mode == 4) {
	    if (g1->datatype == REAL_GRID) {
		FOREACH(c, g1)
		    RGD_CELL(g, c) = RGD_CELL(g1, c);
	    } else if (g1->datatype == INTEGER_GRID) {
		FOREACH(c, g1)
		    RGD_CELL(g, c) = IGD_CELL(g1, c);
	    }
	    if (g2->datatype == REAL_GRID) {
		FOREACH(c, g2)
		    RGD_AT(g,c.i,g1->N+c.j) = RGD_CELL(g2, c);
	    } else if (g2->datatype == INTEGER_GRID) {
		FOREACH(c, g2)
		    RGD_AT(g,c.i,g1->N+c.j) = IGD_CELL(g2, c);
	    }
	}
    }
    return g;
}


grid *gdtransform(grid *gd, double tr[], int M, int N) {
    /* i0=ai + bi * i + ci * j				
       j0=aj + bj * i + cj * j
       tr = (ai,bi,ci,aj,bj,cj) */
    static char *fct = "gdtransform";
    grid *g;
    cell c;
    int i0,j0;
    if (!(gd->datatype == INTEGER_GRID OR gd->datatype == REAL_GRID)) {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return NULL;
    }
    if (!(g = gdnew(gd->datatype, M, N))) return NULL;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, g) {
	    i0 = round(tr[0]+tr[1]*c.i+tr[2]*c.j);
	    j0 = round(tr[3]+tr[4]*c.i+tr[5]*c.j);
	    if (i0 >= 0 AND i0 < gd->M AND j0 >= 0 AND j0 < gd->N)
		IGD_CELL(g, c) = IGD_AT(gd,i0,j0);
	}
    } else {
	FOREACH(c, g) {
	    i0 = round(tr[0]+tr[1]*c.i+tr[2]*c.j);
	    j0 = round(tr[3]+tr[4]*c.i+tr[5]*c.j);
	    if (i0 >= 0 AND i0 < gd->M AND j0 >= 0 AND j0 < gd->N)
		RGD_CELL(g, c) = RGD_AT(gd,i0,j0);
	}
    }
    return g;
}


int parse_int(char **s, int *i) {
    char *p = *s, t;
    int r;
    while (*p AND strchr("+-0123456789",*p)) p++;
    t = *p;
    *p = '\0';
    r = sscanf(*s, "%i", i);
    if (r != 1) fprintf(stderr,"parse_int: WARNING: %s: not an integer\n", *s);
    *p = t;
    *s = p;
    return r;
}


int parse_float(char **s, float *d) {
    char *p = *s, t;
    int r;
    while (*p AND strchr("+-0123456789Ee.",*p)) p++;
    t = *p;
    *p = '\0';
    r = sscanf(*s, "%f", d);
    if (r != 1) fprintf(stderr,"parse_float: WARNING: %s: not a floating point number\n", *s);
    *p = t;
    *s = p;
    return r;
}


grid *a2gd(int datatype, int M, int N, char *infile, int mode)
{
    const char *fct = "a2gd";
    grid *gd = NULL;
    int i = 0, j = 0, lnr = 0, ok = 0;
    char *ln;
    FILE *f = fopen(infile,"r");
    if (!f) {
	fprintf(stderr,"%s: %s: %s\n", fct, infile, strerror(errno)); 
	return NULL;
    }
    if (!(gd = gdnew(datatype, M, N))) {
	fclose(f);
	return NULL;
    }
    if (mode == 1) {
	/* skip the header, leave it for perl */
	while (lnr < 6) {
	    ln = getln(f);
	    if (ln) 
		free(ln);
	    else
		break;
	    lnr++;
	}
	ok = lnr == 6;
    } else if (mode == 2) {
	/* look for GRD part */
	while ((ln = getln(f))) {
	    if (strncmp(ln, "GRD", 3) == 0) {
		free(ln);
		break;
	    }
	    free(ln);
	}
	/* skip the header, leave it for perl */
	while (lnr < 4) {
	    ln = getln(f);
	    if (ln) 
		free(ln);
	    else
		break;
	    lnr++;
	}
	ok = lnr == 4;
    } else {
	fprintf(stderr,"%s: %i unknown import mode\n", fct, mode);
    }
    if (ok) {
	while ((ln = getln(f))) {
	    char *s = btrim(ln);
	    lnr++;
	    if ((mode == 2) AND (strncmp(ln, "EOG", 3) == 0)) {
		free(ln);
		break;
	    }
	    if (datatype == INTEGER_GRID AND strchr("+-0123456789",*s)) {
		if (i == M) {
		    fprintf(stderr,"%s: WARNING: %s:%i: too much data\n", 
			    fct, infile, lnr);
		    free(ln);
		    break;
		}
		while (*s) {
		    char *tok = s;
		    int x;
		    parse_int(&s, &x);
		    debug(2,"%i,%i: %i tok=%s\n",i,j,x,tok);
		    if (x == -2147483647)
			x = INTEGER_MIN;
		    gdset2(gd, i, j, x);
		    GDINCINDEX(i,j,N);
		    s = btrim(s);
		}
	    } else if (datatype == REAL_GRID AND strchr("+-0123456789Ee.",*s)) {
		if (i == M) {
		    fprintf(stderr,"%s: WARNING: %s:%i: too much data\n", 
			    fct, infile, lnr);
		    free(ln);
		    break;
		}
		while (*s) {
		    char *tok = s;
		    float x;
		    parse_float(&s, &x);
		    debug(2,"%i,%i: %f tok=%s\n",i,j,x,tok);
		    gdset(gd, i, j, x);
		    GDINCINDEX(i,j,N);
		    s = btrim(s);
		}
	    } else 
		fprintf(stderr,"%s: WARNING: %s:%i: skipping line:%s", 
			fct, infile, lnr, ln);
	    free(ln);
	    SIGCANCEL;
	}
    }
    fclose(f);
    if (i < M) 
	fprintf(stderr,"%s: WARNING: imported data only from %i/%i lines\n", 
		fct, i, M);
    return gd;
 cancel:
    gddestroy(gd);
    return NULL;
}


int gd2a(grid *gd, char *outfile)
{
    const char *fct = "gd2a";
    int i = 0, j = 0;
    FILE *f = fopen(outfile,"w");
    if (!f) {
	fprintf(stderr,"%s: %s: %s\n", fct, outfile, strerror(errno)); 
	return 0;
    }
    /* the header */
    fprintf(f,"ncols         %i\n",gd->N);
    fprintf(f,"nrows         %i\n",gd->M);
    fprintf(f,"xllcorner     %f\n",gd->xy.minX);
    fprintf(f,"yllcorner     %f\n",gd->xy.minY);
    fprintf(f,"cellsize      %f\n",gd->unitdist);

    if (gd->datatype == INTEGER_GRID) {
	fprintf(f,"NODATA_value  %i\n",gd->snodata);
	for (i = 0; i < gd->M; i++) {
	    for (j = 0; j < gd->N; j++) {
		fprintf(f,"%i ",IGD_AT(gd, i, j));
	    }
	    fprintf(f,"\n");
	}
    } else if (gd->datatype == REAL_GRID) {
	fprintf(f,"NODATA_value  %f\n",gd->fnodata);
	for (i = 0; i < gd->M; i++) {
	    for (j = 0; j < gd->N; j++) {
		fprintf(f,"%f ",RGD_AT(gd, i, j));
	    }
	    fprintf(f,"\n");
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
    }

    fclose(f);
    return 1;
}

/* Bresenham as presented in Foley & Van Dam */

int gdline(grid *gd, int i1, int j1, int i2, int j2, double pen)
{
    int di, dj, incr1, incr2, d, i, j, iend, jend, idirflag, jdirflag;
    di = abs(i2-i1);
    dj = abs(j2-j1);
    if (dj <= di) {
	d = 2*dj - di;
	incr1 = 2*dj;
	incr2 = 2 * (dj - di);
	if (i1 > i2) {
	    i = i2;
	    j = j2;
	    jdirflag = (-1);
	    iend = i1;
	} else {
	    i = i1;
	    j = j1;
	    jdirflag = 1;
	    iend = i2;
	}
	gdset(gd, i, j, pen);
	if (((j2 - j1) * jdirflag) > 0) {
	    while (i < iend) {
		i++;
		if (d <0) {
		    d+=incr1;
		} else {
		    j++;
		    d+=incr2;
		}
		gdset(gd, i, j, pen);
	    }
	} else {
	    while (i < iend) {
		i++;
		if (d <0) {
		    d+=incr1;
		} else {
		    j--;
		    d+=incr2;
		}
		gdset(gd, i, j, pen);
	    }
	}		
    } else {
	d = 2*di - dj;
	incr1 = 2*di;
	incr2 = 2 * (di - dj);
	if (j1 > j2) {
	    j = j2;
	    i = i2;
	    jend = j1;
	    idirflag = (-1);
	} else {
	    j = j1;
	    i = i1;
	    jend = j2;
	    idirflag = 1;
	}
	gdset(gd, i, j, pen);
	if (((i2 - i1) * idirflag) > 0) {
	    while (j < jend) {
		j++;
		if (d <0) {
		    d+=incr1;
		} else {
		    i++;
		    d+=incr2;
		}
		gdset(gd, i, j, pen);
	    }
	} else {
	    while (j < jend) {
		j++;
		if (d <0) {
		    d+=incr1;
		} else {
		    i--;
		    d+=incr2;
		}
		gdset(gd, i, j, pen);
	    }
	}
    }
    return 1;
}


int gdfilledrect(grid *gd, int i1, int j1, int i2, int j2, double pen) 
{
    int i, j;
    if (i2 < i1) swap(i1,i2,i);
    if (j2 < j1) swap(j1,j2,j);
    for (i = max(0,i1); i < min(gd->M,i2+1); i++) {
	for (j = max(0,j1); j < min(gd->N,j2+1); j++) {
	    gdset(gd, i, j, pen);
	}
    }
    return 1;
}


void gdfilledcircle(grid *gd, int i, int j, int r, int r2, double pen)
{
    int a, b;
    int di, dj;
    di = max(-r, -i);
    a = r2 - di*di;
    while (1) {
	dj = max(-r, -j);
	b = dj*dj;
	while (1) {
	    if (b < a) gdset(gd, i+di, j+dj, pen);
	    dj++;
	    if (dj > r OR j+dj >= gd->N) break;
	    b += 2*dj - 1;
	}
	di++;
	if (di > r OR i+di >= gd->M) break;
	a -= 2*di - 1;
    }
}


int gdfloodfill(grid *gd, cell c, int icolor, double rcolor, int connectivity)
{
  const char *fct = "gdfloodfill";
  if (!(gd->datatype == INTEGER_GRID OR gd->datatype == REAL_GRID)) {
      fprintf(stderr,"%s: grid datatype error\n", fct);
      return 0;
  }
  if (gd->datatype == INTEGER_GRID) {
      return igdfloodfill(gd, NULL, c, icolor, connectivity);
  } else {
      return rgdfloodfill(gd, NULL, c, rcolor, connectivity);
  }
  return 0;
}


/* after gdImageFill in gd.c of http://www.boutell.com/gd/ */
int igdfloodfill(grid *gd, grid *done, cell c, int color, int connectivity)
{
    int lastBorder;
    int old;
    int leftLimit, rightLimit;
    int i2;
    old = IGD_CELL(gd, c);
    if (old == color) {
	/* Nothing to be done */
        if (done) IGD_CELL(done, c) = 1;
	return 1;
    }
    /* Seek left */
    leftLimit = (-1);
    for (i2 = c.i; (i2 >= 0); i2--) {
	if (IGD_AT(gd, i2, c.j) != old) {
	    break;
	}
	IGD_AT(gd, i2, c.j) = color;
	if (done) IGD_AT(done, i2, c.j) = 1;
	if (connectivity == 8)
	    leftLimit = max(0,i2-1);
	else
	    leftLimit = i2;
    }
    if (leftLimit == (-1)) {
	return 1;
    }
    /* Seek right */
    if (connectivity == 8)
	rightLimit = min(gd->M-1,c.i+1);
    else
	rightLimit = c.i;
    for (i2 = (c.i+1); (i2 < gd->M); i2++) {	
	if (IGD_AT(gd, i2, c.j) != old) {
	    break;
	}
	IGD_AT(gd, i2, c.j) = color;
	if (done) IGD_AT(done, i2, c.j) = 1;
	if (connectivity == 8)
	    rightLimit = min(gd->M-1,i2+1);
	else
	    rightLimit = i2;
    }
    /* Look at lines above and below and start paints */
    /* Above */
    if (c.j > 0) {
	cell c2;
	c2.j = c.j-1;
	lastBorder = 1;
	for (c2.i = leftLimit; (c2.i <= rightLimit); c2.i++) {
	    int a;
	    a = IGD_CELL(gd, c2);
	    if (lastBorder) {
		if (a == old) {	
		    igdfloodfill(gd, done, c2, color, connectivity);
		    lastBorder = 0;
		}
	    } else if (a != old) {
		lastBorder = 1;
	    }
	}
    }
    /* Below */
    if (c.j < ((gd->N) - 1)) {
	cell c2;
	c2.j = c.j+1;
	lastBorder = 1;
	for (c2.i = leftLimit; (c2.i <= rightLimit); c2.i++) {
	    int a;
	    a = IGD_CELL(gd, c2);
	    if (lastBorder) {
		if (a == old) {
		    igdfloodfill(gd, done, c2, color, connectivity);
		    lastBorder = 0;
		}
	    } else if (a != old) {
		lastBorder = 1;
	    }
	}
    }
    return 1;
}


int rgdfloodfill(grid *gd, grid *done, cell c, double color, int connectivity)
{
    int lastBorder;
    double old;
    int leftLimit, rightLimit;
    int i2;
    old = RGD_CELL(gd, c);
    if (old == color) {
	/* Nothing to be done */
        if (done) IGD_CELL(done, c) = 1;
	return 1;
    }
    /* Seek left */
    leftLimit = (-1);
    for (i2 = c.i; (i2 >= 0); i2--) {
	if (RGD_AT(gd, i2, c.j) != old) {
	    break;
	}
	RGD_AT(gd, i2, c.j) = color;
	if (done) IGD_AT(done, i2, c.j) = 1;
	if (connectivity == 8)
	    leftLimit = max(0,i2-1);
	else
	    leftLimit = i2;
    }
    if (leftLimit == (-1)) {
	return 1;
    }
    /* Seek right */
    if (connectivity == 8)
	rightLimit = min(gd->M-1,c.i+1);
    else
	rightLimit = c.i;
    for (i2 = (c.i+1); (i2 < gd->M); i2++) {	
	if (RGD_AT(gd, i2, c.j) != old) {
	    break;
	}
	RGD_AT(gd, i2, c.j) = color;
	if (done) IGD_AT(done, i2, c.j) = 1;
	if (connectivity == 8)
	    rightLimit = min(gd->M-1,i2+1);
	else
	    rightLimit = i2;
    }
    /* Look at lines above and below and start paints */
    /* Above */
    if (c.j > 0) {
	cell c2;
	c2.j = c.j-1;
	lastBorder = 1;
	for (c2.i = leftLimit; (c2.i <= rightLimit); c2.i++) {
	    double a;
	    a = RGD_CELL(gd, c2);
	    if (lastBorder) {
		if (a == old) {	
		    igdfloodfill(gd, done, c2, color, connectivity);
		    lastBorder = 0;
		}
	    } else if (a != old) {
		lastBorder = 1;
	    }
	}
    }
    /* Below */
    if (c.j < ((gd->N) - 1)) {
	cell c2;
	c2.j = c.j+1;
	lastBorder = 1;
	for (c2.i = leftLimit; (c2.i <= rightLimit); c2.i++) {
	    double a;
	    a = RGD_CELL(gd, c2);
	    if (lastBorder) {
		if (a == old) {
		    igdfloodfill(gd, done, c2, color, connectivity);
		    lastBorder = 0;
		}
	    } else if (a != old) {
		lastBorder = 1;
	    }
	}
    }
    return 1;
}


int gdprint(grid *gd)
{
    static char *fct = "gdprint";
    int i, j;
    if (gd->datatype == INTEGER_GRID) {
	for (i = 0; i < gd->M; i++) {
	    for (j = 0; j < gd->N; j++) {
		SIGCANCEL;
		printf("%i ",IGD_AT(gd, i, j));
	    }
	    printf("\n");
	}
    } else if (gd->datatype == REAL_GRID) {
	for (i = 0; i < gd->M; i++) {
	    for (j = 0; j < gd->N; j++) {
		SIGCANCEL;
		printf("%f ",RGD_AT(gd, i, j));
	    }
	    printf("\n");
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
 cancel:
    return 1;
}


int gdprint1(grid *gd, double **p, int *psize, int quiet, int wc)
{
    static char *fct = "gdprint1";
    cell c;
    *psize = 0;
    *p = NULL;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd) {
	    if (IGD_DATACELL(gd, c) AND IGD_CELL(gd, c)) {
		SIGCANCEL;
		if (!quiet) {
		    if (wc) 
			printf("%f,%f,%i\n", gdj2x(gd,c.j), gdi2y(gd,c.i), IGD_CELL(gd, c));
		    else
			printf("%i,%i,%i\n", c.i, c.j, IGD_CELL(gd, c));
		}
		(*psize)++;
	    }
	}
	*psize *= 3;
	if (!(*p = (double *)calloc(*psize,sizeof(double)))) {
	    fprintf(stderr,"%s: Out of memory!\n", fct);
	    return 0;
	}
	*psize = 0;
	FOREACH(c, gd) {
	    SIGCANCEL;
	    if (IGD_DATACELL(gd, c) AND IGD_CELL(gd, c)) {
		if (wc) {
		    (*p)[3*(*psize)] = gdj2x(gd,c.j);
		    (*p)[3*(*psize)+1] = gdi2y(gd,c.i);
		} else {
		    (*p)[3*(*psize)] = c.i;
		    (*p)[3*(*psize)+1] = c.j;
		}
		(*p)[3*(*psize)+2] = IGD_CELL(gd, c);
		(*psize)++;
	    }
	}
	*psize *= 3;
    } else if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd) {
	    SIGCANCEL;
	    if (RGD_DATACELL(gd, c) AND fabs(RGD_CELL(gd, c)) > 0) {
		if (!quiet) {
		    if (wc) 
			printf("%f,%f,%f\n", gdj2x(gd,c.j), gdi2y(gd,c.i), RGD_CELL(gd, c));
		    else
			printf("%i,%i,%f\n", c.i, c.j, RGD_CELL(gd, c));
		}
		(*psize)++;
	    }
	}
	*psize *= 3;
	if (!(*p = (double *)calloc(*psize,sizeof(double)))) {
	    fprintf(stderr,"%s: Out of memory!\n", fct);
	    return 0;
	}
	*psize = 0;
	FOREACH(c, gd) {
	    SIGCANCEL;
	    if (RGD_DATACELL(gd, c) AND fabs(RGD_CELL(gd, c)) > 0) {
		if (wc) {
		    (*p)[3*(*psize)] = gdj2x(gd,c.j);
		    (*p)[3*(*psize)+1] = gdi2y(gd,c.i);
		} else {
		    (*p)[3*(*psize)] = c.i;
		    (*p)[3*(*psize)+1] = c.j;
		}
		(*p)[3*(*psize)+2] = RGD_CELL(gd, c);
		(*psize)++;
	    }
	}
	*psize *= 3;
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
 cancel:
    if (*p) free(*p);
    *psize = 0;
    return 0;
}


int gdhistogram(grid *gd, double *h, int *c, int n)
{
    static char *fct = "gdhistogram";
    cell p;
    int k;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(p, gd) {
	    if (IGD_DATACELL(gd, p)) {
		k = 0;
		while (IGD_CELL(gd, p) > h[k] AND k < n) k++;
		c[k]++;
	    }
	}
    } else if (gd->datatype == REAL_GRID) {
	FOREACH(p, gd) {
	    if (RGD_DATACELL(gd, p)) {
		k = 0;
		while (RGD_CELL(gd, p) > h[k] AND k < n) k++;
		c[k]++;
	    }
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return 1;
}


int gdcontents(grid *gd, hash *table) 
{
    static char *fct = "gdcontents";
    cell p;
    int ret = 0;
    if (gd->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: not an integer grid\n",fct);
	return 0;
    }
    FOREACH(p, gd) {
	int *count;
	if (IGD_NODATACELL(gd,p)) continue;
	if (hash_int_lookup(table, IGD_CELL(gd, p), &count))
	    (*count)++;
	else {
	    if (!hash_int_insert(table, IGD_CELL(gd, p), 1)) goto fail;
	}
    }
    ret = 1;
 fail:
    return ret;
}


int gdzonalcount(grid *gd, grid *zones, hash *table)
{
    static char *fct = "gdzonalcount";
    cell p;
    int ret = 0;
    if (zones->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: zones grid is not an integer grid\n",fct);
	return 0;
    }
    FOREACH(p, gd) {
	int *count;
	if (IGD_NODATACELL(zones,p) OR GD_NODATACELL(gd,p)) continue;
	if (hash_int_lookup(table, IGD_CELL(zones, p), &count))
	    (*count)++;
	else {
	    if (!hash_int_insert(table, IGD_CELL(zones, p), 1)) goto fail;
	}
    }
    ret = 1;
 fail:
    return ret;
}


int gdzonalsum(grid *gd, grid *zones, hash *table)
{
    static char *fct = "gdzonalsum";
    cell p;
    int ret = 0;
    if (zones->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: zones grid is not an integer grid\n",fct);
	return 0;
    }
    FOREACH(p, gd) {
	double *sum;
	if (IGD_NODATACELL(zones,p) OR GD_NODATACELL(gd,p)) continue;
	if (hash_double_lookup(table, IGD_CELL(zones, p), &sum))
	    *sum += GD_CELL(gd, p);
	else {
	    if (!hash_double_insert(table, IGD_CELL(zones, p), GD_CELL(gd, p))) goto fail;
	}
    }
    ret = 1;
 fail:
    return ret;
}


int gdzonalmin(grid *gd, grid *zones, hash *table)
{
    static char *fct = "gdzonalmin";
    cell p;
    int ret = 0;
    if (zones->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: zones grid is not an integer grid\n",fct);
	return 0;
    }
    FOREACH(p, gd) {
	double *min;
	if (IGD_NODATACELL(zones,p) OR GD_NODATACELL(gd,p)) continue;
	if (hash_double_lookup(table, IGD_CELL(zones, p), &min)) {
	    if (GD_CELL(gd, p) < *min) *min = GD_CELL(gd, p);
	} else {
	    if (!hash_double_insert(table, IGD_CELL(zones, p), GD_CELL(gd, p))) goto fail;
	}
    }
    ret = 1;
 fail:
    return ret;
}


int gdzonalmax(grid *gd, grid *zones, hash *table)
{
    static char *fct = "gdzonalmax";
    cell p;
    int ret = 0;
    if (zones->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: zones grid is not an integer grid\n",fct);
	return 0;
    }
    FOREACH(p, gd) {
	double *max;
	if (IGD_NODATACELL(zones,p) OR GD_NODATACELL(gd,p)) continue;
	if (hash_double_lookup(table, IGD_CELL(zones, p), &max)) {
	    if (GD_CELL(gd, p) > *max) *max = GD_CELL(gd, p);
	} else {
	    if (!hash_double_insert(table, IGD_CELL(zones, p), GD_CELL(gd, p))) goto fail;
	}
    }
    ret = 1;
 fail:
    return ret;
}


int gdzonalmean(grid *gd, grid *zones, hash *table)
{
    static char *fct = "gdzonalmean";
    cell p;
    hash count;
    int ret = 0;
    if (zones->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: zones grid is not an integer grid\n",fct);
	return 0;
    }
    if (!hash_create(&count, 200)) return 0;
    FOREACH(p, gd) {
	double *mean;
	int *n;
	if (IGD_NODATACELL(zones,p) OR GD_NODATACELL(gd,p)) continue;
	if (hash_double_lookup(table, IGD_CELL(zones, p), &mean)) {
	    hash_int_lookup(table, IGD_CELL(zones, p), &n);
	    (*n)++;
	    *mean += (GD_CELL(gd, p) - *mean) / *n;
	} else {
	    if (!hash_double_insert(table, IGD_CELL(zones, p), GD_CELL(gd, p))) goto fail;
	    if (!hash_int_insert(&count, IGD_CELL(zones, p), 1)) goto fail;
	}
    }
    ret = 1;
 fail:
    hash_destroy(&count);
    return ret;
}


int gdzonalvariance(grid *gd, grid *zones, hash *table)
{
    static char *fct = "gdzonalvariance";
    cell p;
    hash count, mean;
    int i, ret = 0;
    if (zones->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: zones grid is not an integer grid\n",fct);
	return 0;
    }
    if (!hash_create(&count, 200)) return 0;
    if (!hash_create(&mean, 200)) goto fail;
    FOREACH(p, gd) {
	double oldmean;
	int *n;
	double *m, *variance;
	if (IGD_NODATACELL(zones,p) OR GD_NODATACELL(gd,p)) continue;
	if (hash_double_lookup(table, IGD_CELL(zones, p), &variance)) {
	    hash_double_lookup(&mean, IGD_CELL(zones, p), &m);
	    hash_int_lookup(&count, IGD_CELL(zones, p), &n);
	    (*n)++;
	    oldmean = *m;
	    *m += (GD_CELL(gd, p) - oldmean) / *n;
	    *variance += (GD_CELL(gd, p) - oldmean) * (GD_CELL(gd, p) - *m);
	} else {
	    if (!hash_double_insert(table, IGD_CELL(zones, p), 0)) goto fail;
	    if (!hash_double_insert(&mean, IGD_CELL(zones, p), GD_CELL(gd, p))) goto fail;
	    if (!hash_int_insert(&count, IGD_CELL(zones, p), 1)) goto fail;
	}
    }
    for (i = 0; i < table->size; i++) {
	hash_double_item *a = (hash_double_item *)table->table[i];
	while (a) {
	    int *n;
	    hash_int_lookup(&count, a->key, &n);
	    if (*n > 1) a->value /= (*n - 1);
	    a = a->next;
	}
    }
    ret = 1;
 fail:
    hash_destroy(&count);
    hash_destroy(&mean);
    return ret;
}


void rgrow(grid *zones, grid *grow, cell c, int connectivity)
{
    if (connectivity == 8) {
	int dir;
	for (dir = 1; dir < 9; dir++) {
	    cell t = cellmove(c, dir);
	    if (GD_CELLIN(zones, t) AND 
		IGD_CELL(grow, t) AND 
		IGD_CELL(zones, t) != IGD_CELL(zones, c)) {
		IGD_CELL(zones, t) = IGD_CELL(zones, c);
		rgrow(zones, grow, t, connectivity);
	    }
	}
    } else {
	int dir;
	for (dir = 1; dir < 9; dir += 2) {
	    cell t = cellmove(c, dir);
	    if (GD_CELLIN(zones, t) AND 
		IGD_CELL(grow, t) AND
		IGD_CELL(zones, t) != IGD_CELL(zones, c)) {
		IGD_CELL(zones, t) = IGD_CELL(zones, c);
		rgrow(zones, grow, t, connectivity);
	    }
	}
    }
}


int gdgrowzones(grid *zones, grid *grow, int connectivity)
{
    static char *fct = "gdgrowzones";
    cell c;
    if (zones->datatype != INTEGER_GRID OR grow->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: the grids have to be integer grids\n",fct);
	return 0;
    }
    if (!gdoverlayable(zones, grow)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    FOREACH(c, zones) {
	SIGCANCEL;
	if (IGD_NODATACELL(zones, c)) continue;
	rgrow(zones, grow, c, connectivity);
    }
 cancel:
    return 1;
}

int resize_c(int **c, int oldsize, int newsize)
{
    int i,*tmp;
    if (*c == NULL) {
	if (!((*c) = (int *)calloc(newsize,sizeof(int)))) {
	    fprintf(stderr,"Out of memory!\n");
	    return 0;
	}
    } else {
	if (!(tmp = (int *)realloc(*c,newsize*sizeof(int)))) {
	    fprintf(stderr,"Out of memory!\n");
	    free(*c);
	    return 0;
	}
	*c = tmp;
	for (i=oldsize;i<newsize;i++) (*c)[i] = 0;
    }
    return 1;
}


int gdneighbors(grid *gd, hash ***b, int **c, int *n)
{
    static char *fct = "gdneighbors";
    hash contents;
    cell d;
    int i, j;
    if (gd->datatype != INTEGER_GRID) {
	fprintf(stderr,"%s: not an integer grid\n", fct);
	return 0;
    }
    *b = NULL;
    *n = 0;
    if (!hash_create(&contents, 200)) goto fail;
    if (!gdcontents(gd, &contents)) goto fail;
    *n = hash_count(&contents);
    if (!(*b = (hash **)calloc(*n,sizeof(hash *)))) goto outofmem;
    if (!(*c = (int *)calloc(*n,sizeof(int *)))) goto outofmem;
    for (i = 0; i < *n; i++) {
	if (!((*b)[i] = (hash *)malloc(sizeof(hash)))) goto outofmem;
	if (!hash_create((*b)[i], 100)) goto fail;
    }
    j = 0;
    for (i = 0; i < contents.size; i++) {
	hash_int_item *a = (hash_int_item *)contents.table[i];
	while (a) {
	    (*c)[j] = a->key;
	    a->value = j;
	    j++;
	    a = a->next;
	}
    }
    FOREACH(d, gd) {
	int *i, z = IGD_CELL(gd, d), dir;
	if (!hash_int_lookup(&contents, z, &i)) continue;
	FOREACHDIR(dir) {
	    cell t = cellmove(d, dir);
	    if (GD_CELLIN(gd, t)) { /* MASKING!!! */
		int *na, a = IGD_CELL(gd, t);
		if (hash_int_lookup((*b)[*i], a, &na))
		    (*na)++;
		else {
		    if (!hash_int_insert((*b)[*i], a, 1)) goto fail;
		}
	    }
	}
    }
    hash_destroy(&contents);
    return 1;
 outofmem: 
    fprintf(stderr,"%s: Out of memory!\n", fct);
 fail:
    if (*b) {
	for (i = 0; i < *n; i++) {
	    if ((*b)[i]) {
		hash_destroy((*b)[i]);
		free((*b)[i]);
	    }
	}
	free(*b);
    }
    hash_destroy(&contents);
    return 0;
}


grid *gdbufferzone(grid *gd, int z, double w)
{
    static char *fct = "gdbufferzone";
    grid *d;
    cell c;
    int r2 = round(w*w);
    int r = round(w);
    PROGRESS_BAR_VARS(gd->M/40);
    if (!(d = gdnew2(INTEGER_GRID, gd))) return NULL;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd) {
	    PROGRESS_CHECK(c.i);
	    SIGCANCEL;
	    if (IGD_CELL(gd, c) == z)
		gdfilledcircle(d, c.i, c.j, r, r2, 1);
	}
    } else {
	fprintf(stderr,"%s: not an integer grid\n", fct);
	goto fail;
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    return d;
 fail:
    gddestroy(d);
    return NULL;
}


int gdcount(grid *gd)
{
    static char *fct = "gdcount";
    cell c;
    int count = 0;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c))
		count++;
    } else if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd)
	    if (RGD_DATACELL(gd, c))
		count++;
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return count;
}


double gdsum(grid *gd)
{
    static char *fct = "gdsum";
    cell c;
    double sum = 0;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c))
		sum += IGD_CELL(gd, c);
    } else if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd)
	    if (RGD_DATACELL(gd, c))
		sum += RGD_CELL(gd, c);
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return sum;
}


double gdmean(grid *gd)
{
    static char *fct = "gdmean";
    cell c;
    double mean = 0;
    int n = 0;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c)) {
		n++;
		mean += (IGD_CELL(gd, c) - mean) / n;
	    }
    } else if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd)
	    if (RGD_DATACELL(gd, c)) {
		n++;
		mean += (RGD_CELL(gd, c) - mean) / n;
	    }
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    return mean;
}


double gdvariance(grid *gd)
{
    static char *fct = "gdsum";
    cell c;
    double variance = 0, mean = 0;
    int n = 0;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd)
	    if (IGD_DATACELL(gd, c)) {
		double oldmean = mean;
		n++;
		mean += (IGD_CELL(gd, c) - oldmean) / n;
		variance += (IGD_CELL(gd, c) - oldmean) * (IGD_CELL(gd, c) - mean);
	    }
    } else if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd)
	    if (RGD_DATACELL(gd, c)) {
		double oldmean = mean;
		n++;
		mean += (RGD_CELL(gd, c) - oldmean) / n;
		variance += (RGD_CELL(gd, c) - oldmean) * (RGD_CELL(gd, c) - mean);
	    }
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return 0;
    }
    if (n > 1) variance /= (n - 1);
    return variance;
}


#define COMPARISON { if (IGD_CELL(gd, x) AND IGD_DATACELL(gd, x)) { \
                         double t = DISTANCE_BETWEEN_CELLS(c, x); \
			 if (d < 0 OR t < d) { \
                             d = t; \
                             ret = x; \
                         } \
                     } \
		   }

#define RCOMPARISON { if (fabs(RGD_CELL(gd, x)) > EPSILON AND RGD_DATACELL(gd, x)) { \
                         double t = DISTANCE_BETWEEN_CELLS(c, x); \
			 if (d < 0 OR t < d) { \
                             d = t; \
                             ret = x; \
                         } \
                     } \
		   }


cell gdnearest_neighbor(grid *gd, cell c)
{
    static char *fct = "gdnearest_neighbor";
    int r = 1;
    double d = -1;
    cell ret = {-1, -1};
    if (gd->datatype == INTEGER_GRID) {
	/* the first candidate */
	while (d < 0 AND r < min(gd->M, gd->N)) {
	    cell x;
	    int imin = c.i-r, jmin = c.j-r, imax = c.i+r+1, jmax = c.j+r+1;
	    int left = 1, right = 1, top = 1, bottom = 1;
	    if (imin < 0) {imin = 0;top = 0;}
	    if (jmin < 0) {jmin = -1;left = 0;}
	    if (imax > gd->M) {imax = gd->M;bottom = 0;}
	    if (jmax > gd->N) {jmax = gd->N+1;right = 0;}
	    if (left) {
		x.j = jmin;
		for (x.i = imin; x.i < imax; x.i++) COMPARISON;
	    }
	    if (right) {
		x.j = jmax-1;
		for (x.i = imin; x.i < imax; x.i++) COMPARISON;
	    }
	    if (top) {
		x.i = imin;
		for (x.j = jmin+1; x.j < jmax-1; x.j++) COMPARISON;
	    }
	    if (bottom) {
		x.i = imax - 1;
		for (x.j = jmin+1; x.j < jmax-1; x.j++) COMPARISON;
	    }
	    r++;
	}
	if (d < 0) {
	    fprintf(stderr,"%s: no non-zero cells in the grid\n", fct);
	    return ret;
	}
	/* we must now enlarge the rectangle until it is
	   larger than d and check */
	while (r < d) {
	    cell x;
	    int imin = c.i-r, jmin = c.j-r, imax = c.i+r+1, jmax = c.j+r+1;
	    int left = 1, right = 1, top = 1, bottom = 1;
	    if (imin < 0) {imin = 0;top = 0;}
	    if (jmin < 0) {jmin = -1;left = 0;}
	    if (imax > gd->M) {imax = gd->M;bottom = 0;}
	    if (jmax > gd->N) {jmax = gd->N+1;right = 0;}
	    if (left) {
		x.j = jmin;
		for (x.i = imin; x.i < imax; x.i++) COMPARISON;
	    }
	    if (right) {
		x.j = jmax-1;
		for (x.i = imin; x.i < imax; x.i++) COMPARISON;
	    }
	    if (top) {
		x.i = imin;
		for (x.j = jmin+1; x.j < jmax-1; x.j++) COMPARISON;
	    }
	    if (bottom) {
		x.i = imax - 1;
		for (x.j = jmin+1; x.j < jmax-1; x.j++) COMPARISON;
	    }
	    r++;
	}
    } else if (gd->datatype == REAL_GRID) {
	while (d < 0 AND r < min(gd->M, gd->N)) {
	    cell x;
	    int imin = c.i-r, jmin = c.j-r, imax = c.i+r+1, jmax = c.j+r+1;
	    int left = 1, right = 1, top = 1, bottom = 1;
	    if (imin < 0) {imin = 0;top = 0;}
	    if (jmin < 0) {jmin = -1;left = 0;}
	    if (imax > gd->M) {imax = gd->M;bottom = 0;}
	    if (jmax > gd->N) {jmax = gd->N+1;right = 0;}
	    if (left) {
		x.j = jmin;
		for (x.i = imin; x.i < imax; x.i++) RCOMPARISON;
	    }
	    if (right) {
		x.j = jmax-1;
		for (x.i = imin; x.i < imax; x.i++) RCOMPARISON;
	    }
	    if (top) {
		x.i = imin;
		for (x.j = jmin+1; x.j < jmax-1; x.j++) RCOMPARISON;
	    }
	    if (bottom) {
		x.i = imax - 1;
		for (x.j = jmin+1; x.j < jmax-1; x.j++) RCOMPARISON;
	    }
	    r++;
	}
	if (d < 0) {
	    fprintf(stderr,"%s: no non-zero cells in the grid\n", fct);
	    return ret;
	}
	/* we must now enlarge the rectangle until it is
	   larger than d and check */
	while (r < d) {
	    cell x;
	    int imin = c.i-r, jmin = c.j-r, imax = c.i+r+1, jmax = c.j+r+1;
	    int left = 1, right = 1, top = 1, bottom = 1;
	    if (imin < 0) {imin = 0;top = 0;}
	    if (jmin < 0) {jmin = -1;left = 0;}
	    if (imax > gd->M) {imax = gd->M;bottom = 0;}
	    if (jmax > gd->N) {jmax = gd->N+1;right = 0;}
	    if (left) {
		x.j = jmin;
		for (x.i = imin; x.i < imax; x.i++) RCOMPARISON;
	    }
	    if (right) {
		x.j = jmax-1;
		for (x.i = imin; x.i < imax; x.i++) RCOMPARISON;
	    }
	    if (top) {
		x.i = imin;
		for (x.j = jmin+1; x.j < jmax-1; x.j++) RCOMPARISON;
	    }
	    if (bottom) {
		x.i = imax - 1;
		for (x.j = jmin+1; x.j < jmax-1; x.j++)RCOMPARISON;
	    }
	    r++;
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return ret;
    }
    return ret;
}


grid *gddistances(grid *gd)
{
    static char *fct = "gddistances";
    grid *d;
    cell c;
    PROGRESS_BAR_VARS(gd->M/40);
    if (!(d = gdnew2(REAL_GRID, gd))) return NULL;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd) {
	    PROGRESS_CHECK(c.i);
	    SIGCANCEL;
	    if (!IGD_CELL(gd, c) AND IGD_DATACELL(gd, c)) {
		cell nn = gdnearest_neighbor(gd, c);
		if (nn.i < 0) goto fail;
		RGD_CELL(d, c) = DISTANCE_BETWEEN_CELLS(c, nn) * gd->unitdist;
	    }
	}
    } else if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd) {
	    PROGRESS_CHECK(c.i);
	    SIGCANCEL;
	    if (fabs(RGD_CELL(gd, c)) < EPSILON AND RGD_DATACELL(gd, c)) {
		cell nn = gdnearest_neighbor(gd, c);
		if (nn.i < 0) goto fail;
		RGD_CELL(d, c) = DISTANCE_BETWEEN_CELLS(c, nn) * gd->unitdist;
	    }
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	goto fail;
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    return d;
 fail:
    gddestroy(d);
    return NULL;
}


grid *gddirections(grid *gd)
{
    static char *fct = "gddirections";
    grid *d;
    cell c;
    PROGRESS_BAR_VARS(gd->M/40);
    if (!(d = gdnew2(REAL_GRID, gd))) return NULL;
    if (gd->datatype == INTEGER_GRID) {
	FOREACH(c, gd) {
	    PROGRESS_CHECK(c.i);
	    SIGCANCEL;
	    if (!IGD_CELL(gd, c) AND IGD_DATACELL(gd, c)) {
		cell nn = gdnearest_neighbor(gd, c);
		if (nn.i < 0) goto fail;
		RGD_CELL(d, c) = atan2(c.i-nn.i, nn.j-c.j);
	    }
	}
    } else if (gd->datatype == REAL_GRID) {
	FOREACH(c, gd) {
	    PROGRESS_CHECK(c.i);
	    SIGCANCEL;
	    if (fabs(RGD_CELL(gd, c)) < EPSILON AND RGD_DATACELL(gd, c)) {
		cell nn = gdnearest_neighbor(gd, c);
		if (nn.i < 0) goto fail;
		RGD_CELL(d, c) = atan2(c.i-nn.i, nn.j-c.j);
	    }
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	goto fail;
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    return d;
 fail:
    gddestroy(d);
    return NULL;
}


grid *gdnn(grid *gd)
{
    static char *fct = "gdnn";
    cell c;
    grid *n = NULL;
    PROGRESS_BAR_VARS(gd->M/40);
    if (gd->datatype == INTEGER_GRID) {
	if (!(n = gdcreatecopy(gd, INTEGER_GRID))) return NULL;
	FOREACH(c, gd) {
	    PROGRESS_CHECK(c.i);
	    SIGCANCEL;
	    if (!IGD_CELL(gd, c) AND IGD_DATACELL(gd, c)) {
		cell nn = gdnearest_neighbor(gd, c);
		if (nn.i < 0) goto fail;
		IGD_CELL(n, c) = IGD_CELL(n, nn);
	    }
	}
    } else if (gd->datatype == REAL_GRID) {
	if (!(n = gdcreatecopy(gd, REAL_GRID))) return NULL;
	FOREACH(c, gd) {
	    PROGRESS_CHECK(c.i);
	    SIGCANCEL;
	    if (fabs(RGD_CELL(gd, c)) < EPSILON AND RGD_DATACELL(gd, c)) {
		cell nn = gdnearest_neighbor(gd, c);
		if (nn.i < 0) goto fail;
		RGD_CELL(n, c) = RGD_CELL(n, nn);
	    }
	}
    } else {
	fprintf(stderr,"%s: grid datatype error\n", fct);
	return NULL;
    }
    PROGRESS_BAR_CLOSE;
 cancel:
    return n;
 fail:
    gddestroy(n);
    return NULL;
}

int gdzones(grid *gd, grid *z, double ***tot, int **c, int **k, int *n)
{
    static char *fct = "gdzones";
    cell d;
    int i, j, *itot = NULL;
    hash table;
    if (!gdoverlayable(gd, z)) {
	fprintf(stderr,"%s: the grids are not overlayable\n", fct);
	return 0;
    }
    if (!hash_create(&table, 200)) goto cancel;
    if (!gdcontents(z, &table)) goto cancel;
    *n = hash_count(&table);
    if (!(*tot = (double **)calloc(*n, sizeof(double *))))
	goto fail;
    if (!(itot = (int *)calloc(*n, sizeof(int))))
	goto fail;
    if (!(*c = (int *)calloc(*n, sizeof(int))))
	goto fail;
    if (!(*k = (int *)calloc(*n, sizeof(int))))
	goto fail;
    j = 0;
    for (i = 0; i < table.size; i++) {
	hash_int_item *a = (hash_int_item *)table.table[i];
	while (a) {
	    (*c)[j] = a->key;
	    (*k)[j] = a->value;
	    a->value = j;
	    j++;
	    a = a->next;
	}
    }
    for (i = 0; i < *n; i++) {
	if (!((*tot)[i] = (double *)calloc((*k)[i], sizeof(double))))
	    goto fail;
    }
    FOREACH(d, z) {
	SIGCANCEL;
	if (IGD_DATACELL(z, d) AND GD_DATACELL(gd, d)) {
	    int *value;
	    if (hash_int_lookup(&table, IGD_CELL(z, d), &value)) {		
		(*tot)[*value][itot[*value]] = GD_CELL(gd, d);
		itot[*value]++;
	    }
	}
    }
    for (i = 0; i < *n; i++) 
	(*k)[i] = itot[i];
    free(itot);
    return 1;
 fail:
    fprintf(stderr,"%s: Out of memory!\n", fct);
 cancel:
    if (**tot) {
	for (i = 0; i < *n; i++)
	    if ((*tot)[i]) free((*tot)[i]);
	free(**tot);
    }
    if (itot) free(itot);
    if (*c) free(*c);
    if (*k) free(*k);
    return 0;
 
}
