#ifndef GRID_H
#define GRID_H

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include "hash.h"
#include "vd.h"

#ifdef DEBUG
#define debug(level,msg,x...) \
    {if (debug > level) fprintf(stderr,""msg"", ## x);}
#else
#define debug(level,msg,x...) \
    {;}
#endif

#define ASSERT(test,msg) if (!(test)) { fprintf(stderr,"%s: "msg, fct); goto fail; }

/* do these work if this is a shared library, NO? they work now but then this is not shared? */
extern int debug;
extern int hashmarks;
extern int sigint;

#define SIGCANCEL \
    if (sigint) {sigint = 0;fprintf(stderr,"CANCEL\n");goto cancel;}

#define SQRT2 1.414213562

#define EPSILON 1.0E-12

/*
                   j = 0..N-1  
               ------------------>
              |
i = 0..M-1    |
              |
              |
              V

              ^
              |
              |
              |
            y |
               ------------------>
               x

 */

#define AND &&
#define OR ||

#define EVEN(a) ((a) % 2 == 0)
#define ODD(a) ((a) % 2 == 1)

#define GDINDEX(i, j, N) (j)+(N)*(i)
#define GDINCINDEX(i, j, N) {(j)++; if ((j) >= (N)) {(j) = 0;(i)++;}}

#ifndef _PBMPLUS_H_
#define max(x, y) ((x)>(y) ? (x) : (y))
#define min(x, y) ((x)<(y) ? (x) : (y))
#endif

#define swap(a, b, temp) \
    {(temp) = (a);(a) = (b);(b) = (temp);}

#define round(x) \
    ((x)<0 ? ((long)((x)-0.5)) : ((long)((x)+0.5)))

/* 
   datatypes, HMM.. should have also byte, int, and double
   do we need INTEGER?, i.e., these defines?
*/

/* the two byte integer */
#define INTEGER short
/* the four byte float */
#define REAL float

int gddebug(int dbg);
int gdhashmarks(int h);

/* 
   datatype constants, 
   let's keep this in the order of increasing expressive capacity
   the number is not stored on the disk, the string is
*/
#define INTEGER_GRID 1
#define REAL_GRID 2

#define IGD_AT(gd, i, j) \
    ((gd)->sdata[GDINDEX((i),(j),((gd)->N))])

#define RGD_AT(gd, i, j) \
    ((gd)->fdata[GDINDEX((i),(j),((gd)->N))])

#define INTEGER_STRING "integer"
#define REAL_STRING "real"

#define INTEGER_MIN SHRT_MIN
#define INTEGER_MAX SHRT_MAX

typedef struct {
    int i;
    int j;
} cell;

#define SAME_CELL(a, b) \
    ((a).i == (b).i AND (a).j == (b).j)

#define DISTANCE_BETWEEN_CELLS(a, b) \
    (sqrt(((double)((b).i)-(double)((a).i)) * \
	   ((double)((b).i)-(double)((a).i)) + \
	   ((double)((b).j)-(double)((a).j)) * \
	   ((double)((b).j)-(double)((a).j))))

cell *cellnew();
int celldestroy(cell *c);

#define UP 1
#define UPRIGHT 2
#define RIGHT 3
#define DOWNRIGHT 4
#define DOWN 5
#define DOWNLEFT 6
#define LEFT 7
#define UPLEFT 8

#define FOREACHDIR(dir) \
    for ((dir) = 1; (dir) < 9; (dir)++)

#define DISTANCE_UNIT(dir) \
    (EVEN((dir)) ? SQRT2 : 1)

/* 1 up, 2 up-right, 3 right, 4 down-right, etc */
cell cellmove(cell c, int dir); 

/* 1: b is straigt up from a, ... */
int celldirto(cell a, cell b); 

#define NEXTDIR(d) \
    ((d) > 7 ? 1 : (d) + 1)

#define PREVDIR(d) \
    ((d) > 1 ? (d) - 1 : 8)

#define INVDIR(d) \
    ((d) > 4 ? (d) - 4 : (d) + 4)

typedef struct {
    float minX;
    float maxX;
    float minY;
    float maxY;
} xyrange;

typedef struct {
  char *title;
  int datatype;
  char *filetype;     /* binary */
  int M;
  int N;
  double unitdist;
  xyrange xy;         /* minX is the left edge of first cell in line,
		         maxX is the right edge of the last cell in a
		         line */
  double minval,maxval;
  INTEGER snodata;
  INTEGER *sdata;
  REAL fnodata;
  REAL *fdata;
} grid;

extern grid *mask; /* this affects all methods below if set, SHOULD EFFECT ALSO GRAPHICS? */
/* exceptions:
   gdfilledrect
   gdfloodfill
   gdprint
   gdgrowzones
   gdneighbors HALF WORKING HALF NOT!!
   gdnearestneighbor
   gddirections
*/

int gdsetmask(grid *gd);
grid *gdgetmask();

/* should the nodata have an effect on these...? */

#define GD_CELLIN(gd, c) \
    ((c).i >= 0 AND (c).j >= 0 AND (c).i < (gd)->M AND (c).j < (gd)->N)

#define GD_CELLOUT(gd, c) \
    ((c).i < 0 OR (c).j < 0 OR (c).i >= (gd)->M OR (c).j >= (gd)->N)

#define FOREACH(c, gd) \
    for((c).i=0;(c).i<(gd)->M;(c).i++) for((c).j=0;(c).j<(gd)->N;(c).j++) \
    if (!mask OR (GD_CELLIN(mask, (c)) AND IGD_CELL(mask, (c))))

#define IGD_CELL(gd, c) \
    ((gd)->sdata[GDINDEX(((c).i),((c).j),((gd)->N))])

#define RGD_CELL(gd, c) \
    ((gd)->fdata[GDINDEX(((c).i),((c).j),((gd)->N))])

#define GD_CELL(gd, c) \
    ((gd)->datatype == INTEGER_GRID ? \
     (gd)->sdata[GDINDEX(((c).i),((c).j),((gd)->N))] : \
     (gd)->fdata[GDINDEX(((c).i),((c).j),((gd)->N))])

#define IGD_DATACELL(gd, c) \
    (IGD_CELL((gd), (c)) != (gd)->snodata)

#define RGD_DATACELL(gd, c) \
    (RGD_CELL((gd), (c)) != (gd)->fnodata)

#define IGD_NODATACELL(gd, c) \
    (IGD_CELL((gd), (c)) == (gd)->snodata)

#define IGD_SETNODATACELL(gd, c) \
    (IGD_CELL((gd), (c)) = (gd)->snodata)

#define RGD_NODATACELL(gd, c) \
    (RGD_CELL((gd), (c)) == (gd)->fnodata)

#define RGD_SETNODATACELL(gd, c) \
    (RGD_CELL((gd), (c)) = (gd)->fnodata)

#define GD_DATACELL(gd, c) \
    ((gd)->datatype == INTEGER_GRID ? \
     IGD_CELL((gd), (c)) != (gd)->snodata : \
     RGD_CELL((gd), (c)) != (gd)->fnodata)

#define GD_NODATACELL(gd, c) \
    ((gd)->datatype == INTEGER_GRID ? \
     IGD_CELL((gd), (c)) == (gd)->snodata : \
     RGD_CELL((gd), (c)) == (gd)->fnodata)

int gddestroy(grid *gd);

grid *gdnew(int datatype, int M, int N);
grid *gdnew2(int datatype, grid *gd); /* creates an overlayable grid */
int gdcopy(grid *to, grid *from);
grid *gdcreatecopy(grid *gd, int datatype);
grid *gdopen(char *basename);

int gdread(grid *gd, char *basename, char *ext, int el_size, int byteorder);
int gdsave(grid *gd, char *basename);
int gdwrite(grid *gd, char *basename, char *ext);

int gdsetnodata(grid *gd, double nodata);
double gdgetnodata(grid *gd);

/* a simple progress indicator */
#define PROGRESS_BAR_VARS(step) \
    int tick = (100*step); int steptick = tick;

#define PROGRESS_CHECK(counter) \
    if (hashmarks AND (100*counter) > steptick) \
        {fprintf(stderr,"#");steptick += tick;}

#define PROGRESS_BAR_CLOSE \
    if (hashmarks) fprintf(stderr,"\n");

int gdoverlayable(grid *g1, grid *g2);

int gdgetM(grid *gd);
int gdgetN(grid *gd);
int gddatatype(grid *gd);
double gdunitdist(grid *gd);
double gdminX(grid *gd);
double gdmaxX(grid *gd);
double gdminY(grid *gd);
double gdmaxY(grid *gd);

int gdsetbounds(grid *gd, double unitdist, double minX, double minY);
int gdsetbounds2(grid *gd, double unitdist, double minX, double maxY);
int gdsetbounds3(grid *gd, double unitdist, double maxX, double minY);
int gdsetbounds4(grid *gd, double minX, double maxX, double minY);
int gdsetbounds5(grid *gd, double minX, double maxX, double maxY);
int gdsetbounds6(grid *gd, double minX, double minY, double maxY);
int gdsetbounds7(grid *gd, double maxX, double minY, double maxY);

int gdcopybounds(grid *from, grid *to);

int gdx2j(grid *gd, double x);
int gdy2i(grid *gd, double y);
double gdj2x(grid *gd, int j);
double gdi2y(grid *gd, int i);

int gdset(grid *gd, int i, int j, double x);
int gdset2(grid *gd, int i, int j, int x);
double gdget(grid *gd, int i, int j);
int gdget2(grid *gd, int i, int j);

#define GDGET(gd, c) gdget((gd), (c).i, (c).j)
#define GDGET2(gd, c) gdget2((gd), (c).i, (c).j)

#define GDSET(gd, c, x) gdset((gd), (c).i, (c).j, (x))
#define GDSET2(gd, c, x) gdset2((gd), (c).i, (c).j, (x))

int gdsetminmax(grid *gd);
double gdgetminval(grid *gd);
double gdgetmaxval(grid *gd);

/* where is the min/max value in the grid? */
cell gdgetmin(grid *gd);
cell gdgetmax(grid *gd);

int gdsetall(grid *gd, double x);

/* makes a binary grid: 0: there is no data, 1: there is data */
int gddata(grid *gd);

int gdnot(grid *gd);
int gdandgd(grid *gd1, grid *gd2);
int gdorgd(grid *gd1, grid *gd2);

int gdaddsv(grid *gd, double x);
int gdaddgd(grid *gd1, grid *gd2);
int gdsubgd(grid *gd1, grid *gd2);
int gdmultsv(grid *gd, double x);
int gdmultgd(grid *gd1, grid *gd2);
int gddivsv(grid *gd, double x);
int svdivgd(double x, grid *gd);
int gddivgd(grid *gd1, grid *gd2);
int gdmodulussv(grid *gd, int x);
int svmodulusgd(int x, grid *gd);
int gdmodulusgd(grid *gd1, grid *gd2);
int gdpowersv(grid *gd, double x);
int svpowergd(double x, grid *gd);
int gdpowergd(grid *gd1, grid *gd2);

int gdabs(grid *gd);
int gdacos(grid *gd);
int gdatan(grid *gd);
int gdatan2(grid *gd1, grid *gd2);
int gdceil(grid *gd);
int gdcos(grid *gd);
int gdcosh(grid *gd);
int gdexp(grid *gd);
int gdfloor(grid *gd);
int gdlog(grid *gd);
int gdlog10(grid *gd);
int gdpow(grid *gd, double b);
int gdsin(grid *gd);
int gdsinh(grid *gd);
int gdsqrt(grid *gd);
int gdtan(grid *gd);
int gdtanh(grid *gd);

grid *gdround(grid *gd);

/* 

   all comparison operations silently transform compared REAL grids
   (gd or gd1) into INTEGER grids... is this a polite thing to do? 

   in  Perl this is "fixed" by creating a copy of the original if
   the original is wished to be unchanged

*/

int gdltsv(grid *gd, double x);
int gdgtsv(grid *gd, double x);
int gdlesv(grid *gd, double x);
int gdgesv(grid *gd, double x);
int gdeqsv(grid *gd, double x);
int gdnesv(grid *gd, double x);
int gdcmpsv(grid *gd, double x);

int gdltgd(grid *gd1, grid *gd2);
int gdgtgd(grid *gd1, grid *gd2);
int gdlegd(grid *gd1, grid *gd2);
int gdgegd(grid *gd1, grid *gd2);
int gdeqgd(grid *gd1, grid *gd2);
int gdnegd(grid *gd1, grid *gd2);
int gdcmpgd(grid *gd1, grid *gd2);

int gdminsv(grid *gd, double x);
int gdmaxsv(grid *gd, double x);

int gdmingd(grid *gd1, grid *gd2);
int gdmaxgd(grid *gd1, grid *gd2);

grid *gdcross(grid *a, grid *b);

/* if a then b = c */
int gdif_thensv(grid *a, grid *b, double c);
int gdif_thengd(grid *a, grid *b, grid *c);
/* if a = k then b = v */
int gdzonal_if_then(grid *a, grid *b, int *k, double *v, int n);

int gdbinary(grid *gd);
int gdapplytempl(grid *gd, int *templ, int new_val); /* nodata has no effect */

/* the n-valued map s->d should be sorted to ascending order */
int gdmap(grid *gd, int *s, int *d, int n);

double gdzonesize(grid *gd, cell c);

/* a scanline method, suitable in general for a single area */
grid *gdborders(grid *gd);  

/* a recursive algorithm, suitable for multiple areas */
grid *gdborders2(grid *gd); 

grid *gdareas(grid *gd, int k);

int gdconnect(grid *gd);

int gdnrareas(grid *gd, int connectivity);

grid *gdclip(grid *gd, int i1, int j1, int i2, int j2);
grid *gdjoin(grid *g1, grid *g2);

grid *gdtransform(grid *gd, double tr[], int M, int N);

char *getln(FILE *stream);

grid *a2gd(int datatype, int M, int N, char *infile, int mode);
int gd2a(grid *gd, char *outfile);

int gdline(grid *gd, int i1, int j1, int i2, int j2, double pen);
int gdfilledrect(grid *gd, int i1, int j1, int i2, int j2, double pen);
void gdfilledcircle(grid *gd, int i, int j, int r, int r2, double pen);
/* 
   marks the flooded area in done if given - it may be null or an integer grid 
   connectivity is either 8 or 4 
   dir should be 0 when calling from outside
 */
int igdfloodfill(grid *gd, grid *done, cell c, int color, int connectivity);
int rgdfloodfill(grid *gd, grid *done, cell c, double color, int connectivity);
int gdfloodfill(grid *gd, cell c, int icolor, double rcolor, int connectivity);

int gdprint(grid *gd);
int gdprint1(grid *gd, double **p, int *psize, int quiet, int wc);
/* print only nonzeros, returns i0,j0,x0,i1,j1,x1,...*/

int gdhistogram(grid *gd, double *h, int *c, int n);

/* c is the contents, k is their counts, n is the logical size */
int gdcontents(grid *gd, hash *table);

/*********** ZONAL FUNCTIONS ***********/

int gdzonalcount(grid *gd, grid *zones, hash *table);
int gdzonalsum(grid *gd, grid *zones, hash *table);
int gdzonalmin(grid *gd, grid *zones, hash *table);
int gdzonalmax(grid *gd, grid *zones, hash *table);
int gdzonalmean(grid *gd, grid *zones, hash *table);
int gdzonalvariance(grid *gd, grid *zones, hash *table);

/* 
   each zone in the zones grid is iteratively "grown" to areas
   designated by the (binary) grid grow 
*/
int gdgrowzones(grid *zones, grid *grow, int connectivity);

/* 
   b will contain the neighboring zones of each zone
   b's size is n
   c will contain the ones
   c's size is n
*/
int gdneighbors(grid *gd, hash ***b, int **c, int *n);

/* 
   a buffer zone around cells with value z
   returns a binary grid
*/
grid *gdbufferzone(grid *gd, int z, double w);

/* count, sum, mean, and variance of all cell values (instead of nodata cells)*/
int gdcount(grid *gd);
double gdsum(grid *gd);
double gdmean(grid *gd);
double gdvariance(grid *gd);

/* the returned grid has in the zero cells of gd
   the distance to the nearest non-zero cell in gd
 */
grid *gddistances(grid *gd);

/* the returned grid has in the zero cells of gd
   the direction to the nearest non-zero cell in gd
 */
grid *gddirections(grid *gd);

/* interpolation methods */
/* nearest neighbor, each cell with zero value will get the value of
   its nearest neighbor (in distance sense) 
*/
grid *gdnn(grid *gd); 

/* creates a table of tables, where there is one table for each zone, 
   gd is either the grid from which the values are picked
   z is the zones grid (has to be an integer grid)
   tot is the table of tables
   c is the zone values
   k contains the sizes of the tables in the table
   n is the size of the table of the tables */
int gdzones(grid *gd, grid *z, double ***tot, int **c, int **k, int *n);

#endif
