#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include "vd.h"


/* counterclockwise from Sedgewick: Algorithms in C */
int ccw(point p0, point p1, point p2)
{
    int dx1, dx2, dy1, dy2;
    dx1 = p1.x - p0.x; dy1 = p1.y - p0.y;
    dx2 = p2.x - p0.x; dy2 = p2.y - p0.y;
    if (dx1*dy2 > dy1*dx2) return +1;
    if (dx1*dy2 < dy1*dx2) return -1;
    if ((dx1*dx2 < 0) || (dy1*dy2 < 0)) return -1;
    if ((dx1*dx1+dy1*dy1) < (dx2*dx2+dy2*dy2)) return +1;
    return 0;
}


point_data *pointcreate(double x, 
			double y, 
			int color, 
			int line_width, 
			int symbol, 
			double symbol_height, 
			int symbol_font, 
			char *label, 
			int label_placement)
{
    static char *fct = "pointcreate";
    point_data *p = (point_data *)malloc(sizeof(point_data));
    if (!p) goto fail;

    p->p.x = x;
    p->p.y = y;
    p->color = color;
    p->line_width = line_width;
    p->symbol = 0;
    p->symbol = symbol;
    p->symbol_height = symbol_height;
    p->symbol_font = symbol_font;
    p->label = NULL;
    p->label_placement = label_placement;
    
    if (!(p->label = strdup(label)))
	goto fail;
    
    return p;
 fail:
    fprintf(stderr,"%s: Out of memory!\n", fct);
    pointdestroy(p);
    return NULL;
}


int pointdestroy(point_data *p)
{
    if (p) {
	if (p->label) free(p->label);
	free(p);
    }
    return 1;
}


/* from Sedgewick: Algorithms in C */
int intersect(line l1, line l2)
{
    return ((ccw(l1.p[0], l1.p[1], l2.p[0])
	     *ccw(l1.p[0], l1.p[1], l2.p[1])) <= 0)
	&& ((ccw(l2.p[0], l2.p[1], l1.p[0])
	     *ccw(l2.p[0], l2.p[1], l1.p[1])) <= 0);
}


lseg_data *lsegcreate(double x1, 
		      double y1, 
		      double x2, 
		      double y2,
		      int color,
		      int style,
		      int width,
		      int arrow,
		      double arrow_angle,
		      double arrow_barb,
		      char *label, 
		      int label_placement)
{
    static char *fct = "lsegcreate";
    lseg_data *lseg = (lseg_data *)malloc(sizeof(lseg_data));
    if (!lseg) goto fail;
    lseg->l.p[0].x = x1;
    lseg->l.p[0].y = y1;
    lseg->l.p[1].x = x2;    
    lseg->l.p[1].y = y2;
    lseg->color = color;
    lseg->style = style;
    lseg->width = width;
    lseg->arrow = arrow;
    lseg->arrow_angle = arrow_angle;
    lseg->arrow_barb = arrow_barb;
    lseg->label = NULL;
    lseg->label_placement = label_placement;
    
    if (!(lseg->label = strdup(label)))
	goto fail;
    
    return lseg;
 fail:
    fprintf(stderr,"%s: Out of memory!\n", fct);
    lsegdestroy(lseg);
    return NULL;
}


int lsegdestroy(lseg_data *lseg)
{
    if (lseg) {
	if (lseg->label) free(lseg->label);
	free(lseg);
    }
    return 1;
}


polygon_data *polygoncreate(int n, 
			    double *x, 
			    double *y,			    
			    int color,
			    int line_style,
			    int line_width,
			    int fill_style,
			    double fill_angle,
			    double fill_sepn,
			    double fill_phase,
			    char *label, 
			    int label_placement)
{
    static char *fct = "polygoncreate";
    polygon_data *polygon = (polygon_data *)malloc(sizeof(polygon_data));
    if (!polygon) goto fail;
    polygon->n = 0;
    polygon->points = NULL;
    polygon->color = color;
    polygon->line_style = line_style;
    polygon->line_width = line_width;
    polygon->fill_style = fill_style;
    polygon->fill_angle = fill_angle; 
    polygon->fill_sepn = fill_sepn;
    polygon->fill_phase = fill_phase;
    polygon->label = NULL;
    polygon->label_placement = label_placement;

    if (!(polygon->points = (point *)calloc(n, sizeof(point)))) 
	goto fail;

    if (!(polygon->label = strdup(label)))
	goto fail;
    {
	int i;
	for (i = 0; i < n; i++) {
	    polygon->points[i].x = x[i];
	    polygon->points[i].y = y[i];
	}
    }
    polygon->n = n;
    
    return polygon;
 fail:
    fprintf(stderr,"%s: Out of memory!\n", fct);
    polygondestroy(polygon);
    return NULL;
}


int polygondestroy(polygon_data *polygon)
{
    if (polygon) {
	if (polygon->points) free(polygon->points);
	if (polygon->label) free(polygon->label);
	free(polygon);
    }
    return 1;
}

/* 
   is point inside a convex polygon?  
   true if line t->infinity intersect
   with one and only one lines of the polygon
*/
int insideconvex(point t, point *p, int N)
{
    int i, count = 0;
    line l;
    line *lp;
    l.p[0] = l.p[1] = t; l.p[1].x = FLT_MAX;
    for (i = 0; i < N-1; i++) {
	lp = (line *)&p[i];
	if (intersect(l, *lp)) {
	    count++;
	    if (count > 1) break;
	}
    }
    return count == 1;
}


vector_data *vdnull() 
{
    static char *fct = "vdnull";
    vector_data *vd;
    if (!(vd = (vector_data *)malloc(sizeof(vector_data)))) goto fail;
    vd->name = NULL;
    vd->id = 0;
    vd->next = NULL;

    vd->n_points = 0;
    vd->pointtablesize = 0;
    vd->n_lsegs = 0;
    vd->lsegtablesize = 0;
    vd->n_polygons = 0;
    vd->polygontablesize = 0;
    vd->pointtable = NULL;
    vd->lsegtable = NULL;
    vd->polygontable = NULL;

    vd->label_color = 0;
    vd->label_line_width = 0;
    vd->label_font = 0;
    vd->label_height = 0;

    return vd;
 fail:
    fprintf(stderr,"%s: Out of memory!\n", fct);
    vddestroy(vd);
    return NULL;
}


vector_data *vdcreate(char *name, 
		      int id, 
		      int label_color, 
		      int label_line_width,
		      int label_font, 
		      int label_height)
{
    static char *fct = "vdcreate";
    vector_data *vd = (vector_data *)malloc(sizeof(vector_data));
    if (!vd) goto fail;

    vd->name = NULL;
    vd->id = id;
    vd->next = NULL;

    vd->pointtable = NULL;
    vd->lsegtable = NULL;
    vd->polygontable = NULL;

    vd->label_color = label_color;
    vd->label_line_width = label_line_width;
    vd->label_font = label_font;
    vd->label_height = label_height;

    if (!(vd->name = strdup(name)))
	goto fail;
    
    vd->n_points = 0;
    if (!(vd->pointtable = (point_data **)calloc(50,sizeof(point_data *)))) 
	goto fail;
    vd->pointtablesize = 50;

    vd->n_lsegs = 0;
    if (!(vd->lsegtable = (lseg_data **)calloc(50,sizeof(lseg_data *)))) 
	goto fail;
    vd->lsegtablesize = 50;

    vd->n_polygons = 0;
    if (!(vd->polygontable = (polygon_data **)calloc(50,sizeof(polygon_data *)))) 
	goto fail;
    vd->polygontablesize = 50;

    return vd;
 fail:
    fprintf(stderr,"%s: Out of memory!\n", fct);
    vddestroy(vd);
    return NULL;
}


int vddestroy(vector_data *vd)
{
    if (vd) {
	if (vd->next) vddestroy(vd->next);
	if (vd->pointtable) {
	    int i;
	    for (i = 0; i < vd->pointtablesize; i++) {
		if (vd->pointtable[i]) pointdestroy(vd->pointtable[i]);
	    }
	    free(vd->pointtable);
	}
	if (vd->lsegtable) {
	    int i;
	    for (i = 0; i < vd->lsegtablesize; i++) {
		if (vd->lsegtable[i]) lsegdestroy(vd->lsegtable[i]);
	    }
	    free(vd->lsegtable);
	}
	if (vd->polygontable) {
	    int i;
	    for (i = 0; i < vd->polygontablesize; i++) {
		if (vd->polygontable[i]) polygondestroy(vd->polygontable[i]);
	    }
	    free(vd->polygontable);
	}
	if (vd->name) free(vd->name);
	free(vd);
    }
    return 1;
}


char *vdget_name(vector_data *vd)
{
    return vd->name;
}

int vdget_id(vector_data *vd)
{
    return vd->id;
}


vector_data *vdnext(vector_data *vd)
{
    return vd->next;
}


int vdbreaklist(vector_data *vd)
{
    vd->next = NULL;
    return 1;
}


int vdaddvd(vector_data *root, vector_data *vd)
{
    while (root->next) root = root->next;
    root->next = vd;
    vd->next = NULL;
    return 1;
}


int vdaddpoint(vector_data *vd, point_data *point)
{
    static char *fct = "vdaddpoint";
    if (vd->n_points == vd->pointtablesize) {
	if (!(vd->pointtable = 
	      (point_data **)realloc(vd->pointtable, 
				     (vd->pointtablesize+50)*sizeof(point_data *))))
	    goto fail;
	{
	    int i;
	    for (i = vd->pointtablesize; i < vd->pointtablesize+50; i++) 
		vd->pointtable[i] = NULL;
	}
	vd->pointtablesize += 50;
    }
    vd->pointtable[vd->n_points] = point;
    vd->n_points++;
    return 1;
 fail:
    fprintf(stderr,"%s: Out of memory!\n", fct);
    return 0;
}


int vdaddlseg(vector_data *vd, lseg_data *lseg)
{
    static char *fct = "vdaddlseg";
    if (vd->n_lsegs == vd->lsegtablesize) {
	if (!(vd->lsegtable = 
	      (lseg_data **)
	      realloc(vd->lsegtable, 
		      (vd->lsegtablesize+50)*sizeof(lseg_data *))))
	    goto fail;
	{
	    int i;
	    for (i = vd->lsegtablesize; i < vd->lsegtablesize+50; i++) 
		vd->lsegtable[i] = NULL;
	}
	vd->lsegtablesize += 50;
    }
    vd->lsegtable[vd->n_lsegs] = lseg;
    vd->n_lsegs++;
    return 1;
 fail:
    fprintf(stderr,"%s: Out of memory!\n", fct);
    return 0;
}


int vdaddpolygon(vector_data *vd, polygon_data *polygon)
{
    static char *fct = "vdaddpolygon";
    if (vd->n_polygons == vd->polygontablesize) {
	if (!(vd->polygontable = 
	      (polygon_data **)realloc(vd->polygontable, 
				       (vd->polygontablesize+50)*sizeof(polygon_data *))))
	    goto fail;
	{
	    int i;
	    for (i = vd->polygontablesize; i < vd->polygontablesize+50; i++) 
		vd->polygontable[i] = NULL;
	}
	vd->polygontablesize += 50;
    }
    vd->polygontable[vd->n_polygons] = polygon;
    vd->n_polygons++;
    return 1;
 fail:
    fprintf(stderr,"%s: Out of memory!\n", fct);
    return 0;
}



