#ifndef VD_H
#define VD_H

/* 
   all data to constructors 
   label placement + vert. align. in plotting
   some explanations from pgplot.doc 
*/

typedef struct {
    float x;
    float y;
} point;

int ccw(point p0, point p1, point p2);

typedef struct {
    point p;
    int color;           /* index to the colortable */
    int line_width;      /* width of line, in units of 0.005 inch (0.13 mm)
			    in range 1-201. */
    int symbol;          /* -1, -2  : a single dot (diameter = current
                		      line width).
			    -3..-31 : a regular polygon with ABS(SYMBOL)
			              edges (style set by current fill style).
			    0..31   : standard marker symbols.
			    32..127 : ASCII characters (in current font).
			              e.g. to use letter F as a marker, let
			    SYMBOL = ICHAR('F'). 
			    > 127  :  a Hershey symbol number. */
    float symbol_height;/* The default character size is 1.0, 
			   corresponding to a character height about 
			   1/40 the height of the view surface. */
    int symbol_font;    /* 1: a simple single-stroke font ("normal" font)
			   2: roman font
			   3: italic font
			   4: script font */
    char *label;
    int label_placement;
} point_data;

point_data *pointcreate(double x, 
			double y, 
			int color, 
			int line_width, 
			int symbol, 
			double symbol_height, 
			int symbol_font, 
			char *label, 
			int label_placement);
int pointdestroy(point_data *p);

typedef struct {
    point p[2];
} line;

int intersect(line l1, line l2);

typedef struct {
    line l;
    int color;           /* index to the colortable */
    int style;           /* 1 (full line), 2 (dashed), 3 (dot-dash-dot-dash), 
			    4 (dotted), 5 (dash-dot-dot-dot) */
    int width;           /* width of line, in units of 0.005 inch (0.13 mm)
			    in range 1-201. */
    int arrow;           /* 0 no arrow head, 1 filled, 2 outline */
    float arrow_angle;   /* the acute angle of the arrow point, in degrees;
			    angles in the range 20.0 to 90.0 give reasonable
			    results. Use 45.0. */
    float arrow_barb;    /* the fraction of the triangular arrow-head that
			    is cut away from the back. 0.0 gives a triangular
			    wedge arrow-head; 1.0 gives an open >. Values 0.3
			    to 0.7 give reasonable results. Use 0.3. */
    char *label;
    int label_placement;
} lseg_data;

lseg_data *lsegcreate(double x1, 
		      double y1, 
		      double x2, 
		      double y2,
		      int color,
		      int style,
		      int width,
		      int arrow,
		      double arrow_angle,
		      double arrow_barb,
		      char *label, 
		      int label_placement);
int lsegdestroy(lseg_data *lseg);

typedef struct {
    point *points;         /* it is the users responsibility to 
			      make sure that points[0] == points[n-1] */
    int n;
    int color;             /* index to the colortable */
    int line_style;        /* 1 (full line), 
                              2 (dashed), 
                              3 (dot-dash-dot-dash), 
			      4 (dotted), 
                              5 (dash-dot-dot-dot) */
    int line_width;        /* width of line, in units of 0.005 inch
                              (0.13 mm) in range 1-201. */
    int fill_style;        /* FS = 1 => solid (default)
			      FS = 2 => outline
			      FS = 3 => hatched
			      FS = 4 => cross-hatched */
    float fill_angle;      /* the angle the hatch lines make with the
			      horizontal, in degrees, increasing
			      counterclockwise (this is an angle on
			      the view surface, not in
			      world-coordinate space). */
    float fill_sepn;       /* the spacing of the hatch lines. The unit
			      spacing is 1 percent of the smaller of
			      the height or width of the view
			      surface. This should not be zero. */
    float fill_phase;      /* a real number between 0 and 1; the hatch
			      lines are displaced by this fraction of
			      SEPN from a fixed reference.  Adjacent
			      regions hatched with the same PHASE have
			      contiguous hatch lines. To hatch a
			      region with alternating lines of two
			      colors, fill the area twice, with
			      PHASE=0.0 for one color and PHASE=0.5
			      for the other color. */
    char *label;
    int label_placement;
} polygon_data;

polygon_data *polygoncreate(int n, 
			    double *x, 
			    double *y,			    
			    int color,
			    int line_style,
			    int line_width,
			    int fill_style,
			    double fill_angle,
			    double fill_sepn,
			    double fill_phase,
			    char *label, 
			    int label_placement);

int polygondestroy(polygon_data *polygon);

int inside(point t, point *p, int N);

struct _vector_data {
    char *name;
    int id;
    struct _vector_data *next;
    int label_color;         /* index to the colortable */
    int label_line_width;    /* width of line, in units of 0.005 inch
				(0.13 mm) in range 1-201. */
    int label_font;          /* 1: a simple single-stroke font ("normal" font) 
                                2: roman font 
                                3: italic font 
                                4: script font */
    int label_height;        /* The default character size is 1.0,
				corresponding to a character height
				about 1/40 the height of the view
				surface. */

    /* points */
    int pointtablesize;
    int n_points;
    point_data **pointtable;
    
    /* line segments */
    int lsegtablesize;
    int n_lsegs;
    lseg_data **lsegtable;

    /* polylines */

    /* polygons */
    int polygontablesize;
    int n_polygons;
    polygon_data **polygontable;
};

#define vector_data struct _vector_data

vector_data *vdnull();
vector_data *vdcreate(char *name, 
		      int id, 
		      int label_color, 
		      int label_line_width,
		      int label_font, 
		      int label_height);
int vddestroy(vector_data *vd);                  /* destroys the whole list 
						    from this node on */

char *vdget_name(vector_data *vd);
int vdget_id(vector_data *vd);

vector_data *vdnext(vector_data *vd);
int vdbreaklist(vector_data *vd);                /* clears next and prev */
int vdaddvd(vector_data *root, vector_data *vd); /* adds vd to the end of the 
						    list root, sets next and 
						    prev in list and in vd */

int vdaddpoint(vector_data *vd, point_data *point);
int vdaddlseg(vector_data *vd, lseg_data *lseg);
int vdaddpolygon(vector_data *vd, polygon_data *polygon);

#endif
